/* Description:
 *      This file implements the Secure Hash Signature Standard
 *      algorithms as defined in the National Institute of Standards
 *      and Technology Federal Information Processing Standards
 *      Publication (FIPS PUB) 180-1 published on April 17, 1995, 180-2
 *      published on August 1, 2002, and the FIPS PUB 180-2 Change
 *      Notice published on February 28, 2004.
 *
 *      A combined document showing all algorithms is available at
 *              http://csrc.nist.gov/publications/fips/
 *              fips180-2/fips180-2withchangenotice.pdf
 *
 * xmlfy authors note:
 *   This work is derived from RFC4634 implementing just the
 *   "SHA512 Message-Digest Algorithm", with only trivial
 *   peripheral customisations to work with xmlfy.
 *     -- Arthur Gouros.
 */

#ifndef _SHA_H_
#define _SHA_H_

#include "digest.h"

#define shaSuccess 0
#define shaStateError 1
#define SHA512_Message_Block_Size 128
#define SHA512HashSize 64

/* SHA512 Context */
typedef struct SHA512Context {
    UINT4 Intermediate_Hash[SHA512HashSize/4]; /* Message Digest  */
    UINT4 Length[4];                 /* Message length in bits */
    UINT2 Message_Block_Index;  /* Message_Block array index */
    UINT1 Message_Block[SHA512_Message_Block_Size];
    int Computed;                       /* Is the digest computed?*/
    int Corrupted;                      /* Is the digest corrupted? */
} SHA512Context;

/* Function Prototypes */
/* SHA-512 */
extern void SHA512Reset(SHA512Context *);
extern int SHA512Input(SHA512Context *, const UINT1 *bytes, unsigned int bytecount);
extern int SHA512Result(SHA512Context *, UINT1 Message_Digest[SHA512HashSize]);

#endif /* _SHA_H_ */
