/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2013, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * noxml - Convert reserved XML characters to their escaped equivalent    *
 *         and/or convert the input to wide UTF encoding.                 *
 *         This is almost a standalone function with no XML handling.     *
 *         Used primarily to convert test results for reporting purposes. *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "../xmlfy.h"
#include "../utf_io/utf_io.h"

/* function templates */
void noxml();

void noxml() 
{
  /* No XML just convert the input stream's escaped characters 
   * and/or encoding. */
  long i, j;
  int  m;

  /* try and get the encoding */
  if (!sw_ascii_to_utf)
    utf_get_encoding();
  else
    input_record.charwidth = UTF_8_INCREMENTER;

  if (sw_bom)
    utf_print_bom(STDOUT_FD);

  while (!utf_flag_EOF)
  {
    input_record.length = 0;

    do
    {
      input_record.line[input_record.length++] = utf_getchar(STDIN_FD);

      /* mem checks */
      if (input_record.length > memseg_input_record)
        memadd_input_record();
    }
    while (!utf_flag_EOF && input_record.length < LINESIZE / 2);

    if (utf_flag_EOF)
    {
      if (input_record.length > 1)
        input_record.length--; /* do-while logic adds EOF so ignore it */
      else
        break; /* don't process EOF */
    }

    output_record.length = 0;

    for (i = 0 ; i < input_record.length ; i += input_record.charwidth)
    {
      /* XML characters that need escaping */
      if (utf_char_equals(i, '<') && sw_char_escape_lt)
        strcat_output_record((unsigned char *) "&lt;");
      else if (utf_char_equals(i, '>') && sw_char_escape_gt)
        strcat_output_record((unsigned char *) "&gt;");
      else if (utf_char_equals(i, '&') && sw_char_escape_amp)
        strcat_output_record((unsigned char *) "&amp;");
      else if (utf_char_equals(i, '"') && sw_char_escape_quot)
        strcat_output_record((unsigned char *) "&quot;");
      else if (utf_char_equals(i, '|') && sw_char_escape_brvbar)
        strcat_output_record((unsigned char *) "&brvbar;");
      else if (utf_char_equals(i, '\'') && sw_char_escape_apos)
        strcat_output_record((unsigned char *) "&apos;");
      else
      {
        if (output_record.encoding == UTF_8)
        {
          output_record.line[output_record.length++] = input_record.line[i];
        }
        else if (input_record.encoding == UTF_8) /* and wide encoding */
        {
          if (input_record.line[i])
            strncat_output_record(&input_record.line[i], 1);
          else
          {
            /* print a null. NOTE: strncat does not handle nulls */
            output_record.line[output_record.length++] = 0;
            output_record.line[output_record.length++] = 0;
            if (output_record.encoding == UTF_32BE || output_record.encoding == UTF_32LE)
            {
              output_record.line[output_record.length++] = 0;
              output_record.line[output_record.length++] = 0;
            }
          }
        }
        else 
        {
          /* Wide encoding (UTF-16 or UTF-32).
           * NOTE: We don't do code point error correction and assume the
           * wide input chars fall within their correct boundaries.
           * Any input stream corruption or purposely formed delimiter
           * strings that match across code point boundaries will 
           * cause the ouput to get out of synchronisation and
           * warrant some investigation by the user. */
           for (j = 0 ; j < output_record.charwidth && i + j < input_record.length ; j++)
             output_record.line[output_record.length++] = input_record.line[i + j];
        }

        /* mem checks */
        if (output_record.length > memseg_output_record)
          memadd_output_record(output_record.length);
      }
    }
    m = write(STDOUT_FD, output_record.line, output_record.length);
  }

  /* Fudge to shut up gcc >= 4.6.
   * For security reasons (buffer overflow), certain distros 
   * don't allow you to cast to (void) the return value of write(),
   * so I assign it to an unused variable m, but now the new version 
   * of gcc warns that variable m is declared but not used. */
  i = m;
}
