/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2013, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert to XML on the fly.                                     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#ifndef SCHEMA_H
#define SCHEMA_H

#include "../xmlfy.h"
#include "../utf_io/utf_io.h"

extern int schema_fd;
extern int flag_comments, flag_char;
extern unsigned char *start_element_name;

/* scratch memory for processing schema files */
typedef struct {
  unsigned char *value; /* token name */
  int match_control;    /* wildcard token */
  int choice;           /* XSD choice handling */
} struct_tmp_field;

extern struct_tmp_field *tmp_field;
extern long tmp_nfields;
extern long memseg_tmp_nfields;


/* function templates */
extern void load_dtd();
extern void load_rnc();
extern void load_xsd();
extern void create_elements();
extern void build_element_hierachy();
extern long purge_comments_blanklines(int encoding);
extern int  match_control_token(unsigned char c1);
extern int  is_legal_xml_element_name_char(unsigned char c1);
extern void schema_args();
extern void helper_calc_tree_paths();
extern void alloc_tmp_field();
extern void memadd_tmp_nfields();
extern void debug_print_helper_xpaths(char *msg, int i_helper, int h_helper);
/* function templates external to this object */
extern void increase_nlevels(int newlevel);

#endif /* SCHEMA_H */
