/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2013, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert to XML on the fly.                                     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "schema.h"

/* function templates */
void schema_args();


void schema_args()
{
  /* Look for xmlfy args specified inside the schema file */
  long i, j, p0, p1, prev_p1, end;
  int  args_token_length, characters, sgl_quoted, dbl_quoted;
  char *args_token;

  /* initialise */
  end               = 0;
  args_token_length = 0;
  args_token        = NULL;

  if (sw_schema == SCHEMA_DTD || sw_schema == SCHEMA_XSD)
  {
    args_token_length = DTD_ARGS_TOKEN_LENGTH;
    args_token        = DTD_ARGS_TOKEN;
  }
  else if (sw_schema == SCHEMA_RNC)
  {
    args_token_length = RNC_ARGS_TOKEN_LENGTH;
    args_token        = RNC_ARGS_TOKEN;
  }

  /* look for line in schema file with args string token and only process once */
  if (args_token_length
      && input_record.length > args_token_length
      && xstrncmp((unsigned char *) args_token, input_record.line, args_token_length) == 0
      && schema_argc == 0)
  {
    if (sw_debug)
      fprintf(stderr, "DEBUG: Schema: schema_args: Found more args to process.%s", newline);

    /* locate end */
    if (sw_schema == SCHEMA_DTD || sw_schema == SCHEMA_XSD)
    {
      for (end = DTD_ARGS_TOKEN_LENGTH ; end < input_record.length ; end++)
      {
        if (end + 2 < input_record.length && input_record.line[end] == '-' && input_record.line[end + 1] == '-' && input_record.line[end + 2] == '>')
        {
          break;
        }
      }
    }
    else if (sw_schema == SCHEMA_RNC)
      end = input_record.length;

    /* break token into argv components */
    schema_argc = 1; /* validate_args() ignores argv[0] */
    schema_argv = (unsigned char **) malloc(schema_argc * sizeof(unsigned char *));
    if (schema_argv == NULL) out_of_mem();
    p0 = 0; p1 = 0, prev_p1 = 0;
    characters = FALSE; sgl_quoted = FALSE; dbl_quoted = FALSE;
    for (i = args_token_length ; i < end ; i++)
    {
      /* Note this portion of code was originally implemented as a switch() but the
       * "fall through" feature got too complicated so it was re-factored 
       * with if/continue code. */

      /* space separator */
      if (input_record.line[i] == ' ' || input_record.line[i] == '\t')
      {
        if (!sgl_quoted && !dbl_quoted)
        {
          characters = FALSE;
          continue; /* continue for/i */
        }
      }

      /* double quote */
      if (input_record.line[i] == '"')
      {
        if (i && input_record.line[i - 1] != '\\') 
        {
          if (dbl_quoted)
            dbl_quoted = FALSE;
          else
            dbl_quoted = TRUE;
          continue; /* continue for/i */
        }
      }

      /* single quote */
      if (input_record.line[i] == '\'')
      {
        if (i && input_record.line[i - 1] != '\\') 
        {
          if (sgl_quoted)
            sgl_quoted = FALSE;
          else
            sgl_quoted = TRUE;
          continue; /* continue for/i */
        }
      }

      /* data */
      if (!characters)
      {
        if (p1 > 0)
        {
          schema_argv = (unsigned char **) realloc(schema_argv, (schema_argc + 1) * sizeof(unsigned char *));
          if (schema_argv == NULL) out_of_mem();
          schema_argv[schema_argc] = (unsigned char *) malloc(((p1 - p0) + 2) * sizeof(unsigned char));
          if (schema_argv[schema_argc] == NULL) out_of_mem();
          for (j = p0 ; j <= p1 ; j++)
            schema_argv[schema_argc][j - p0] = input_record.line[j];
          schema_argv[schema_argc][j - p0] = '\0';
          prev_p1 = p1;
          schema_argc++;
        }
        p0 = i;
      }
      p1 = i;
      characters = TRUE;
    } /* end for/i */


    if (p1 > prev_p1)
    {
      schema_argv = (unsigned char **) realloc(schema_argv, (schema_argc + 1) * sizeof(unsigned char *));
      if (schema_argv == NULL) out_of_mem();
      schema_argv[schema_argc] = (unsigned char *) malloc(((p1 - p0) + 2) * sizeof(unsigned char));
      if (schema_argv[schema_argc] == NULL) out_of_mem();
      for (j = p0 ; j <= p1 ; j++)
        schema_argv[schema_argc][j - p0] = input_record.line[j];
      schema_argv[schema_argc][j - p0] = '\0';
      schema_argc++;
    }

    /* scan through args and remove the escape character from escaped quotes */
    for (i = 1 ; i < schema_argc ; i++)
    {
      p1 = xstrlen(schema_argv[i]);
      for (j = 0 ; j < p1 - 1 ; j++)
      {
        if (schema_argv[i][j] == '\\' && (schema_argv[i][j + 1] == '"' || schema_argv[i][j + 1] == '\''))
        {
          for (p0 = j ; p0 < p1 - 1 ; p0++)
            schema_argv[i][p0] = schema_argv[i][p0 + 1];
          schema_argv[i][p0] = '\0';
        }
      }
    }

    /* validate these extra args */
    if (schema_argc > 1)
      validate_args(schema_argc, schema_argv, ARGS_SCHEMA);
  }
}
