/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2013, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert to XML on the fly.                                     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "utf_io.h"

/* function templates */
void utf_get_encoding();


void utf_get_encoding()
{
  /* try figure out the encoding used in the input stream 
   * NOTE: This routine is only used by the --unxml and --noxml options
   * as it assumes an XML formatted input stream. */

  if (!utf_read_bom(STDIN_FD) && !utf_flag_EOF)
  {
    /* XML documents usually start with <? */
    if (utf_read > 1
        && utf_buf[0] == '<' && utf_buf[1] == '?')
    {
      sw_encoding = UTF_8;
      utf_set_encoding();
    }
    else if (utf_read > 3
        && utf_buf[0] == 0 && utf_buf[1] == '<'
        && utf_buf[2] == 0 && utf_buf[3] == '?')
    {
      sw_encoding = UTF_16BE;
      utf_set_encoding();
    }
    else if (utf_read > 3
        && utf_buf[0] == '<' && utf_buf[1] == 0
        && utf_buf[2] == '?' && utf_buf[3] == 0)
    {
      sw_encoding = UTF_16LE;
      utf_set_encoding();
    }
    else if (utf_read > 7
        && utf_buf[0] == 0 && utf_buf[1] == 0
        && utf_buf[2] == 0 && utf_buf[3] == '<'
        && utf_buf[4] == 0 && utf_buf[5] == 0
        && utf_buf[6] == 0 && utf_buf[7] == '?')
    {
      sw_encoding = UTF_32BE;
      utf_set_encoding();
    }
    else if (utf_read > 7
        && utf_buf[0] == '<' && utf_buf[1] == 0
        && utf_buf[2] == 0   && utf_buf[3] == 0
        && utf_buf[4] == '?' && utf_buf[5] == 0
        && utf_buf[6] == 0   && utf_buf[7] == 0)
    {
      sw_encoding = UTF_32LE;
      utf_set_encoding();
    }
    /* else .... user specified options prevail */
  }
}
