/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2013, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert to XML on the fly.                                     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#ifndef UTF_IO_H
#define UTF_IO_H

#include "../xmlfy.h"


/* function templates */
extern void utf_getchar_init(int fd);
extern unsigned char utf_getchar(int fd);
extern void utf_print(int fd, unsigned char *buf, long buf_bytes);
extern int  utf_read_bom(int fd);
extern void utf_print_bom(int fd);
/* pointer function targets */
extern void strcpy_utf8(int level, const unsigned char *s_src);
extern void strncpy_utf8(const unsigned char *s_src, long n);
extern void strcat_utf8(const unsigned char *s_src);
extern void strncat_utf8(const unsigned char *s_src, long n);
extern void strcpy_utf16be(int level, const unsigned char *s_src);
extern void strncpy_utf16be(const unsigned char *s_src, long n);
extern void strcat_utf16be(const unsigned char *s_src);
extern void strncat_utf16be(const unsigned char *s_src, long n);
extern void strcpy_utf16le(int level, const unsigned char *s_src);
extern void strncpy_utf16le(const unsigned char *s_src, long n);
extern void strcat_utf16le(const unsigned char *s_src);
extern void strncat_utf16le(const unsigned char *s_src, long n);
extern void strcpy_utf32be(int level, const unsigned char *s_src);
extern void strncpy_utf32be(const unsigned char *s_src, long n);
extern void strcat_utf32be(const unsigned char *s_src);
extern void strncat_utf32be(const unsigned char *s_src, long n);
extern void strcpy_utf32le(int level, const unsigned char *s_src);
extern void strncpy_utf32le(const unsigned char *s_src, long n);
extern void strcat_utf32le(const unsigned char *s_src);
extern void strncat_utf32le(const unsigned char *s_src, long n);
/* UTF/ascii functions */
extern int utf_xstrncmp(const unsigned char *utf_s, long utf_s_length, int utf_s_encoding, const unsigned char *ascii_s, long n);
extern int utf_char_equals(long i, unsigned char byte_c);
extern int utf_char_equals_2(long i, unsigned char byte_c, const unsigned char *utf_s, long utf_s_length, int utf_s_encoding);
extern int utf_char_equals_level1(long i, unsigned char byte_c);
extern int utf_isspace(long i);
extern int utf_iscomposite(long i);
extern void utf_set_encoding();
extern void utf_get_encoding();
extern long utf_get_first_word_length(const unsigned char *utf_s, long utf_s_length, int utf_s_encoding);

#endif /* UTF_IO_H */
