/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2013, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert to XML on the fly.                                     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "utf_io.h"

/* function templates */
void utf_set_encoding();


void utf_set_encoding()
{
  /* Set UTF handlers and populate the UTF delimiter strings 
   * in requested/detected encoding */
  int level, i, j, k, byte_value;
  char value[4];

  /* set the output encoding */
  if (sw_encoding == UTF_8)
  {
    strcpy_output_record    = strcpy_utf8;
    strncpy_output_record   = strncpy_utf8;
    strcat_output_record    = strcat_utf8;
    strncat_output_record   = strncat_utf8;
    input_record.encoding   = UTF_8;
    input_record.charwidth  = UTF_8_INCREMENTER;
    output_record.encoding  = UTF_8;
    output_record.charwidth = UTF_8_INCREMENTER;

  }
  else if (sw_encoding == UTF_16BE)
  {
    strcpy_output_record    = strcpy_utf16be;
    strncpy_output_record   = strncpy_utf16be;
    strcat_output_record    = strcat_utf16be;
    strncat_output_record   = strncat_utf16be;
    input_record.encoding   = UTF_16BE;
    input_record.charwidth  = UTF_16_INCREMENTER;
    output_record.encoding  = UTF_16BE;
    output_record.charwidth = UTF_16_INCREMENTER;
  }
  else if (sw_encoding == UTF_16LE)
  {
    strcpy_output_record    = strcpy_utf16le;
    strncpy_output_record   = strncpy_utf16le;
    strcat_output_record    = strcat_utf16le;
    strncat_output_record   = strncat_utf16le;
    input_record.encoding   = UTF_16LE;
    input_record.charwidth  = UTF_16_INCREMENTER;
    output_record.encoding  = UTF_16LE;
    output_record.charwidth = UTF_16_INCREMENTER;
  }
  else if (sw_encoding == UTF_32BE)
  {
    strcpy_output_record    = strcpy_utf32be;
    strncpy_output_record   = strncpy_utf32be;
    strcat_output_record    = strcat_utf32be;
    strncat_output_record   = strncat_utf32be;
    input_record.encoding   = UTF_32BE;
    input_record.charwidth  = UTF_32_INCREMENTER;
    output_record.encoding  = UTF_32BE;
    output_record.charwidth = UTF_32_INCREMENTER;
  }
  else if (sw_encoding == UTF_32LE)
  {
    strcpy_output_record    = strcpy_utf32le;
    strncpy_output_record   = strncpy_utf32le;
    strcat_output_record    = strcat_utf32le;
    strncat_output_record   = strncat_utf32le;
    input_record.encoding   = UTF_32LE;
    input_record.charwidth  = UTF_32_INCREMENTER;
    output_record.encoding  = UTF_32LE;
    output_record.charwidth = UTF_32_INCREMENTER;
  }

  /* set input_record.charwidth if doing ASCIItoUTF conversion */
  if (sw_ascii_to_utf)
  {
    input_record.encoding  = UTF_8;
    input_record.charwidth = UTF_8_INCREMENTER;
  }

  /* assign correctly encoded delimiters 
   * NOTE: memory to handle up to UTF-32 is already allocated */
  for (level = 1 ; level < nlevels ; level++)
  {
    for (i = 0 ; i < delimiter[level]->ndelimiters ; i++)
    {
      if (delimiter[level]->byte_mode[i])
      {
        /* convert any escaped decimal characters to their equivalent value */
        k = 0;
        for (j = 0 ; j < delimiter[level]->arg_length[i] ; j++)
        {
          if (delimiter[level]->arg_d[i][j] == '\\')
          {
            if (j + 3 < delimiter[level]->arg_length[i]
               && isdigit(delimiter[level]->arg_d[i][j + 1])
               && isdigit(delimiter[level]->arg_d[i][j + 2])
               && isdigit(delimiter[level]->arg_d[i][j + 3]))
            {
              value[0] = delimiter[level]->arg_d[i][j + 1];
              value[1] = delimiter[level]->arg_d[i][j + 2];
              value[2] = delimiter[level]->arg_d[i][j + 3];
              value[3] = '\0';
              byte_value = atoi(value);
              if (byte_value >= 0 && byte_value < 256)
              {
                delimiter[level]->utf_d[i][k++] = (unsigned char) byte_value;
                j += 3;
              }
              else
              {
                delimiter[level]->utf_d[i][k++] = delimiter[level]->arg_d[i][j];
              }
            }
            else if (j + 1 < delimiter[level]->arg_length[i] && delimiter[level]->arg_d[i][j + 1] == '\\')
            {
              j++;
              delimiter[level]->utf_d[i][k++] = delimiter[level]->arg_d[i][j];
            }
          }
          else
          {
            delimiter[level]->utf_d[i][k++] = delimiter[level]->arg_d[i][j];
          }
        }
        delimiter[level]->utf_length[i] = k;
      }
      else if (input_record.encoding == UTF_8)
      {
        for (j = 0 ; j < delimiter[level]->arg_length[i] ; j++)
          delimiter[level]->utf_d[i][j] = delimiter[level]->arg_d[i][j];
        delimiter[level]->utf_length[i] = delimiter[level]->arg_length[i];
      }
      else if (input_record.encoding == UTF_16BE)
      {
        k = 0;
        for (j = 0 ; j < delimiter[level]->arg_length[i] ; j++)
        {
          delimiter[level]->utf_d[i][k++] = 0;
          delimiter[level]->utf_d[i][k++] = delimiter[level]->arg_d[i][j];
        }
        delimiter[level]->utf_length[i] = k;
      }
      else if (input_record.encoding == UTF_16LE)
      {
        k = 0;
        for (j = 0 ; j < delimiter[level]->arg_length[i] ; j++)
        {
          delimiter[level]->utf_d[i][k++] = delimiter[level]->arg_d[i][j];
          delimiter[level]->utf_d[i][k++] = 0;
        }
        delimiter[level]->utf_length[i] = k;
      }
      else if (input_record.encoding == UTF_32BE)
      {
        k = 0;
        for (j = 0 ; j < delimiter[level]->arg_length[i] ; j++)
        {
          delimiter[level]->utf_d[i][k++] = 0;
          delimiter[level]->utf_d[i][k++] = 0;
          delimiter[level]->utf_d[i][k++] = 0;
          delimiter[level]->utf_d[i][k++] = delimiter[level]->arg_d[i][j];
        }
        delimiter[level]->utf_length[i] = k;
      }
      else if (input_record.encoding == UTF_32LE)
      {
        k = 0;
        for (j = 0 ; j < delimiter[level]->arg_length[i] ; j++)
        {
          delimiter[level]->utf_d[i][k++] = delimiter[level]->arg_d[i][j];
          delimiter[level]->utf_d[i][k++] = 0;
          delimiter[level]->utf_d[i][k++] = 0;
          delimiter[level]->utf_d[i][k++] = 0;
        }
        delimiter[level]->utf_length[i] = k;
      }
    }
  }
}
