/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2013, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert to XML on the fly.                                     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "validate_args.h"

/* function templates */
void allocate_insert_attr(int level, unsigned char *name, unsigned char *value);
int  allocate_insert_tag(int type, int level, int flag_reinsert, unsigned char *arg1, unsigned char *arg2);
int  load_insert_tag_file(int tag_index, unsigned char *tag_filename);
void reprocess_insert_tag_file();


void allocate_insert_attr(level, name, value)
int level;
unsigned char *name;
unsigned char *value;
{
  /* Allocate storage for custom attribute names */

  if (ninsert_attr)
    insert_attr = (struct_insert_attr *) realloc(insert_attr, (ninsert_attr + 1) * sizeof(struct_insert_attr));
  else
    insert_attr = (struct_insert_attr *) malloc((ninsert_attr + 1) * sizeof(struct_insert_attr));
  if (insert_attr == NULL) out_of_mem();

  insert_attr[ninsert_attr].level        = level;
  insert_attr[ninsert_attr].name         = name;
  insert_attr[ninsert_attr].name_length  = xstrlen(insert_attr[ninsert_attr].name);
  insert_attr[ninsert_attr].value        = value;
  insert_attr[ninsert_attr].value_length = xstrlen(insert_attr[ninsert_attr].value);
  insert_attr[ninsert_attr].nscope       = nscope;
  insert_attr[ninsert_attr].scope        = scope;
  insert_attr[ninsert_attr].scope_seq    = 0;

  ninsert_attr++;
}


int allocate_insert_tag(type, level, flag_reinsert, arg1, arg2)
int type;
int level;
int flag_reinsert;
unsigned char *arg1;
unsigned char *arg2;
{
  /* Allocate storage for custom element tags */
  int j, n;
  int ret;

  n   = 0;
  ret = 0;

  if (type == TAG_INSERT_XMLFILE)
  {
    if (xstrcmp(arg1, (unsigned char *) "raw") == 0)
      n = -1;
    else
    {
      n = atoi((char *) arg1);
      /* atoi() is too smart */
      for (j = 0 ; j < xstrlen(arg1) ; j++)
        if (!isdigit(arg1[j]))
          ret = -2; /* invalid level */
    }
  }

  if (ret == 0)
  {
    if (ninsert_tag)
      insert_tag = (struct_insert_tag *) realloc(insert_tag, (ninsert_tag + 1) * sizeof(struct_insert_tag));
    else
      insert_tag = (struct_insert_tag *) malloc(sizeof(struct_insert_tag));
    if (insert_tag == NULL) out_of_mem();

    insert_tag[ninsert_tag].type         = type;
    insert_tag[ninsert_tag].level        = level;

    /* set name */
    if (type == TAG_INSERT_ELEMENT || type == TAG_INSERT_FILE)
    {
      insert_tag[ninsert_tag].name         = arg1;
      insert_tag[ninsert_tag].name_length  = xstrlen(insert_tag[ninsert_tag].name);
    }
    else
    {
      insert_tag[ninsert_tag].name         = NULL;
      insert_tag[ninsert_tag].name_length  = 0;
    }

    /* set value */
    if (type == TAG_INSERT_ELEMENT)
    {
      insert_tag[ninsert_tag].value        = arg2;
      insert_tag[ninsert_tag].value_length = xstrlen(insert_tag[ninsert_tag].value);
      insert_tag[ninsert_tag].filename     = NULL;
    }
    else if (!load_insert_tag_file(ninsert_tag, arg2))
    {
      ret = -1; /* -1 cannot open file */
    }

    insert_tag[ninsert_tag].indent       = n;
    insert_tag[ninsert_tag].reinsert     = flag_reinsert;
    insert_tag[ninsert_tag].nscope       = nscope;
    insert_tag[ninsert_tag].scope        = scope;
    insert_tag[ninsert_tag].scope_seq    = 0;

    ninsert_tag++;
  }

  return(ret);
}


int load_insert_tag_file(tag_index, tag_filename)
int tag_index;
unsigned char *tag_filename;
{
  /* Allocate storage for contents of specified file */
  int start_char, ret;
  long i;
  int tag_fd;

  /* initialise */
  start_char          = 0;
  input_record.length = 0;
  insert_tag[tag_index].filename = tag_filename;
  ret                 = TRUE;

  if ((tag_fd = open((char *) tag_filename, OPEN_R)) > 0)
  {
    utf_getchar_init(tag_fd);

    /* load file */
    while (!utf_flag_EOF)
    {
      input_record.line[input_record.length++] = utf_getchar(tag_fd);

      /* mem checks */
      if (input_record.length > memseg_input_record)
        memadd_input_record();
    }
    /* correct for EOF */
    if (input_record.length)
      input_record.length--;

    /* skip any BOM in the file */
     if (input_record.length > 2 && xstrncmp((unsigned char *) UTF8_BOM, input_record.line, 3) == 0)
       start_char = 3;
     else if (input_record.length > 1 && xstrncmp((unsigned char *) UTF16BE_BOM, input_record.line, 2) == 0)
       start_char = 2;
     else if (input_record.length > 1 && xstrncmp((unsigned char *) UTF16LE_BOM, input_record.line, 2) == 0)
     {
       /* check for UTF-32LE */
       if (input_record.length > 3 && input_record.line[2] == 0x00 && input_record.line[3] == 0x00)
         start_char = 4;
       else
         start_char = 2;
     }
     else if (input_record.length > 3 && input_record.line[0] == 0x00 && input_record.line[1] == 0x00 
                                      && input_record.line[2] == 0xfe && input_record.line[3] == 0xff)
       start_char = 4;

    /* allocate file storage */
    insert_tag[tag_index].value = (unsigned char *) malloc((input_record.length - start_char + 1) * sizeof(unsigned char));
    if (insert_tag[tag_index].value == NULL) out_of_mem();

    /* copy loaded file to file storage location */
    insert_tag[tag_index].value_length = 0;
    for(i = start_char ; i < input_record.length ; i++)
      insert_tag[tag_index].value[insert_tag[tag_index].value_length++] = input_record.line[i];

    /* NOTE: we cannot further process an inserted file until we know what encoding will eventually be used.
     * This means we have to reprocess the file storage contents when the output encoding
     * is known - which may occur after reading the first input record (i.e. BOM). */

    (void) close(tag_fd);
  }
  else
    ret = FALSE;

  return(ret);
}


void reprocess_insert_tag_file()
{
  /* Reprocess the insert_tag xmlfile file contents once the encoding has been finalised.
   * I.e. Convert line endings to those being used and/or remove prologue and add indentation. */
  int i, k;
  long j, indent_lines, newline_adjust;
  unsigned char *new_value;
  long new_value_length;
  int flag_xmlnewline, flag_newline_unix;

  /* for expediency use a flag for line termintation handling */
  if (xstrcmp(newline, (unsigned char *) NEWLINE_UNIX) == 0)
    flag_newline_unix = TRUE;
  else
    flag_newline_unix = FALSE;

  for (i = 0 ; i < ninsert_tag ; i++)
  {
    if ((insert_tag[i].type == TAG_INSERT_XMLFILE || insert_tag[i].type == TAG_INSERT_FILE) 
       && insert_tag[i].value_length > 0 && insert_tag[i].indent >= 0)
    {
      /* initialise */
      indent_lines        = 1; /* begin count from line 1 */
      newline_adjust      = 0;

      /* determine the number of lines needing attention */
      flag_xmlnewline = TRUE;
      for(j = 0 ; j < insert_tag[i].value_length ; j += output_record.charwidth)
      {
        if (utf_char_equals_2(j, '\n', insert_tag[i].value, insert_tag[i].value_length, output_record.encoding))
        {
          flag_xmlnewline = TRUE;
          /* check for newline adjustments */
          if (j > output_record.charwidth && utf_char_equals_2(j - output_record.charwidth, '\r', insert_tag[i].value, insert_tag[i].value_length, output_record.encoding))
          {
            if (flag_newline_unix)
              newline_adjust--; /* remove DOS line endings */
          }
          else
          {
            if (!flag_newline_unix)
              newline_adjust++; /* add DOS line endings */
          }
        } 
        else if (insert_tag[i].type == TAG_INSERT_XMLFILE && insert_tag[i].indent > 0 
                && flag_xmlnewline && utf_char_equals_2(j, '<', insert_tag[i].value, insert_tag[i].value_length, output_record.encoding))
        {
          indent_lines++;
          flag_xmlnewline = FALSE;
        }
      }

      /* filter out any prologue for xml files */
      if (insert_tag[i].type == TAG_INSERT_XMLFILE)
      {
        new_value_length = 0;
        for (j = 0 ; j < insert_tag[i].value_length ; j += output_record.charwidth)
        {
          if (utf_xstrncmp(&insert_tag[i].value[j], insert_tag[i].value_length, output_record.encoding, (unsigned char *) "<?xml", 5) == 0
             || utf_xstrncmp(&insert_tag[i].value[j], insert_tag[i].value_length, output_record.encoding, (unsigned char *) "<!DOCTYPE", 9) == 0)
          {
            /* read until close bracket is reached */
            while (j < insert_tag[i].value_length && !utf_char_equals_2(j, '>', insert_tag[i].value, insert_tag[i].value_length, output_record.encoding))
              j += output_record.charwidth;
            /* read until the newline of this prologue is reached */
            while (j < insert_tag[i].value_length && !utf_char_equals_2(j, '\n', insert_tag[i].value, insert_tag[i].value_length, output_record.encoding))
              j += output_record.charwidth;
            /* skip this newline */
            if (j < insert_tag[i].value_length && utf_char_equals_2(j, '\n', insert_tag[i].value, insert_tag[i].value_length, output_record.encoding))
              continue; /* for/j */
            /* check array bounds */
            if (j >= insert_tag[i].value_length)
              break; /* for/j */
          }
          /* left shift array characters over the old prologue characters */
          if (j > new_value_length)
            for (k = 0 ; k < output_record.charwidth ; k++)
              insert_tag[i].value[new_value_length + k] = insert_tag[i].value[j + k];

          new_value_length += output_record.charwidth;
        } /* for/j */
        /* update the length of the revised array */
        insert_tag[i].value_length = new_value_length;
      }

      /* Check if further processing is required */
      if (newline_adjust != 0 || (insert_tag[i].type == TAG_INSERT_XMLFILE && insert_tag[i].indent > 0))
      {
        /* allocate enough mem */
        new_value_length = insert_tag[i].value_length + ((indent_lines * insert_tag[i].indent) + newline_adjust) * output_record.charwidth;
        new_value = (unsigned char *) malloc(new_value_length * sizeof(unsigned char));
        if (new_value == NULL) out_of_mem();

        /* copy xml file array to new xml file array with adjustments */
        new_value_length = 0;
        flag_xmlnewline = TRUE;
        for (j = 0 ; j < insert_tag[i].value_length ; j += output_record.charwidth)
        {
          if (utf_char_equals_2(j, '\n', insert_tag[i].value, insert_tag[i].value_length, output_record.encoding))
          {
            flag_xmlnewline = TRUE;
            /* check for newline adjustments */
            if (j > output_record.charwidth && utf_char_equals_2(j - output_record.charwidth, '\r', insert_tag[i].value, insert_tag[i].value_length, output_record.encoding))
            {
              if (flag_newline_unix)
              {
                /* remove DOS line endings */
                if (new_value_length > output_record.charwidth)
                  new_value_length -= output_record.charwidth;
              }
            }
            else
            {
              if (!flag_newline_unix)
              {
                /* add DOS line endings */
                if (output_record.encoding == UTF_8)
                {
                  new_value[new_value_length++] = '\r';
                }
                else if (output_record.encoding == UTF_16BE)
                {
                  new_value[new_value_length++] = 0;
                  new_value[new_value_length++] = '\r';
                }
                else if (output_record.encoding == UTF_16LE)
                {
                  new_value[new_value_length++] = '\r';
                  new_value[new_value_length++] = 0;
                }
                else if (output_record.encoding == UTF_32BE)
                {
                  new_value[new_value_length++] = 0;
                  new_value[new_value_length++] = 0;
                  new_value[new_value_length++] = 0;
                  new_value[new_value_length++] = '\r';
                }
                else if (output_record.encoding == UTF_32LE)
                {
                  new_value[new_value_length++] = '\r';
                  new_value[new_value_length++] = 0;
                  new_value[new_value_length++] = 0;
                  new_value[new_value_length++] = 0;
                }
              }
            }
          }
          else if (insert_tag[i].type == TAG_INSERT_XMLFILE
                  && flag_xmlnewline && utf_char_equals_2(j, '<', insert_tag[i].value, insert_tag[i].value_length, output_record.encoding))
          {
            /* process indentation for XML file */
            for (k = 0 ; k < insert_tag[i].indent ; k++)
            {
              if (output_record.encoding == UTF_8)
                new_value[new_value_length++] = ' ';
              else if (output_record.encoding == UTF_16BE)
              {
                new_value[new_value_length++] = 0;
                new_value[new_value_length++] = ' ';
              }
              else if (output_record.encoding == UTF_16LE)
              {
                new_value[new_value_length++] = ' ';
                new_value[new_value_length++] = 0;
              }
              else if (output_record.encoding == UTF_32BE)
              {
                new_value[new_value_length++] = 0;
                new_value[new_value_length++] = 0;
                new_value[new_value_length++] = 0;
                new_value[new_value_length++] = ' ';
              }
              else if (output_record.encoding == UTF_32LE)
              {
                new_value[new_value_length++] = ' ';
                new_value[new_value_length++] = 0;
                new_value[new_value_length++] = 0;
                new_value[new_value_length++] = 0;
              }
            }
            flag_xmlnewline = FALSE;
          }

          /* add char to new file mem array */
          for (k = 0 ; k < output_record.charwidth ; k++)
            new_value[new_value_length++] = insert_tag[i].value[j + k];
        } /* end for/j */

        /* repoint to new file contents */
        (void) free(insert_tag[i].value);
        insert_tag[i].value        = new_value;
        insert_tag[i].value_length = new_value_length;
      }
    }
  } /* end for/i */
}
