/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2013, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert to XML on the fly.                                     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "validate_args.h"
#ifdef LANG_EN
#include "messages_EN.h"
#endif

/* function templates */
void populate_delimiter(int level, unsigned char *d);
void increase_nlevels(int newlevel);
void version();
void usage(int invocation);

void validate_args(argc, argv, invocation)
int argc;
unsigned char **argv;
int invocation;
{
  /* validate all the parameters supplied to this command */
  int i, j, level, i_prev, flag_tmp;

  /* only initialise once */
  if (invocation == ARGS_CMD)
  {
    /* initialise some infamous delimiter */
    delimiter_newline[0] = '\n'; delimiter_newline[1] = '\0';
    delimiter_cr[0]      = '\r'; delimiter_cr[1]      = '\0';
    delimiter_tab[0]     = '\t'; delimiter_tab[1]     = '\0';
    delimiter_space[0]   = ' ' ; delimiter_space[1]   = '\0';

    /* initialise default print switches */
    sw_print_default.tagname        = default_schema_field_name;
    sw_print_default.numbered_attr  = FALSE;
    sw_print_default.numbered_tag   = FALSE;
    sw_print_default.delimiter_attr = FALSE;
    sw_print_default.level_attr     = FALSE;
    sw_print_default.level_tag      = FALSE;
    sw_print_default.timestamp_attr = FALSE;

    /* initialise level delimiters and level switches */
    for (level = 0 ; level < nlevels ; level++)
    {
      delimiter[level]->ndelimiters = 0;

      if (level == 0)
      {
        sw_print[0]->tagname        = default_schema_parent_name;
        sw_print[0]->tagname_length = xstrlen(sw_print[0]->tagname);
        kvp_tag[0]->enabled         = FALSE;
        kvp_tag[0]->set             = FALSE;
        kvp_tag[0]->name            = sw_print[0]->tagname;
        kvp_tag[0]->length          = sw_print[0]->tagname_length;
        kvp_tag[0]->encoding        = UTF_8;
        kvp_tag[0]->level_d         = 0;
        kvp_tag[0]->match_d         = DELIMITER_INDEX_UNMATCHED;
      }
      else if (level == 1)
      {
        delimiter[1]->match_d       = DELIMITER_INDEX_NEWLINE;
        sw_print[1]->tagname        = default_schema_line_name;
        sw_print[1]->tagname_length = xstrlen(sw_print[1]->tagname);
        kvp_tag[1]->enabled         = FALSE;
        kvp_tag[1]->set             = FALSE;
        kvp_tag[1]->name            = sw_print[1]->tagname;
        kvp_tag[1]->length          = sw_print[1]->tagname_length;
        kvp_tag[1]->encoding        = UTF_8;
        kvp_tag[1]->level_d         = 0;
        kvp_tag[1]->match_d         = DELIMITER_INDEX_UNMATCHED;
      }
      else 
      {
        sw_print[level]->tagname        = sw_print_default.tagname;
        sw_print[level]->tagname_length = xstrlen(sw_print[level]->tagname);
        kvp_tag[level]->enabled         = FALSE;
        kvp_tag[level]->set             = FALSE;
        kvp_tag[level]->name            = sw_print[level]->tagname;
        kvp_tag[level]->length          = sw_print[level]->tagname_length;
        kvp_tag[level]->encoding        = UTF_8;
        kvp_tag[level]->level_d         = 0;
        kvp_tag[level]->match_d         = DELIMITER_INDEX_UNMATCHED;
      }
      sw_print[level]->numbered_attr  = sw_print_default.numbered_attr;
      sw_print[level]->numbered_tag   = sw_print_default.numbered_tag;
      sw_print[level]->delimiter_attr = sw_print_default.delimiter_attr;
      sw_print[level]->level_attr     = sw_print_default.level_attr;
      sw_print[level]->level_tag      = sw_print_default.level_tag;
      sw_print[level]->timestamp_attr = sw_print_default.timestamp_attr;

      output_record.field_cntr[level] = 0;
    }

    /* initialise input BOM value */
    xstrcpy(input_record.bom_name, (unsigned char *) BOM_IGNORE_NAME);

    /* initialise default utf format (UTF-8) */
    sw_encoding = UTF_8;
    xstrcpy(output_record.encoding_name, (unsigned char *) UTF_8_NAME);
    utf_set_encoding();

    /* initialise schema options */
    schema_argc = 0;

    /* initialise expel options */
    nexpel = 0;

    /* initialise column options */
    ncolumn = 0;

    /* initialise custom attribute/tag options */
    ninsert_attr = 0;
    ninsert_tag  = 0;

    /* initialise quotechar options */
    quotechar_type    = 0;
    quotechars_length = 0;
    i_quotechar       = 0;
  }


  /* basic parameter handling  - deliberately avoiding getopt dependancy */
  flag_tmp = FALSE;
  i = 1;
  while (i < argc)
  {
    /* control */
    i_prev = i;


    /*
     * Option: Help
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--help") == 0 || strcmp((char *) argv[i], "-h") == 0))
    {
      if (invocation != ARGS_SCHEMA)
      {
        usage(invocation);
        i += 1;
      }
      else
      {
        fprintf(stderr, MSG_SCHEMA_INVALID_OPTION, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Version
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--version") == 0 || strcmp((char *) argv[i], "-v") == 0))
    {
      if (invocation != ARGS_SCHEMA)
      {
        version();
        i += 1;
      }
      else
      {
        fprintf(stderr, MSG_SCHEMA_INVALID_OPTION, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: License
     */
    if (i < argc
        && strcmp((char *) argv[i], "--license") == 0)
    {
      if (invocation != ARGS_SCHEMA)
      {
        show_license();
        i += 1;
      }
      else
      {
        fprintf(stderr, MSG_SCHEMA_INVALID_OPTION, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Debug
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--debug") == 0 || strcmp((char *) argv[i], "-debug") == 0))
    {
      if (invocation != ARGS_SCHEMA)
      {
        sw_debug = TRUE;
        i += 1;
      }
      else
      {
        fprintf(stderr, MSG_SCHEMA_INVALID_OPTION, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Linetags
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--linetags") == 0 || strcmp((char *) argv[i], "-L") == 0))
    {
      sw_line_tags = TRUE;
      i += 1;
    }


    /*
     * Option: Linenumbers
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--linenumbers") == 0 || strcmp((char *) argv[i], "-l") == 0))
    {
      sw_print[1]->numbered_tag = TRUE;
      i += 1;
    }


    /*
     * Option: Fieldnumbers
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--fieldnumbers") == 0 || strcmp((char *) argv[i], "-f") == 0))
    {
      sw_print[2]->numbered_tag = TRUE;
      i += 1;
    }


    /*
     * Option: Blanklines
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--blanklines") == 0 || strcmp((char *) argv[i], "-b") == 0))
    {
      sw_blank_lines = TRUE;
      i += 1;
    }


    /*
     * Option: Quoted fields
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--quotedfields") == 0 || strcmp((char *) argv[i], "-q") == 0))
    {
      sw_field_quoting = TRUE;
      i += 1;
    }


    /*
     * Option: Quoted fields spanning multiple input records
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--quotedfields2") == 0 || strcmp((char *) argv[i], "-q2") == 0))
    {
      sw_field_quoting_span  = TRUE;
      sw_field_quoting       = TRUE;
      i += 1;
    }


    /*
     * Option: Trim fields
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--trim") == 0 || strcmp((char *) argv[i], "-t") == 0))
    {
      sw_field_trim = TRUE;
      i += 1;
    }


    /*
     * Option: Expel empty fields
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--expelempty") == 0 || strcmp((char *) argv[i], "-e") == 0))
    {
      sw_field_expel_empty = TRUE;
      i += 1;
    }


    /*
     * Option: Print summary
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--summary") == 0 || strcmp((char *) argv[i], "-s") == 0))
    {
      summary_fd = STDOUT_FD;
      sw_summary = SUMMARY_NORMAL;
      i += 1;
    }
    /* Option: Print summary2 - i.e. to stderr */
    if (i < argc
        && (strcmp((char *) argv[i], "--summary2") == 0 || strcmp((char *) argv[i], "-s2") == 0))
    {
      summary_fd = STDERR_FD;
      if (!sw_summary)
        sw_summary = SUMMARY_NORMAL;
      i += 1;
    }
    /* Option: Print summaryc - i.e. as comment */
    if (i < argc
        && (strcmp((char *) argv[i], "--summaryc") == 0 || strcmp((char *) argv[i], "-sc") == 0))
    {
      sw_summary = SUMMARY_COMMENT;
      i += 1;
    }
    /* Option: Print summaryn - i.e. no digests */
    if (i < argc
        && (strcmp((char *) argv[i], "--summaryn") == 0 || strcmp((char *) argv[i], "-sn") == 0))
    {
      if (!sw_summary)
        sw_summary = SUMMARY_NORMAL;
      sw_digest = FALSE;
      i += 1;
    }
    /* Option: Print summaryf  - i.e. to file */
    if (i < argc
        && (strcmp((char *) argv[i], "--summaryf") == 0 || strcmp((char *) argv[i], "-sf") == 0))
    {
      if (i < argc - 1)
      {
        sw_summary = SUMMARY_FILE;
        if ((summary_fd = open((char *) argv[i + 1], OPEN_W, S_IREAD|S_IWRITE)) > 0)
          i += 2;
        else
        {
          fprintf(stderr, MSG_CANNOT_WRITE_FILE, progname, argv[i], argv[i + 1], newline);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, MSG_PARAMETER_MISSING, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: no xml
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--noxml") == 0))
    {
      if (invocation != ARGS_SCHEMA)
      {
        sw_xml_format = SW_NOXML;
        i += 1;
      }
      else
      {
        fprintf(stderr, MSG_SCHEMA_INVALID_OPTION, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Undo xml
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--unxml") == 0 || strcmp((char *) argv[i], "-U") == 0))
    {
      if (invocation != ARGS_SCHEMA)
      {
        sw_xml_format = SW_UNXML;
        i += 1;
      }
      else
      {
        fprintf(stderr, MSG_SCHEMA_INVALID_OPTION, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: DTD schema file
     */
    if (i < argc
        && ((strcmp((char *) argv[i], "--schemadtd") == 0 || strcmp((char *) argv[i], "-Sd") == 0)
         || (strcmp((char *) argv[i], "--dtd") == 0 || strcmp((char *) argv[i], "-d") == 0)))
    {
      if (invocation != ARGS_SCHEMA)
      {
        if (sw_schema == SCHEMA_DEFAULT)
        {
          if (sw_key_value_pairs == FALSE)
          {
            if (i < argc - 1)
            {
              schema_filename = argv[i + 1];
              sw_schema = SCHEMA_DTD;
              i += 2;
            }
            else
            {
              fprintf(stderr, MSG_SCHEMA_DTD_NO_FILE, progname, argv[i], newline);
              err_status = -1;
              usage(invocation);
            }
          }
          else
          {
            fprintf(stderr, MSG_SCHEMA_KVP_INVALID_OPTION, progname, argv[i], newline);
            err_status = -1;
            usage(invocation);
          }
        }
        else
        {
          fprintf(stderr, MSG_SCHEMA_MULTIPLE_FILES, progname, argv[i], newline);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, MSG_SCHEMA_INVALID_OPTION, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: RNC schema file
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--schemarnc") == 0 || strcmp((char *) argv[i], "-Sr") == 0))
    {
      if (invocation != ARGS_SCHEMA)
      {
        if (sw_schema == SCHEMA_DEFAULT)
        {
          if (sw_key_value_pairs == FALSE)
          {
            if (i < argc - 1)
            {
              schema_filename = argv[i + 1];
              sw_schema = SCHEMA_RNC;
              i += 2;
            }
            else
            {
              fprintf(stderr, MSG_SCHEMA_RNC_NO_FILE, progname, argv[i], newline);
              err_status = -1;
              usage(invocation);
            }
          }
          else
          {
            fprintf(stderr, MSG_SCHEMA_KVP_INVALID_OPTION, progname, argv[i], newline);
            err_status = -1;
            usage(invocation);
          }
        }
        else
        {
          fprintf(stderr, MSG_SCHEMA_MULTIPLE_FILES, progname, argv[i], newline);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, MSG_SCHEMA_INVALID_OPTION, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: XSD schema file
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--schemaxsd") == 0 || strcmp((char *) argv[i], "-Sx") == 0))
    {
      if (invocation != ARGS_SCHEMA)
      {
        if (sw_schema == SCHEMA_DEFAULT)
        {
          if (sw_key_value_pairs == FALSE)
          {
            if (i < argc - 1)
            {
              schema_filename = argv[i + 1];
              sw_schema = SCHEMA_XSD;
              i += 2;
            }
            else
            {
              fprintf(stderr, MSG_SCHEMA_XSD_NO_FILE, progname, argv[i], newline);
              err_status = -1;
              usage(invocation);
            }
          }
          else
          {
            fprintf(stderr, MSG_SCHEMA_KVP_INVALID_OPTION, progname, argv[i], newline);
            err_status = -1;
            usage(invocation);
          }
        }
        else
        {
          fprintf(stderr, MSG_SCHEMA_MULTIPLE_FILES, progname, argv[i], newline);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, MSG_SCHEMA_INVALID_OPTION, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }
    

    /*
     * Option: DTD or RNC or XSD schema file
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--schema") == 0 || strcmp((char *) argv[i], "-S") == 0))
    {
      if (invocation != ARGS_SCHEMA)
      {
        if (sw_schema == SCHEMA_DEFAULT)
        {
          if (sw_key_value_pairs == FALSE)
          {
            if (i < argc - 1)
            {
              j = xstrlen(argv[i + 1]);
              if (j > 4 && argv[i + 1][j - 4] == '.' && tolower(argv[i + 1][j - 3]) == 'd' && tolower(argv[i + 1][j - 2]) == 't' && tolower(argv[i + 1][j - 1]) == 'd')
                sw_schema = SCHEMA_DTD;
              else if (j > 4 && argv[i + 1][j - 4] == '.' && tolower(argv[i + 1][j - 3]) == 'r' && tolower(argv[i + 1][j - 2]) == 'n' && tolower(argv[i + 1][j - 1]) == 'c')
                sw_schema = SCHEMA_RNC;
              else if (j > 4 && argv[i + 1][j - 4] == '.' && tolower(argv[i + 1][j - 3]) == 'x' && tolower(argv[i + 1][j - 2]) == 's' && tolower(argv[i + 1][j - 1]) == 'd')
                sw_schema = SCHEMA_XSD;
              else
              {
                fprintf(stderr, MSG_SCHEMA_UNSUPPORTED_EXTENSION, progname, argv[i], argv[i + 1], newline);
                err_status = -1;
                usage(invocation);
              }
              schema_filename = argv[i + 1];
              i += 2;
            }
            else
            {
              fprintf(stderr, MSG_SCHEMA_NO_FILE, progname, argv[i], newline);
              err_status = -1;
              usage(invocation);
            }
          }
          else
          {
            fprintf(stderr, MSG_SCHEMA_KVP_INVALID_OPTION, progname, argv[i], newline);
            err_status = -1;
            usage(invocation);
          }
        }
        else
        {
          fprintf(stderr, MSG_SCHEMA_MULTIPLE_FILES, progname, argv[i], newline);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, MSG_SCHEMA_INVALID_OPTION, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Make parent element of schema 
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--makeparent") == 0 || strcmp((char *) argv[i], "-M") == 0))
    {
      if (i < argc - 1)
      {
        schema_match_start_level = 0;
        schema_match_start_name  = argv[i + 1];
        i += 2;
      }
      else
      {
        fprintf(stderr, MSG_PARAMETER_MISSING, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Record separator
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--recordseparator") == 0 || strcmp((char *) argv[i], "-R") == 0))
    {
      if (i < argc - 1)
      {
        nscope = 0;
        scope  = NULL;
        populate_delimiter(1, argv[i + 1]);
        i += 2;
      }
      else
      {
        fprintf(stderr, MSG_DELIMITER_NO_DELIMITER, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Field separator
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--fieldseparator") == 0 || strcmp((char *) argv[i], "-F") == 0))
    {
      if (ncolumn == 0)
      {
        if (i < argc - 1)
        {
          nscope = 0;
          scope  = NULL;
          populate_delimiter(2, argv[i + 1]);
          /* allow zero length fields */
          sw_field_empty_valid = TRUE;
          i += 2;
        }
        else
        {
          fprintf(stderr, MSG_DELIMITER_NO_DELIMITER, progname, argv[i], newline);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, MSG_DELIMITER_CONFLICT_WITH_COLUMN_REGEX_OPTION, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Delimiter
     */
    if (i < argc
        && ((strncmp((char *) argv[i], "--fieldseparator", 16) == 0 && xstrlen(argv[i]) > 16)
        || (strncmp((char *) argv[i], "-F", 2) == 0 && xstrlen(argv[i]) > 2)))
    {
      if (i < argc - 1)
      {
        level = process_arg_scoped(OPT_DELIMITER, argv[i]);
        if (ncolumn == 0 || (ncolumn > 0 && level < 2))
        {
          if (level >= nlevels)
            increase_nlevels(level + 1);

          if (level > 0 && level < nlevels)
          {
            populate_delimiter(level, argv[i + 1]);
            if (level == 2) /* allow zero length fields */
              sw_field_empty_valid = TRUE;
            i += 2;
          }
          else 
          {
            if (level == -15)
              fprintf(stderr, MSG_SCOPE_INVALID, progname, argv[i], newline);
            else if (level == -13)
              fprintf(stderr, MSG_SCOPE_MISSING, progname, argv[i], newline);
            else if (level == -12)
              fprintf(stderr, MSG_SCOPE_NO_LEVEL, progname, argv[i], newline);
            else
              fprintf(stderr, MSG_INVALID_LEVEL, progname, argv[i], newline);
            err_status = -1;
            usage(invocation);
          }
        }
        else 
        {
          fprintf(stderr, MSG_DELIMITER_CONFLICT_WITH_COLUMN_REGEX_OPTION, progname, argv[i], newline);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, MSG_DELIMITER_NO_DELIMITER, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Column mapping
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--column") == 0 || strcmp((char *) argv[i], "-C") == 0
        || strncmp((char *) argv[i], "--column:", 9) == 0 || strncmp((char *) argv[i], "-C:", 3) == 0))
    {
      if (nlevels <= DEFAULT_NLEVELS && delimiter[2]->ndelimiters == 0)
      {
        if (i < argc - 1)
        {
          j = process_arg_column(argv[i], argv[i + 1]);
          if (j > 0)
            i += 2;
          else
          {
            if (j == -15)
              fprintf(stderr, MSG_SCOPE_INVALID, progname, argv[i], newline);
            else
              fprintf(stderr, MSG_INVALID_RANGE, progname, argv[i], argv[i + 1], newline);
            err_status = -1;
            usage(invocation);
          }
        }
        else
        {
          fprintf(stderr, MSG_NO_RANGE, progname, argv[i], newline);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, MSG_LEVEL_ONE_ONLY_OPTION, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Regex mapping
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--regex") == 0 || strcmp((char *) argv[i], "-W") == 0
        || strncmp((char *) argv[i], "--regex:", 8) == 0 || strncmp((char *) argv[i], "-W:", 3) == 0))
    {
      if (nlevels <= DEFAULT_NLEVELS && delimiter[2]->ndelimiters == 0)
      {
        if (i < argc - 1)
        {
          j = process_arg_regex(argv[i], argv[i + 1]);
          if (j > 0)
            i += 2;
          else
          {
            if (j == -15)
              fprintf(stderr, MSG_SCOPE_INVALID, progname, argv[i], newline);
            else if (j == -22)
              fprintf(stderr, MSG_REGEX_NO_PATTERN_DELIMITERS, progname, argv[i], argv[i + 1], newline);
            else if (j == -23)
              fprintf(stderr, MSG_REGEX_NO_PATTERN, progname, argv[i], argv[i + 1], newline);
            else if (j == -24)
              fprintf(stderr, MSG_REGEX_INVALID_CFLAGS, progname, argv[i], argv[i + 1], newline);
            else if (j == -25)
              fprintf(stderr, MSG_REGEX_PATTERN_INVALID, progname, argv[i], argv[i + 1], newline);
            else
              fprintf(stderr, MSG_REGEX_INVALID, progname, argv[i], argv[i + 1], newline);
            err_status = -1;
            usage(invocation);
          }
        }
        else
        {
          fprintf(stderr, MSG_PARAMETER_MISSING, progname, argv[i], newline);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, MSG_LEVEL_ONE_ONLY_OPTION, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Tag name handling
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--tag") == 0 || strcmp((char *) argv[i], "-T") == 0))
    {
      if (i < argc - 1)
      {
        nscope = 0;
        scope  = NULL;
        if (strcmp((char *) argv[i + 1], "number") == 0)
        {
          sw_print_default.numbered_tag = TRUE;
          for (level = 1 ; level < nlevels ; level++)
            sw_print[level]->numbered_tag = sw_print_default.numbered_tag;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "level") == 0)
        {
          sw_print_default.level_tag = TRUE;
          for (level = 1 ; level < nlevels ; level++)
            sw_print[level]->level_tag = sw_print_default.level_tag;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "name") == 0 || strcmp((char *) argv[i + 1], "tagname") == 0)
        {
          if (i < argc - 2)
          {
            sw_print_default.tagname = argv[i + 2];
            for (level = 1 ; level < nlevels ; level++)
            {
              sw_print[level]->tagname        = sw_print_default.tagname;
              sw_print[level]->tagname_length = xstrlen(sw_print[level]->tagname);
            }
            i += 3;
          }
          else
          {
            fprintf(stderr, MSG_PARAMETER_SECOND_MISSING, progname, argv[i], argv[i + 1], newline);
            err_status = -1;
            usage(invocation);
          }
        }
        else if (strcmp((char *) argv[i + 1], "insert") == 0 || strcmp((char *) argv[i + 1], "reinsert") == 0)
        {
          if (i < argc - 3)
          {
            if (strcmp((char *) argv[i + 1], "reinsert") == 0)
              flag_tmp = TRUE;
            else
              flag_tmp = FALSE;
            (void) allocate_insert_tag(TAG_INSERT_ELEMENT, -1, flag_tmp, argv[i + 2], argv[i + 3]); /* -1 applies to all levels except 0 and 1 */
            i += 4;
          }
          else
          {
            fprintf(stderr, MSG_PARAMETER_MISSING_TWO_REQUIRED, progname, argv[i], argv[i + 1], newline);
            err_status = -1;
            usage(invocation);
          }
        }
        else if (strcmp((char *) argv[i + 1], "insertfile") == 0 || strcmp((char *) argv[i + 1], "reinsertfile") == 0)
        {
          if (i < argc - 3)
          {
            if (strcmp((char *) argv[i + 1], "reinsertfile") == 0)
              flag_tmp = TRUE;
            else
              flag_tmp = FALSE;
            j = allocate_insert_tag(TAG_INSERT_FILE, -1, flag_tmp, argv[i + 2], argv[i + 3]); /* -1 applies to all levels except 0 and 1 */
            if (j == 0)
            {
              i += 4;
            }
            else
            {
              fprintf(stderr, MSG_CANNOT_OPEN_FILE, progname, argv[i], argv[i + 1], argv[i + 2], argv[i + 3], newline);
              err_status = -1;
              usage(invocation);
            }
          }
          else
          {
            fprintf(stderr, MSG_PARAMETER_MISSING_TWO_REQUIRED, progname, argv[i], argv[i + 1], newline);
            err_status = -1;
            usage(invocation);
          }
        }
        else if (strcmp((char *) argv[i + 1], "insertfilexml") == 0 || strcmp((char *) argv[i + 1], "reinsertfilexml") == 0)
        {
          if (i < argc - 3)
          {
            if (strcmp((char *) argv[i + 1], "reinsertfilexml") == 0)
              flag_tmp = TRUE;
            else
              flag_tmp = FALSE;
            j = allocate_insert_tag(TAG_INSERT_XMLFILE, -1, flag_tmp, argv[i + 2], argv[i + 3]); /* -1 applies to all levels except 0 and 1 */
            if (j == 0)
            {
              i += 4;
            }
            else if (j == -2)
            {
              err_status = -1;
              fprintf(stderr, MSG_INVALID_INDENT, progname, argv[i], argv[i + 1], argv[i + 2], argv[i + 3], newline);
              usage(invocation);
            }
            else
            {
              fprintf(stderr, MSG_CANNOT_OPEN_FILE, progname, argv[i], argv[i + 1], argv[i + 2], argv[i + 3], newline);
              err_status = -1;
              usage(invocation);
            }
          }
          else
          {
            fprintf(stderr, MSG_PARAMETER_MISSING_TWO_REQUIRED, progname, argv[i], argv[i + 1], newline);
            err_status = -1;
            usage(invocation);
          }
        }
        else
        {
          fprintf(stderr, MSG_PARAMETER_INVALID, progname, argv[i], argv[i + 1], newline);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, MSG_PARAMETER_MISSING, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }   
   
    if (i < argc
        && ((strncmp((char *) argv[i], "--tag", 5) == 0 && xstrlen(argv[i]) > 5)
        || (strncmp((char *) argv[i], "-T", 2) == 0 && xstrlen(argv[i]) > 2)))
    {
      if (i < argc - 1)
      {
        level = process_arg_scoped(OPT_TAG, argv[i]);

        if (level >= nlevels)
          increase_nlevels(level + 1);

        if (level >= 0 && level < nlevels)
        {
          if (strcmp((char *) argv[i + 1], "number") == 0)
          {
            if (!nscope)
            {
              if (level > 0)
              {
                sw_print[level]->numbered_tag = TRUE;
                i += 2;
              }
              else
              {
                fprintf(stderr, MSG_INVALID_LEVEL_ZERO, progname, argv[i], argv[i + 1], newline);
                err_status = -1;
                usage(invocation);
              }
            }
            else
            {
              fprintf(stderr, MSG_SCOPE_UNSUPPORTED, progname, argv[i], argv[i + 1], newline);
              err_status = -1;
              usage(invocation);
            }
          }
          else if (strcmp((char *) argv[i + 1], "level") == 0 && nscope == 0)
          {
            if (!nscope)
            {
              if (level > 0)
              {
                sw_print[level]->level_tag = TRUE;
                i += 2;
              }
              else
              {
                fprintf(stderr, MSG_INVALID_LEVEL_ZERO, progname, argv[i], argv[i + 1], newline);
                err_status = -1;
                usage(invocation);
              }
            }
            else
            {
              fprintf(stderr, MSG_SCOPE_UNSUPPORTED, progname, argv[i], argv[i + 1], newline);
              err_status = -1;
              usage(invocation);
            }
          }
          else if ((strcmp((char *) argv[i + 1], "name") == 0 || strcmp((char *) argv[i + 1], "tagname") == 0 )&& nscope == 0)
          {
            if (!nscope)
            {
              if (i < argc - 2)
              {
                sw_print[level]->tagname        = argv[i + 2];
                sw_print[level]->tagname_length = xstrlen(sw_print[level]->tagname);
                i += 3;
              }
              else
              {
                fprintf(stderr, MSG_PARAMETER_SECOND_MISSING, progname, argv[i], argv[i + 1], newline);
                err_status = -1;
                usage(invocation);
              }
            }
            else
            {
              fprintf(stderr, MSG_SCOPE_UNSUPPORTED, progname, argv[i], argv[i + 1], newline);
              err_status = -1;
              usage(invocation);
            }
          }
          else if (strcmp((char *) argv[i + 1], "insert") == 0 || strcmp((char *) argv[i + 1], "reinsert") == 0)
          {
            if (i < argc - 3)
            {
              if (level > 0)
              {
                if (strcmp((char *) argv[i + 1], "reinsert") == 0)
                  flag_tmp = TRUE;
                else
                  flag_tmp = FALSE;
                (void) allocate_insert_tag(TAG_INSERT_ELEMENT, level, flag_tmp, argv[i + 2], argv[i + 3]);
                i += 4;
              }
              else
              {
                fprintf(stderr, MSG_INVALID_LEVEL_ZERO, progname, argv[i], argv[i + 1], newline);
                err_status = -1;
                usage(invocation);
              }
            }
            else
            {
              fprintf(stderr, MSG_PARAMETER_MISSING_TWO_REQUIRED, progname, argv[i], argv[i + 1], newline);
              err_status = -1;
              usage(invocation);
            }
          }
          else if (strcmp((char *) argv[i + 1], "insertfile") == 0 || strcmp((char *) argv[i + 1], "reinsertfile") == 0)
          {
            if (i < argc - 3)
            {
              if (level > 0)
              {
                if (strcmp((char *) argv[i + 1], "reinsertfile") == 0)
                  flag_tmp = TRUE;
                else
                  flag_tmp = FALSE;
                j = allocate_insert_tag(TAG_INSERT_FILE, level, flag_tmp, argv[i + 2], argv[i + 3]); 
                if (j == 0)
                {
                  i += 4;
                }
                else
                {
                  fprintf(stderr, MSG_CANNOT_OPEN_FILE, progname, argv[i], argv[i + 1], argv[i + 2], argv[i + 3], newline);
                  err_status = -1;
                  usage(invocation);
                }
              }
              else
              {
                fprintf(stderr, MSG_INVALID_LEVEL_ZERO, progname, argv[i], argv[i + 1], newline);
                err_status = -1;
                usage(invocation);
              }
            }
            else
            {
              fprintf(stderr, MSG_PARAMETER_MISSING_TWO_REQUIRED, progname, argv[i], argv[i + 1], newline);
              err_status = -1;
              usage(invocation);
            }
          }
          else if (strcmp((char *) argv[i + 1], "insertfilexml") == 0 || strcmp((char *) argv[i + 1], "reinsertfilexml") == 0)
          {
            if (i < argc - 3)
            {
              if (level >= 0)
              {
                if (strcmp((char *) argv[i + 1], "reinsertfilexml") == 0)
                  flag_tmp = TRUE;
                else
                  flag_tmp = FALSE;
                j = allocate_insert_tag(TAG_INSERT_XMLFILE, level, flag_tmp, argv[i + 2], argv[i + 3]); 
                if (j == 0)
                {
                  i += 4;
                }
                else if (j == -2)
                {
                  fprintf(stderr, MSG_INVALID_INDENT, progname, argv[i], argv[i + 1], argv[i + 2], argv[i + 3], newline);
                  err_status = -1;
                  usage(invocation);
                }
                else
                {
                  fprintf(stderr, MSG_CANNOT_OPEN_FILE, progname, argv[i], argv[i + 1], argv[i + 2], argv[i + 3], newline);
                  err_status = -1;
                  usage(invocation);
                }
              }
              else
              {
                fprintf(stderr, MSG_INVALID_LEVEL, progname, argv[i], newline);
                err_status = -1;
                usage(invocation);
              }
            }
            else
            {
              fprintf(stderr, MSG_PARAMETER_MISSING_TWO_REQUIRED, progname, argv[i], argv[i + 1], newline);
              err_status = -1;
              usage(invocation);
            }
          }
          else
          {
            fprintf(stderr, MSG_PARAMETER_INVALID, progname, argv[i], argv[i + 1], newline);
            err_status = -1;
            usage(invocation);
          }
        }
        else 
        {
          if (level == -15)
            fprintf(stderr, MSG_SCOPE_INVALID, progname, argv[i], newline);
          else if (level == -13)
            fprintf(stderr, MSG_SCOPE_MISSING, progname, argv[i], newline);
          else if (level == -12)
            fprintf(stderr, MSG_SCOPE_NO_LEVEL, progname, argv[i], newline);
          else
            fprintf(stderr, MSG_INVALID_LEVEL, progname, argv[i], newline);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, MSG_PARAMETER_MISSING, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }

    
    /*
     * Option: Attribute handling
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--attribute") == 0 || strcmp((char *) argv[i], "-A") == 0))
    {
      if (i < argc - 1)
      {
        nscope = 0;
        scope  = NULL;
        if (strcmp((char *) argv[i + 1], "number") == 0)
        {
          sw_print_default.numbered_attr = TRUE;
          for (level = 1 ; level < nlevels ; level++)
            sw_print[level]->numbered_attr = sw_print_default.numbered_attr;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "delimiter") == 0)
        {
          sw_print_default.delimiter_attr = TRUE;
          for (level = 1 ; level < nlevels ; level++)
            sw_print[level]->delimiter_attr = sw_print_default.delimiter_attr;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "level") == 0)
        {
          sw_print_default.level_attr = TRUE;
          for (level = 1 ; level < nlevels ; level++)
            sw_print[level]->level_attr = sw_print_default.level_attr;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "timestamp") == 0)
        {
          sw_print_default.timestamp_attr = TRUE;
          for (level = 1 ; level < nlevels ; level++)
            sw_print[level]->timestamp_attr = sw_print_default.timestamp_attr;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "insert") == 0 || strcmp((char *) argv[i + 1], "attrname") == 0)
        {
          if (i < argc - 3)
          {
            allocate_insert_attr(-1, argv[i + 2], argv[i + 3]); /* -1 applies to all levels except 0 */
            i += 4;
          }
          else
          {
            fprintf(stderr, MSG_PARAMETER_MISSING_TWO_REQUIRED, progname, argv[i], argv[i + 1], newline);
            err_status = -1;
            usage(invocation);
          }
        }
        else
        {
          fprintf(stderr, MSG_PARAMETER_INVALID, progname, argv[i], argv[i + 1], newline);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, MSG_PARAMETER_MISSING, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }

    if (i < argc
        && ((strncmp((char *) argv[i], "--attribute", 11) == 0 && xstrlen(argv[i]) > 11)
        || (strncmp((char *) argv[i], "-A", 2) == 0 && xstrlen(argv[i]) > 2)))
    {
      if (i < argc - 1)
      {
        level = process_arg_scoped(OPT_ATTRIBUTE, argv[i]);

        if (level >= nlevels)
          increase_nlevels(level + 1);

        if (level >= 0 && level < nlevels)
        {
          if (strcmp((char *) argv[i + 1], "number") == 0 && nscope == 0)
          {
            if (!nscope)
            {
              if (level > 0)
              {
                sw_print[level]->numbered_attr = TRUE;
                i += 2;
              }
              else
              {
                fprintf(stderr, MSG_INVALID_LEVEL_ZERO, progname, argv[i], argv[i + 1], newline);
                err_status = -1;
                usage(invocation);
              }
            }
            else
            {
              fprintf(stderr, MSG_SCOPE_UNSUPPORTED, progname, argv[i], argv[i + 1], newline);
              err_status = -1;
              usage(invocation);
            }
          }
          else if (strcmp((char *) argv[i + 1], "delimiter") == 0 && nscope == 0)
          {
            if (!nscope)
            {
              if (level > 0)
              {
                sw_print[level]->delimiter_attr = TRUE;
                i += 2;
              }
              else
              {
                fprintf(stderr, MSG_INVALID_LEVEL_ZERO, progname, argv[i], argv[i + 1], newline);
                err_status = -1;
                usage(invocation);
              }
            }
            else
            {
              fprintf(stderr, MSG_SCOPE_UNSUPPORTED, progname, argv[i], argv[i + 1], newline);
              err_status = -1;
              usage(invocation);
            }
          }
          else if (strcmp((char *) argv[i + 1], "level") == 0 && nscope == 0)
          {
            if (!nscope)
            {
              if (level > 0)
              {
                sw_print[level]->level_attr = TRUE;
                i += 2;
              }
              else
              {
                fprintf(stderr, MSG_INVALID_LEVEL_ZERO, progname, argv[i], argv[i + 1], newline);
                err_status = -1;
                usage(invocation);
              }
            }
            else
            {
              fprintf(stderr, MSG_SCOPE_UNSUPPORTED, progname, argv[i], argv[i + 1], newline);
              err_status = -1;
              usage(invocation);
            }
          }
          else if (strcmp((char *) argv[i + 1], "timestamp") == 0 && nscope == 0)
          {
            if (!nscope)
            {
              sw_print[level]->timestamp_attr = TRUE;
              i += 2;
            }
            else
            {
              fprintf(stderr, MSG_SCOPE_UNSUPPORTED, progname, argv[i], argv[i + 1], newline);
              err_status = -1;
              usage(invocation);
            }
          }
          else if (strcmp((char *) argv[i + 1], "insert") == 0 || strcmp((char *) argv[i + 1], "attrname") == 0)
          {
            if (i < argc - 3)
            {
              allocate_insert_attr(level, argv[i + 2], argv[i + 3]);
              i += 4;
            }
            else
            {
              fprintf(stderr, MSG_PARAMETER_MISSING_TWO_REQUIRED, progname, argv[i], argv[i + 1], newline);
              err_status = -1;
              usage(invocation);
            }
          }
          else
          {
            if (nscope)
              fprintf(stderr, MSG_SCOPE_UNSUPPORTED, progname, argv[i], argv[i + 1], newline);
            else
              fprintf(stderr, MSG_PARAMETER_INVALID, progname, argv[i], argv[i + 1], newline);
            err_status = -1;
            usage(invocation);
          }
        }
        else 
        {
          if (level == -15)
            fprintf(stderr, MSG_SCOPE_INVALID, progname, argv[i], newline);
          else if (level == -13)
            fprintf(stderr, MSG_SCOPE_MISSING, progname, argv[i], newline);
          else if (level == -12)
            fprintf(stderr, MSG_SCOPE_NO_LEVEL, progname, argv[i], newline);
          else
            fprintf(stderr, MSG_INVALID_LEVEL, progname, argv[i], newline);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, MSG_PARAMETER_MISSING, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Print snippets
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--printonly") == 0 || strcmp((char *) argv[i], "-p") == 0))
    {
      /* set all switches to off but only for the first instance */
      if (sw_printonly_first_opt)
      {
        sw_printonly_header        = FALSE;
        sw_printonly_footer        = FALSE;
        sw_printonly_rtagopen      = FALSE;
        sw_printonly_rtagclose     = FALSE;
        sw_printonly_records       = FALSE;
        /* Management switch for the printonly switches */
        sw_printonly_first_opt     = FALSE;
      }

      if (i < argc - 1)
      {
        if (strcmp((char *) argv[i + 1], "header") == 0)
        {
          sw_printonly_header = TRUE;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "footer") == 0)
        {
          sw_printonly_footer = TRUE;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "rtagopen") == 0)
        {
          sw_printonly_rtagopen = TRUE;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "rtagclose") == 0)
        {
          sw_printonly_rtagclose = TRUE;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "records") == 0)
        {
          sw_printonly_records = TRUE;
          i += 2;
        }
        else
        {
          fprintf(stderr, MSG_PARAMETER_INVALID, progname, argv[i], argv[i + 1], newline);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, MSG_PARAMETER_MISSING, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Key/value pairs
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--keyvaluepairs") == 0 || strcmp((char *) argv[i], "-k") == 0))
    {
      if (sw_schema == SCHEMA_DEFAULT)
      {
        sw_key_value_pairs = TRUE;
        sw_kvp_levels_all  = TRUE;
        i += 1;
      }
      else
      {
        fprintf(stderr, MSG_KVP_SCHEMA_INVALID_OPTION, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }

    if (i < argc
        && ((strncmp((char *) argv[i], "--keyvaluepairs", 15) == 0 && xstrlen(argv[i]) > 15)
        || (strncmp((char *) argv[i], "-k", 2) == 0 && xstrlen(argv[i]) > 2)))
    {
      if (sw_schema == SCHEMA_DEFAULT)
      {
        if (xstrlen(argv[i]) > 15)
        {
          level = atoi((char *) &argv[i][15]);
          /* atoi() is too smart */
          for (j = 11 ; j < xstrlen(argv[i]) ; j++)
            if (!isdigit(argv[i][j]))
              level = -1;
        }
        else
        {
          level = atoi((char *) &argv[i][2]);
          /* atoi() is too smart */
          for (j = 2 ; j < xstrlen(argv[i]) ; j++)
            if (!isdigit(argv[i][j]))
              level = -1;
        }
        if (level >= nlevels)
          increase_nlevels(level + 1);
  
        if (level > 1 && level < nlevels)
        {
          sw_key_value_pairs      = TRUE;
          kvp_tag[level]->enabled = TRUE;
          i += 1;
        }
        else 
        {
          fprintf(stderr, MSG_INVALID_LEVEL, progname, argv[i], newline);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, MSG_KVP_SCHEMA_INVALID_OPTION, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Expel records and fields
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--expel") == 0 || strcmp((char *) argv[i], "-E") == 0))
    {
      if (i < argc - 1)
      {
        if (!process_arg_expel(argv[i + 1]))
        {
          fprintf(stderr, MSG_INVALID_RANGE, progname, argv[i], argv[i + 1], newline);
          err_status = -1;
          usage(invocation);
        }
        i += 2;
      }
      else
      {
        fprintf(stderr, MSG_NO_RANGE, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }

    
    /*
     * Option: Specify system identifier
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--identifier") == 0 || strcmp((char *) argv[i], "-I") == 0))
    {
      if (i < argc - 1)
      {
        schema_system_identifier = argv[i + 1];
        i += 2;
      }
      else
      {
        fprintf(stderr, MSG_SYSTEM_IDENTIFIER_PARAMETER_MISSING, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Quotechar
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--quotechars") == 0 || strcmp((char *) argv[i], "-Q") == 0))
    {
      if (!quotechars_length)
      {
        if (i < argc - 1)
        {
          quotechar_type = 1;
          quotechars = argv[i + 1];
          quotechars_length = xstrlen(quotechars);
          i += 2;
        }
        else
        {
          fprintf(stderr, MSG_QUOTECHARS_ARRAY_PARAMETER_MISSING, progname, argv[i], newline);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, MSG_QUOTECHARS_ARRAY_ONLY_ONE_ALLOWED, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Quotechar using paired quotes
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--quotechars2") == 0 || strcmp((char *) argv[i], "-Q2") == 0))
    {
      if (!quotechars_length)
      {
        if (i < argc - 1)
        {
          quotechars = argv[i + 1];
          quotechars_length = xstrlen(quotechars);
          if (quotechars_length && quotechars_length % 2 == 0)
          {
            quotechar_type = 2;
            i += 2;
          }
          else
          {
            fprintf(stderr, MSG_QUOTECHARS_ARRAY_PARAMETER_UNPAIRED, progname, argv[i], newline);
            err_status = -1;
            usage(invocation);
          }
        }
        else
        {
          fprintf(stderr, MSG_QUOTECHARS_ARRAY_PARAMETER_MISSING, progname, argv[i], newline);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, MSG_QUOTECHARS_ARRAY_ONLY_ONE_ALLOWED, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Specify XML handler
     */
    if (i < argc
        && (strcmp((char *) argv[i], "--xmlformat") == 0 || strcmp((char *) argv[i], "-X") == 0))
    {
      if (i < argc - 1)
      {
        if (strcmp((char *) argv[i + 1], "XML1.0") == 0)
        {
          sw_xml_format = SW_XML;
          version_xml   = XML_1_0;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "XML1.1") == 0)
        {
          sw_xml_format = SW_XML;
          version_xml   = XML_1_1;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "SOAP1.1") == 0)
        {
          sw_xml_format = SW_SOAP;
          version_soap  = SOAP_1_1;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "SOAP1.2") == 0)
        {
          sw_xml_format = SW_SOAP;
          version_soap  = SOAP_1_2;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "BOM") == 0)
        {
          sw_bom = TRUE;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], UTF_8_NAME) == 0)
        {
          sw_encoding = UTF_8;
          xstrcpy(output_record.encoding_name, argv[i + 1]);
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], UTF_16_NAME) == 0
                 || strcmp((char *) argv[i + 1], UTF_16BE_NAME) == 0)
        {
          sw_encoding = UTF_16BE;
          xstrcpy(output_record.encoding_name, argv[i + 1]);
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], UTF_16LE_NAME) == 0)
        {
          sw_encoding = UTF_16LE;
          xstrcpy(output_record.encoding_name, argv[i + 1]);
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], UTF_32_NAME) == 0
                 || strcmp((char *) argv[i + 1], UTF_32BE_NAME) == 0)
        {
          sw_encoding = UTF_32BE;
          xstrcpy(output_record.encoding_name, argv[i + 1]);
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], UTF_32LE_NAME) == 0)
        {
          sw_encoding = UTF_32LE;
          xstrcpy(output_record.encoding_name, argv[i + 1]);
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "ASCIItoUTF") == 0)
        {
          sw_ascii_to_utf = TRUE;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "noescape") == 0)
        {
          if (i < argc - 2)
          {
            if (strcmp((char *) argv[i + 2], "all") == 0)
            {
              sw_char_escape_amp    = FALSE;
              sw_char_escape_lt     = FALSE;
              sw_char_escape_gt     = FALSE;
              sw_char_escape_quot   = FALSE;
              sw_char_escape_apos   = FALSE;
              sw_char_escape_brvbar = FALSE;
              i += 3;
            }
            else if (strcmp((char *) argv[i + 2], "amp") == 0)
            {
              sw_char_escape_amp = FALSE;
              i += 3;
            }
            else if (strcmp((char *) argv[i + 2], "lt") == 0)
            {
              sw_char_escape_lt = FALSE;
              i += 3;
            }
            else if (strcmp((char *) argv[i + 2], "gt") == 0)
            {
              sw_char_escape_gt = FALSE;
              i += 3;
            }
            else if (strcmp((char *) argv[i + 2], "quot") == 0)
            {
              sw_char_escape_quot = FALSE;
              i += 3;
            }
            else if (strcmp((char *) argv[i + 2], "apos") == 0)
            {
              sw_char_escape_apos = FALSE;
              i += 3;
            }
            else if (strcmp((char *) argv[i + 2], "brvbar") == 0)
            {
              sw_char_escape_brvbar = FALSE;
              i += 3;
            }
            else
            {
              fprintf(stderr, MSG_PARAMETER_SECOND_INVALID, progname, argv[i], argv[i + 1], argv[i + 2], newline);
              err_status = -1;
              usage(invocation);
            }
          }
          else
          {
            fprintf(stderr, MSG_PARAMETER_MISSING_TWO_REQUIRED, progname, argv[i], argv[i + 1], newline);
            err_status = -1;
            usage(invocation);
          }
        }
        else if (strcmp((char *) argv[i + 1], "trimtagclose") == 0)
        {
          sw_trimtagclose = TRUE;
          i += 2;
        }
        else if (strcmp((char *) argv[i + 1], "newline") == 0)
        {
          if (i < argc - 2)
          {
            if (strcmp((char *) argv[i + 2], "dos") == 0)
            {
              xstrcpy(newline, (unsigned char *) NEWLINE_DOS);
              i += 3;
            }
            else if (strcmp((char *) argv[i + 2], "unix") == 0)
            {
              xstrcpy(newline, (unsigned char *) NEWLINE_UNIX);
              i += 3;
            }
            else
            {
              fprintf(stderr, MSG_PARAMETER_SECOND_INVALID, progname, argv[i], argv[i + 1], argv[i + 2], newline);
              err_status = -1;
              usage(invocation);
            }
          }
          else
          {
            fprintf(stderr, MSG_PARAMETER_MISSING_TWO_REQUIRED, progname, argv[i], argv[i + 1], newline);
            err_status = -1;
            usage(invocation);
          }
        }
        else if (strcmp((char *) argv[i + 1], "HTML") == 0)
        {
          if (i < argc - 2 && sw_xml_format == SW_XML)
          {
            if (strcmp((char *) argv[i + 2], "table") == 0)
            {
              sw_xml_format = SW_HTML;
              /* Overwrite default parent tagname. Ensure new tagname fits inside allocated memory */
              xstrcpy(default_schema_parent_name, (unsigned char *) DEFAULT_STR_HTML_TABLE);
              html_parent_tagname = (unsigned char *) malloc(sizeof(unsigned char) * (xstrlen((unsigned char *) DEFAULT_STR_HTML_TR) + 1));
              html_child_tagname = (unsigned char *) malloc(sizeof(unsigned char) * (xstrlen((unsigned char *) DEFAULT_STR_HTML_TD) + 1));
              xstrcpy(html_parent_tagname, (unsigned char *) DEFAULT_STR_HTML_TR);
              xstrcpy(html_child_tagname, (unsigned char *) DEFAULT_STR_HTML_TD);
              i += 3;
            }
            else if (strcmp((char *) argv[i + 2], "list") == 0)
            {
              sw_xml_format = SW_HTML;
              /* Overwrite default parent tagname. Ensure new tagname fits inside allocated memory */
              xstrcpy(default_schema_parent_name, (unsigned char *) DEFAULT_STR_HTML_P);
              html_parent_tagname = (unsigned char *) malloc(sizeof(unsigned char) * (xstrlen((unsigned char *) DEFAULT_STR_HTML_UL) + 1));
              html_child_tagname = (unsigned char *) malloc(sizeof(unsigned char) * (xstrlen((unsigned char *) DEFAULT_STR_HTML_LI) + 1));
              xstrcpy(html_parent_tagname, (unsigned char *) DEFAULT_STR_HTML_UL);
              xstrcpy(html_child_tagname, (unsigned char *) DEFAULT_STR_HTML_LI);
              i += 3;
            }
            else
            {
              fprintf(stderr, MSG_PARAMETER_SECOND_INVALID, progname, argv[i], argv[i + 1], argv[i + 2], newline);
              err_status = -1;
              usage(invocation);
            }
          }
          else
          {
            fprintf(stderr, MSG_PARAMETER_MISSING_TWO_REQUIRED, progname, argv[i], argv[i + 1], newline);
            err_status = -1;
            usage(invocation);
          }
        }
        else
        {
          fprintf(stderr, MSG_PARAMETER_INVALID, progname, argv[i], argv[i + 1], newline);
          err_status = -1;
          usage(invocation);
        }
      }
      else
      {
        fprintf(stderr, MSG_PARAMETER_MISSING, progname, argv[i], newline);
        err_status = -1;
        usage(invocation);
      }
    }


    /*
     * Option: Unmatched or unsupported option
     */
    if (i == i_prev)
    {
      fprintf(stderr, MSG_UNKNOWN_OPTION, progname, argv[i], newline);
      err_status = -1;
      usage(invocation);
    }
  } /* end while */

  /* set the encoding */
  utf_set_encoding();
}


void populate_delimiter(level, d)
int level;
unsigned char *d;
{
  /* add delimiter string for use at specified level */
  unsigned char *new_d, *new_utf_d;
  int i, j, new_d_length;

  /* assign delimiter */
  if (strcmp((char *) d, "\\n") == 0)       /* new line */
    new_d = delimiter_newline;
  else if (strcmp((char *) d, "\\r") == 0)  /* carriage return */
    new_d = delimiter_cr;
  else if (strcmp((char *) d, "\\t") == 0)  /* tab */
    new_d = delimiter_tab;
  else if (strcmp((char *) d, " ") == 0)    /* space */
    new_d = delimiter_space;
  else
    new_d = d;

  new_d_length = xstrlen(new_d);


  if (delimiter[level]->ndelimiters < 1)
  {
    delimiter[level]->utf_d = (unsigned char **) malloc(sizeof(unsigned char *));
    if (delimiter[level]->utf_d == NULL) out_of_mem();
    delimiter[level]->utf_d[0] = (unsigned char *) malloc(sizeof(unsigned char) * (UTF_32_INCREMENTER * new_d_length + 1));
    if (delimiter[level]->utf_d[0] == NULL) out_of_mem();
    new_utf_d = delimiter[level]->utf_d[0];
    delimiter[level]->utf_length = (int *) malloc(sizeof(int));
    if (delimiter[level]->utf_length == NULL) out_of_mem();
    delimiter[level]->arg_d = (unsigned char **) malloc(sizeof(unsigned char *));
    if (delimiter[level]->arg_d == NULL) out_of_mem();
    delimiter[level]->arg_length = (int *) malloc(sizeof(int));
    if (delimiter[level]->arg_length == NULL) out_of_mem();
    delimiter[level]->byte_mode = (int *) malloc(sizeof(int));
    if (delimiter[level]->byte_mode == NULL) out_of_mem();
    delimiter[level]->nscope = (int *) malloc(sizeof(int));
    if (delimiter[level]->nscope == NULL) out_of_mem();
    delimiter[level]->scope = (struct_scope **) malloc(sizeof(struct_scope *));
    if (delimiter[level]->scope == NULL) out_of_mem();
    delimiter[level]->scope_seq = (long *) malloc(sizeof(long));
    if (delimiter[level]->scope_seq == NULL) out_of_mem();
  }
  else
  {
    delimiter[level]->utf_d = (unsigned char **) realloc(delimiter[level]->utf_d, (delimiter[level]->ndelimiters + 1) * sizeof(unsigned char *));
    if (delimiter[level]->utf_d == NULL) out_of_mem();
    delimiter[level]->utf_d[delimiter[level]->ndelimiters] = (unsigned char *) malloc(sizeof(unsigned char) * (UTF_32_INCREMENTER * new_d_length + 1));
    if (delimiter[level]->utf_d[delimiter[level]->ndelimiters] == NULL) out_of_mem();
    new_utf_d = delimiter[level]->utf_d[delimiter[level]->ndelimiters];
    delimiter[level]->utf_length = (int *) realloc(delimiter[level]->utf_length, (delimiter[level]->ndelimiters + 1) * sizeof(int));
    if (delimiter[level]->utf_length == NULL) out_of_mem();
    delimiter[level]->arg_d = (unsigned char **) realloc(delimiter[level]->arg_d, (delimiter[level]->ndelimiters + 1) * sizeof(unsigned char *));
    if (delimiter[level]->arg_d == NULL) out_of_mem();
    delimiter[level]->arg_length = (int *) realloc(delimiter[level]->arg_length, (delimiter[level]->ndelimiters + 1) * sizeof(int));
    if (delimiter[level]->arg_length == NULL) out_of_mem();
    delimiter[level]->byte_mode = (int *) realloc(delimiter[level]->byte_mode, (delimiter[level]->ndelimiters + 1) * sizeof(int));
    if (delimiter[level]->byte_mode == NULL) out_of_mem();
    delimiter[level]->nscope = (int *) realloc(delimiter[level]->nscope, (delimiter[level]->ndelimiters + 1) * sizeof(int));
    if (delimiter[level]->nscope == NULL) out_of_mem();
    delimiter[level]->scope = (struct_scope **) realloc(delimiter[level]->scope, (delimiter[level]->ndelimiters + 1) * sizeof(struct_scope *));
    if (delimiter[level]->scope == NULL) out_of_mem();
    delimiter[level]->scope_seq = (long *) realloc(delimiter[level]->scope_seq, (delimiter[level]->ndelimiters + 1) * sizeof(long));
    if (delimiter[level]->scope_seq == NULL) out_of_mem();
  }

  delimiter[level]->ndelimiters++;

  /* re-sort the delimiters in size from large to small */
  if (delimiter[level]->ndelimiters == 1)
  {
    delimiter[level]->utf_d[0]      = new_utf_d;
    /* delimiter[level]->utf_length[0] is initialised when encoding is known */
    delimiter[level]->arg_d[0]      = new_d;
    delimiter[level]->arg_length[0] = new_d_length;
    delimiter[level]->byte_mode[0]  = delimiter_byte_mode;
    delimiter[level]->nscope[0]     = nscope;
    delimiter[level]->scope[0]      = scope;
    delimiter[level]->scope_seq[0]  = 0;
  }
  else
  {
    for (i = 0 ; i < delimiter[level]->ndelimiters - 1 ; i++)
    {
      if (new_d_length > delimiter[level]->arg_length[i])
      {
        for (j = delimiter[level]->ndelimiters - 1 ; j > i ; j--)
        {
          delimiter[level]->utf_d[j]      = delimiter[level]->utf_d[j - 1];
          /* delimiter[level]->utf_length[j] is initialised when encoding is known */
          delimiter[level]->arg_d[j]      = delimiter[level]->arg_d[j - 1];
          delimiter[level]->arg_length[j] = delimiter[level]->arg_length[j - 1];
          delimiter[level]->byte_mode[j]  = delimiter[level]->byte_mode[j - 1];
          delimiter[level]->nscope[j]     = delimiter[level]->nscope[j - 1];
          delimiter[level]->scope[j]      = delimiter[level]->scope[j - 1];
          /* delimiter[level]->scope_seq[j] is all initialised as 0 */
        }
        break; /* break for/i */
      }
    } /* end for/i */
    delimiter[level]->utf_d[i]      = new_utf_d;
    /* delimiter[level]->utf_length[i] is initialised when encoding is known */
    delimiter[level]->arg_d[i]      = new_d;
    delimiter[level]->arg_length[i] = new_d_length;
    delimiter[level]->byte_mode[i]  = delimiter_byte_mode;
    delimiter[level]->nscope[i]     = nscope;
    delimiter[level]->scope[i]      = scope;
    delimiter[level]->scope_seq[i]  = 0;
  }
}


void increase_nlevels(newlevel)
int newlevel;
{
  /* allocate level objects for new level */

  /* allocate mem */
  delimiter = (struct_delimiter **) realloc(delimiter, newlevel * sizeof(struct_delimiter *));
  if (delimiter == NULL) out_of_mem();
  sw_print = (struct_sw_print **) realloc(sw_print, newlevel * sizeof(struct_sw_print *));
  if (sw_print == NULL) out_of_mem();
  kvp_tag = (struct_kvp_tag **) realloc(kvp_tag, newlevel * sizeof(struct_kvp_tag *));
  if (kvp_tag == NULL) out_of_mem();
  output_record.field_cntr = (int *) realloc(output_record.field_cntr, newlevel * sizeof(int));
  if (output_record.field_cntr == NULL) out_of_mem();

  /* update attributes of new levels */
  while (nlevels < newlevel)
  {
    delimiter[nlevels] = (struct_delimiter *) malloc(sizeof(struct_delimiter));
    delimiter[nlevels]->ndelimiters = 0;

    sw_print[nlevels] = (struct_sw_print *) malloc(sizeof(struct_sw_print));
    if (sw_print[nlevels] == NULL) out_of_mem();
    sw_print[nlevels]->tagname        = sw_print_default.tagname;
    sw_print[nlevels]->tagname_length = xstrlen(sw_print[nlevels]->tagname);
    sw_print[nlevels]->numbered_attr  = sw_print_default.numbered_attr;
    sw_print[nlevels]->numbered_tag   = sw_print_default.numbered_tag;
    sw_print[nlevels]->delimiter_attr = sw_print_default.delimiter_attr;
    sw_print[nlevels]->level_attr     = sw_print_default.level_attr;
    sw_print[nlevels]->level_tag      = sw_print_default.level_tag;
    sw_print[nlevels]->timestamp_attr = sw_print_default.timestamp_attr;

    kvp_tag[nlevels] = (struct_kvp_tag *) malloc(sizeof(struct_kvp_tag));
    if (kvp_tag[nlevels] == NULL) out_of_mem();
    kvp_tag[nlevels]->enabled  = FALSE;
    kvp_tag[nlevels]->set      = FALSE;
    kvp_tag[nlevels]->name     = sw_print[nlevels]->tagname;
    kvp_tag[nlevels]->encoding = UTF_8;
    kvp_tag[nlevels]->length   = sw_print[nlevels]->tagname_length;
    kvp_tag[nlevels]->level_d  = 0;
    kvp_tag[nlevels]->match_d  = DELIMITER_INDEX_UNMATCHED;

    output_record.field_cntr[nlevels] = 0;

    nlevels++;
  }
}


void usage(invocation)
int invocation;
{
  /* Show quick usage */
  if (invocation == ARGS_SCHEMA)
    fprintf(stderr, MSG_SCHEMA_ILLEGAL_OPTION, progname, schema_filename, newline);
  fprintf(stderr, "Usage: %s [OPTION]...%s", progname, newline);
  fprintf(stderr, "  -h,  --help%s", newline);
  fprintf(stderr, "  -v,  --version%s", newline);
  fprintf(stderr, "  --license%s", newline);
  fprintf(stderr, "  --debug%s", newline);
  fprintf(stderr, "%s", newline);
  fprintf(stderr, "Input options:%s", newline);
  fprintf(stderr, "  -F,  --fieldseparator[<level>[b][:<scope>]] <string>%s", newline);
  fprintf(stderr, "  -R,  --recordseparator <string>%s", newline);
  fprintf(stderr, "  -C,  --column[:<scope>] <c1>-<c2>[:<name>]%s", newline);
  fprintf(stderr, "  -W,  --regex[:<scope>] [E|B][i][l][r][U][n][b][e]/<pattern>/[<name>[,..]]%s", newline);
  fprintf(stderr, "  -e,  --expelempty%s", newline);
  fprintf(stderr, "  -E,  --expel <input_records>[:<input_fields>]%s", newline);
  fprintf(stderr, "  -q,  --quotedfields[2]%s", newline);
  fprintf(stderr, "  -Q,  --quotechars[2] <string>%s", newline);
  fprintf(stderr, "  -b,  --blanklines%s", newline);
  fprintf(stderr, "  -t,  --trim%s", newline);
  fprintf(stderr, "%s", newline);
  fprintf(stderr, "Output options:%s", newline);
  fprintf(stderr, "  -S,  --schema <file>%s", newline);
  fprintf(stderr, "  -Sd, --schemadtd <file>%s", newline);
  fprintf(stderr, "  -Sr, --schemarnc <file>%s", newline);
  fprintf(stderr, "  -Sx, --schemaxsd <file>%s", newline);
  fprintf(stderr, "  -M,  --matchdirect 0|<elementname>%s", newline);
  fprintf(stderr, "  -A,  --attribute[<level>[:<scope>]] delimiter|number|level|timestamp%s", newline);
  fprintf(stderr, "          |insert <name> <value>%s", newline);
  fprintf(stderr, "  -T,  --tag[<level>[:<scope>]] number|level%s", newline);
  fprintf(stderr, "          |name <name>%s", newline);
  fprintf(stderr, "          |[re]insert <name> <value>%s", newline);
  fprintf(stderr, "          |[re]insertfile <name> <filename>%s", newline);
  fprintf(stderr, "          |[re]insertfilexml <indent> <filename>%s", newline);
  fprintf(stderr, "  -k,  --keyvaluepairs[<level>]%s", newline);
  fprintf(stderr, "  -l,  --linenumbers%s", newline);
  fprintf(stderr, "  -f,  --fieldnumbers%s", newline);
  fprintf(stderr, "  -L,  --linetags%s", newline);
  fprintf(stderr, "  -X,  --xmlformat [XML1.0|XML1.1]|[SOAP1.1|SOAP1.2]|[HTML table|list]%s", newline);
  fprintf(stderr, "          |[UTF-8|UTF-16|UTF-16BE|UTF16-LE|UTF-32|UTF-32BE|UTF-32LE]|BOM%s", newline);
  fprintf(stderr, "          |ASCIItoUTF|[noescape all|amp|lt|gt|quot|apos|brvbar]%s", newline);
  fprintf(stderr, "          |trimtagclose|[newline dos|unix]%s", newline);
  fprintf(stderr, "  -p,  --printonly header|footer|rtagopen|rtagclose|records%s", newline);
  fprintf(stderr, "  -I,  --identifier <system_identifier>%s", newline);
  fprintf(stderr, "  -s,  --summary[2|c|n|f <file>]%s", newline);
  fprintf(stderr, "  -U,  --unxml%s", newline);
  fprintf(stderr, "  --noxml%s", newline);

  cleanup();

  exit(err_status);
}


void version()
{
  /* Show version */
  fprintf(stderr, "%s: Version %s%s", progname, VERSION, newline);
  fprintf(stderr, "Copyright (c) 2008-2013, Arthur Gouros.%s", newline);
  fprintf(stderr, "Refer to the %s license for terms and conditions.%s", progname, newline);
  exit(err_status);
}

