/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2013, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert to XML on the fly.                                     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "xmlout.h"

long f_s;
int  p_curr_xpath;
int  j_matching_record;

/* function templates */
void match_schema();
int  populate_element(int i, int level);
int  populate_field_tagnames(int level, int i_helper, int h, int i);


void match_schema()
{
  /* process one record and convert it to XML format if a match with a schema element helper is found */
  int i, k;
  int flag_processed;
  long n;
  struct_input_field *tmpfield;

  if (sw_debug)
    fprintf(stderr, "DEBUG: start function: match_schema%s", newline);

  input_record.nfields_cdata = 0;
  f_s = 0;

  /* With schema processing all input fields are flattened to be just (level 2) fields.
   * Ensure fields of type FIELD_SUBRECORD and FIELD_EXPELLED are at the end of array 
   * for contiguous FIELD_CDATA processing.
   * Most fields of type FIELD_SUBRECORD should have already been recycled by the input
   * processor.
   * Unfortunately this loop acts like a brake when encountering lots of fields that 
   * need repositioning, however its been optimised for both speed and ugliness. */
  if (sw_debug)
    fprintf(stderr, "DEBUG: Sorting %ld input field types%s", input_record.nfields, newline);

  flag_processed = FALSE;
  while (!flag_processed)
  {
    if (input_record.field[f_s]->type == FIELD_CDATA)
    {
      /* we use nfields_cdata instead of nfields for schema processing */
      input_record.nfields_cdata++;
      /* initialise the field's xpath */ 
      input_record.field[f_s]->schema_xpath[0] = 0;
      /* increment */
      f_s++;
      if (f_s >= input_record.nfields)
        flag_processed = TRUE;
    }
    else
    {
      /* set loop to terminate */
      flag_processed = TRUE;
      /* save current field */
      tmpfield = input_record.field[f_s];
      /* shift fields down */
      for (n = f_s ; n < input_record.nfields - 1 ; n++)
      {
        input_record.field[n] = input_record.field[n + 1];
        /* check if we need to set loop to continue */
        if (flag_processed && input_record.field[n]->type == FIELD_CDATA)
          flag_processed = FALSE; 
      }
      /* put saved field at the end */
      input_record.field[n] = tmpfield;
    }
  } /* end while */

  if (sw_debug)
    fprintf(stderr, "DEBUG: Total CDATA input fields=%ld%s", input_record.nfields_cdata, newline);

  flag_processed = FALSE;
  if (schema_match_start_level)
  {
    /* iterate through level 1 elements looking for a nfields constraint match
    * then check their match control */
    for (k = 0 ; k < schema_parent.record->nfields && !flag_processed ; k++)
    {
      /* In this mode (default/legacy) we only process elements of the root element that are records */
      if (schema_parent.record->field[k].type != FIELD_SUBRECORD)
        continue; /* continue for/k */

      for (i = 1 ; i < schema_parent.nelements ; i++)
      {
        if (xstrcmp(schema_parent.record->field[k].name, schema_element[i].name) == 0)
        {
          schema_line_name = schema_element[i].name;
          j_matching_record = 0;
          if (populate_element(i, 2))
          {
            switch (schema_parent.record->field[k].match_control)
            { 
              case MATCH_NONE_OR_ONE: 
              case MATCH_ONE:  
                  /* don't process MATCH_NONE_OR_ONE if already matched */
                  if (schema_parent.record->field[k].field_count == 0)
                  {
                    schema_parent.record->field[k].field_count += 1;
                    schema_element[i].element_count += 1;
                    schema_element[i].record[j_matching_record]->record_count += 1;
                    process_output(schema_element[i].element_count);
                    /* break the for-loop iterations */
                    flag_processed = TRUE;
                  }
                  break; /* break switch */
        
              case MATCH_ONE_OR_MANY:
              case MATCH_NONE_OR_MANY:
                  schema_parent.record->field[k].field_count += 1;
                  schema_element[i].element_count += 1;
                  schema_element[i].record[j_matching_record]->record_count += 1;
                  process_output(schema_element[i].element_count);
                  /* break the for-loop iterations */
                  flag_processed = TRUE;
            } /* end switch */
            break; /* break for/i */
          }
        }
      } /* end for/i */
    } /* end for/k */
  }
  else
  {
    /* In this mode (match element direct) we use the helpers to match and process the
     * specified element record and make it the root element. Only ONE record is
     * matched and printed in this mode as there can only be one root element. */
    if (xstrcmp(schema_match_start_name, (unsigned char *) "0") == 0)
    {
      if (!record_cntr_printed && populate_element(0, 1))
      {
        schema_element[0].element_count += 1;
        process_output(schema_element[0].element_count);
        /* set flag just for the counter updates */
        flag_processed = TRUE; 
      }
    }
    else
    {
      for (i = 0 ; i < schema_parent.nelements && !record_cntr_printed ; i++)
      {
        if (xstrcmp(schema_match_start_name, schema_element[i].name) == 0)
        {
          if (populate_element(i, 1))
          {
            schema_element[i].element_count += 1;
            process_output(schema_element[i].element_count);
            /* set flag just for the counter updates */
            flag_processed = TRUE; 
            break; /* for/i */
          }
        }
      } /* end for/i */
    }
  }

  /* increment unmatched counter if loop exhausted naturally */
  if (!flag_processed)
  {
    if (sw_blank_lines || input_record.nfields)
      record_cntr_unmatched++;
    else
      record_cntr_blank++;
  }

  if (sw_debug)
    fprintf(stderr, "DEBUG: end function: match_schema%s", newline);
}


int populate_element(i, level)
int i;
int level;
{
  /* Iterate through element match helpers and process ones that match the field match constraints */
  int h;
  int flag_match_element;

  flag_match_element = FALSE;
  
  for (h = 0 ; h < schema_element[i].element_helper->nxpaths && !flag_match_element ; h++)
  {
    f_s = 0;
    p_curr_xpath = -1;
    if (input_record.nfields_cdata >= schema_element[i].element_helper->match_helper[h].min)
      if (schema_element[i].element_helper->match_helper[h].max < 0 || input_record.nfields_cdata <= schema_element[i].element_helper->match_helper[h].max)
      {
        if (sw_debug)
          fprintf(stderr, "DEBUG: Initial match on helper: %s.h[%d/%d]%s", schema_element[i].name, h, schema_element[i].element_helper->nxpaths - 1, newline);

        flag_match_element = populate_field_tagnames(level, i, h, i);

        if (sw_debug)
        {
          if (flag_match_element)
            debug_print_helper_xpaths("Full match found:", i, h);
          else
            fprintf(stderr, "DEBUG: Fail match on helper: %s.h[%d/%d]%s", schema_element[i].name, h, schema_element[i].element_helper->nxpaths - 1, newline);
        }
      }
  } /* end for/h */

  return(flag_match_element);
}


int populate_field_tagnames(level, i_helper, h, i)
int level;
int i_helper;
int h;
int i;
{
  /* populate the field tagnames according to field wildcard rules */
  long f_s_parentr, sequence, oops;
  int j, k;
  int i_subr, h_subr, p_subr, p_saved_xpath;
  int min, max, last;
  int flag_match_record, flag_process_deeper;
  
  flag_match_record   = FALSE;
  flag_process_deeper = TRUE;
  oops                = 1000000000L;
  k                   = 0;
  sequence            = 0;
  min                 = -1; /* never match */
  max                 = 0; 

  f_s_parentr = f_s;
  p_curr_xpath++;
  p_saved_xpath = p_curr_xpath;
  j = schema_element[i_helper].element_helper->match_helper[h].xpath[p_curr_xpath];

  /* set the min/max/last values */
  if (level > schema_match_start_level + 1)
  {
    /* find the values of the current subrecord */
    for (h_subr = 0 ; h_subr < schema_element[i].element_helper->nxpaths; h_subr++)
    {
      for (p_subr = 0 ; p_subr < schema_element[i].element_helper->max_xpath_seq && schema_element[i].element_helper->match_helper[h_subr].xpath[p_subr] >= 0 ; p_subr++)
        if (schema_element[i].element_helper->match_helper[h_subr].xpath[p_subr] != schema_element[i_helper].element_helper->match_helper[h].xpath[p_curr_xpath + p_subr])
          break; /* break for/p_subr */
      if (p_subr == schema_element[i].element_helper->max_xpath_seq || schema_element[i].element_helper->match_helper[h_subr].xpath[p_subr] < 0)
      {
        min = schema_element[i].element_helper->match_helper[h_subr].min;
        max = schema_element[i].element_helper->match_helper[h_subr].max;
        break; /* break for/h_subr */
      }
    } /* end for/h_subr */
  }
  else
  {
    /* find the values of the current record */
    min = schema_element[i_helper].element_helper->match_helper[h].min;
    max = schema_element[i_helper].element_helper->match_helper[h].max;
  }
  last = schema_element[i_helper].element_helper->match_helper[h].last;

  while (f_s < input_record.nfields_cdata && k < schema_element[i].record[j]->nfields && flag_process_deeper && oops-- > 1)
  {
    if (schema_element[i].record[j]->field[k].type == FIELD_CDATA)
    {
      switch (schema_element[i].record[j]->field[k].match_control)
      {
        case MATCH_ONE:
              input_record.field[f_s]->tagname = schema_element[i].record[j]->field[k].name;
              input_record.field[f_s]->level = level;
              input_record.field[f_s]->schema_xpath[level - 1] = i;
              input_record.field[f_s]->sequence = sequence++;
              f_s++;
              k++;
              break; /* break switch */

        case MATCH_NONE_OR_ONE:
              if (k < schema_element[i].record[j]->nfields - last && f_s < input_record.nfields_cdata - last)
              {
                input_record.field[f_s]->tagname = schema_element[i].record[j]->field[k].name;
                input_record.field[f_s]->level = level;
                input_record.field[f_s]->schema_xpath[level - 1] = i;
                input_record.field[f_s]->sequence = sequence++;
                f_s++;
              }
              k++;
              break; /* break switch */

        case MATCH_ONE_OR_MANY:
              if (k < schema_element[i].record[j]->nfields - last && f_s < input_record.nfields_cdata - last)
              {
                input_record.field[f_s]->tagname = schema_element[i].record[j]->field[k].name;
                input_record.field[f_s]->level = level;
                input_record.field[f_s]->schema_xpath[level - 1] = i;
                input_record.field[f_s]->sequence = sequence++;
                f_s++;
              }
              else
                k++;
              break; /* break switch */

        case MATCH_NONE_OR_MANY:
              if (k < schema_element[i].record[j]->nfields - last && f_s < input_record.nfields_cdata - last)
              {
                input_record.field[f_s]->tagname = schema_element[i].record[j]->field[k].name;
                input_record.field[f_s]->level = level;
                input_record.field[f_s]->schema_xpath[level - 1] = i;
                input_record.field[f_s]->sequence = sequence++;
                f_s++;
              }
              else
                k++;
              break; /* break switch */
      } /* end switch */
    }
    else /* schema field type is FIELD_SUBRECORD */
    {
      switch (schema_element[i].record[j]->field[k].match_control)
      {
        case MATCH_ONE_OR_MANY:
        case MATCH_NONE_OR_MANY:
              p_saved_xpath = p_curr_xpath;
              break;
      }
      /* find matching element and traverse */
      for (i_subr = schema_match_start_level ; i_subr < schema_parent.nelements ; i_subr++)
      {
        if (xstrcmp(schema_element[i].record[j]->field[k].name, schema_element[i_subr].name) == 0)
        {
          /* add xpath when going deeper */
          input_record.field[f_s]->schema_xpath[level - 1] = i;
          /* go deeper */
          flag_process_deeper = populate_field_tagnames(level + 1, i_helper, h, i_subr);
          /* add xpath when coming back up (could be for a different f_s) */
          input_record.field[f_s]->schema_xpath[level - 1] = i;
          break; /* break for/i_subr */
        }
      } /* end for/i_subr */

      /* process wildcard tokens for FIELD_SUBRECORD */
      switch (schema_element[i].record[j]->field[k].match_control)
      {
        case MATCH_ONE:
        case MATCH_NONE_OR_ONE:
              k++;
              break; /* break switch */

        case MATCH_ONE_OR_MANY:
        case MATCH_NONE_OR_MANY:
              if (k < schema_element[i].record[j]->nfields - last && f_s < input_record.nfields_cdata - last)
              {
                /* reset xpath progress back for another pass */
                p_curr_xpath = p_saved_xpath;
              }
              else
                k++;
              break; /* break switch */

      } /* end switch */
    }
  } /* end while */

  if (!oops) {
    fprintf(stderr, "%s: ERROR: While processing field wildcard sequence the program entered into an infinite loop.%s", progname, newline);
    fprintf(stderr, "       Please check your schema file syntax in file %s is correct.%s", schema_filename, newline);
    err_status = -3;
    exit(err_status);
  }

  if (level > schema_match_start_level + 1) 
  {
    if (f_s - f_s_parentr >= min || max < 0)
      flag_match_record = TRUE;
  }
  else
  {
    if (flag_process_deeper && f_s == input_record.nfields_cdata && (f_s - f_s_parentr >= min || max < 0))
    {
      flag_match_record = TRUE;
      /* increase record_count of matching record of current parent element */
      j_matching_record = j;
    }
  }

  if (sw_debug)
    if (flag_match_record)
      debug_print_helper_xpaths("Record match: OK  ", i, h);

  return(flag_match_record);
}
