/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2013, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert to XML on the fly.                                     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "xmlout.h"

long i_curr_field;

/* function templates */
void process_output_record_default(long record_number);
void process_output_record_kvp(long record_number);
void process_output_record_schema(long record_number);
void process_schema_level(int level);
void process_output_record_html(long record_number);
void mkelem(int level, unsigned char *t, int t_length, int t_encoding, unsigned char *f, long f_length, int f_encoding, long n, int sw_numbered_tag, int sw_numbered_attr, int level_d, int d, int sw_delimiter_attr, int sw_level_tag, int sw_level_attr, int sw_timestamp_attr, int sw_custom_attr, int sw_custom_tag);
void mktag(int tag_control, int level, unsigned char *t, int t_length, int t_encoding, long n, int sw_numbered_tag, int sw_numbered_attr, int level_d, int d, int sw_delimiter_attr, int sw_level_tag, int sw_level_attr, int sw_timestamp_attr, int sw_custom_attr, int sw_custom_tag);
void make_timestamp_array(time_t t, int type);
void append_escaped_string(unsigned char *s, int s_length);
void format_key_value_tagname(long i);
void print_insert_attr(int level);
void print_insert_tag(int level, int flag_reinsert);


void process_output(record_number)
long record_number;
{
  /* print an output record in XML format */

  if (sw_blank_lines || input_record.nfields)
  {
    /* print an output record in the requested format */
    if (sw_schema == SCHEMA_DEFAULT)
    {
      if (!sw_key_value_pairs)
      {
        if (sw_xml_format != SW_HTML)
          process_output_record_default(record_number);
        else
          process_output_record_html(record_number);
      }
      else
        process_output_record_kvp(record_number);
    }
    else
    {
      if (schema_match_start_level)
        /* Process from schema parent element (default legacy mode). */
        process_output_record_schema(record_number);
      else
      {
        /* Process from a user specified schema element. */
        i_curr_field = 0;
        process_schema_level(1);
      }
    }

    /* increment record printed counter */
    record_cntr_printed++;
  }
  else if (ncolumn && input_record.length) /* No level 2 fields created */
    record_cntr_unmatched++;
  else
    record_cntr_blank++;
}


void process_output_record_default(record_number)
long record_number;
{
  /* print an output record in default XML format */
  long i;
  int current_level;

  /* initialise counters */
  for (i = 0 ; i < nlevels ; i++)
    output_record.field_cntr[i] = 0;

  /* open record tag */
  mktag(TAG_START,
    1,
    sw_print[1]->tagname, 
    sw_print[1]->tagname_length, 
    UTF_8,
    record_number,
    sw_print[1]->numbered_tag,
    sw_print[1]->numbered_attr,
    1,
    delimiter[1]->match_d,
    sw_print[1]->delimiter_attr,
    sw_print[1]->level_tag,
    sw_print[1]->level_attr,
    sw_print[1]->timestamp_attr,
    ninsert_attr,
    ninsert_tag);
  utf_print(STDOUT_FD, output_record.line, output_record.length);

  /* init level */
  current_level = 2;

  /* print any insert tags for this new depth level */
  if (ninsert_tag)
    print_insert_tag(current_level, FALSE);

  /* print a line tag */
  if (sw_line_tags) 
  {
    (void) sprintf(temp_array, "<%s>%ld</%s>%s", schema_linenumber, record_number, schema_linenumber, newline);
    strcpy_output_record(2, (unsigned char *) temp_array);
    utf_print(STDOUT_FD, output_record.line, output_record.length);
  }

  for (i = 0 ; i < input_record.nfields ; i++)
  {
    while (input_record.field[i]->level != current_level)
    {
      if (input_record.field[i]->level > current_level)
      {
        /* print tree of opening field tags */
        mktag(TAG_START,
          current_level,
          sw_print[current_level]->tagname, 
          sw_print[current_level]->tagname_length, 
          UTF_8,
          output_record.field_cntr[current_level] + 1,
          sw_print[current_level]->numbered_tag,
          sw_print[current_level]->numbered_attr,
          input_record.field[i - 1]->level_d,  /* See NOTE-ON-SEGV below for explanation */
          input_record.field[i - 1]->match_d,  /* See NOTE-ON-SEGV below for explanation */
          sw_print[current_level]->delimiter_attr,
          sw_print[current_level]->level_tag,
          sw_print[current_level]->level_attr,
          sw_print[current_level]->timestamp_attr,
          ninsert_attr,
          ninsert_tag);
        utf_print(STDOUT_FD, output_record.line, output_record.length);
        current_level++;
        /* print any insert tags for this new depth level */
        if (ninsert_tag)
          print_insert_tag(current_level, FALSE);
        /* initialise next deeper level numbering */
        if (current_level < nlevels)
          output_record.field_cntr[current_level] = 0;
      }
      else if (input_record.field[i]->level < current_level)
      {
        /* print tree of closing field tags */
        current_level--;
        mktag(TAG_END,
          current_level,
          sw_print[current_level]->tagname, 
          sw_print[current_level]->tagname_length, 
          UTF_8,
          output_record.field_cntr[current_level] + 1,
          sw_print[current_level]->numbered_tag,
          0,
          0,
          DELIMITER_INDEX_UNMATCHED,
          0,
          sw_print[current_level]->level_tag,
          0,
          0,
          0,
          0);
        utf_print(STDOUT_FD, output_record.line, output_record.length);
        output_record.field_cntr[current_level] += 1;
      }
    }

    /* print a complete field element at its depth level */
    if (input_record.field[i]->type == FIELD_CDATA)
    {
      mkelem(input_record.field[i]->level,
        input_record.field[i]->tagname, 
        xstrlen(input_record.field[i]->tagname), 
        UTF_8,
        &input_record.line[input_record.field[i]->start], 
        input_record.field[i]->length,
        input_record.encoding,
        output_record.field_cntr[input_record.field[i]->level] + 1,
        sw_print[input_record.field[i]->level]->numbered_tag,
        sw_print[input_record.field[i]->level]->numbered_attr,
        input_record.field[i]->level_d,
        input_record.field[i]->match_d,
        sw_print[input_record.field[i]->level]->delimiter_attr,
        sw_print[input_record.field[i]->level]->level_tag,
        sw_print[input_record.field[i]->level]->level_attr,
        sw_print[input_record.field[i]->level]->timestamp_attr,
        ninsert_attr,
        ninsert_tag);
      utf_print(STDOUT_FD, output_record.line, output_record.length);
      output_record.field_cntr[input_record.field[i]->level] += 1;
    }
  }

  /* print tree of closing field tags */
  while (current_level > 2)
  {
    current_level--;
    mktag(TAG_END,
      current_level,
      sw_print[current_level]->tagname, 
      sw_print[current_level]->tagname_length, 
      UTF_8,
      output_record.field_cntr[current_level] + 1,
      sw_print[current_level]->numbered_tag,
      0,
      0,
      DELIMITER_INDEX_UNMATCHED,
      0,
      sw_print[current_level]->level_tag,
      0,
      0,
      0,
      0);
    utf_print(STDOUT_FD, output_record.line, output_record.length);
    output_record.field_cntr[current_level] += 1;
  }
 
  /* close record tag */
  mktag(TAG_END,
    1,
    sw_print[1]->tagname, 
    sw_print[1]->tagname_length, 
    UTF_8,
    record_number,
    sw_print[1]->numbered_tag,
    0,
    0,
    DELIMITER_INDEX_UNMATCHED,
    0,
    sw_print[1]->level_tag,
    0,
    0,
    0,
    0);
  utf_print(STDOUT_FD, output_record.line, output_record.length);
}


void process_output_record_kvp(record_number)
long record_number;
{
  /* print an output record in key/value pairs XML format */
  long i, j, k;
  int current_level, record_type;
  int flag_drop_level;

  /* initialise */
  for (current_level = 0 ; current_level < nlevels ; current_level++)
  {
    output_record.field_cntr[current_level]   = 0;
    kvp_tag[current_level]->set = FALSE;
  }
  output_record.field_cntr[1] = record_number - 1;

  /* itentify FIELD_KEYTAG_* candidates */
  for (i = 0 ; i < input_record.nfields ; i++)
  {
    if (input_record.field[i]->type == FIELD_CDATA)
    {
      /* check if restricted to specified levels */
      if (!sw_kvp_levels_all)
      {
        if (!kvp_tag[input_record.field[i]->level]->enabled)
          continue;
      }

      /* blank fields don't make good tags so coerce them out */
      if (!sw_blank_lines && !input_record.field[i]->length)
      {
        input_record.field[i]->type = FIELD_SUBRECORD;
        continue;
      }

      /* look ahead */
      k = 0;
      for (j = i + 1 ; j < input_record.nfields ; j++)
      {
        if (k < 2 && input_record.field[j]->level >= input_record.field[i]->level)
        {
          if (input_record.field[j]->type == FIELD_CDATA && kvp_tag[input_record.field[j]->level]->set == FALSE)
            k++;
          /* unset the kvp flag */
          if (input_record.field[j]->level > input_record.field[i]->level)
            kvp_tag[input_record.field[i]->level]->set = FALSE;
        }
        else
          break;
      } /* end for/j */

      if (k == 1)
      {
        input_record.field[i]->type = FIELD_KEYTAG_ELEMENT;
        format_key_value_tagname(i);

        i = j;
      }
      else if (k > 1)
      {
        if (!kvp_tag[input_record.field[i]->level]->set)
        {
          input_record.field[i]->type = FIELD_KEYTAG_RECORD;
          format_key_value_tagname(i);

          kvp_tag[input_record.field[i]->level]->set = TRUE;
          for (current_level = input_record.field[i]->level + 1 ; current_level < nlevels ; current_level++)
            kvp_tag[current_level]->set = FALSE;
        }

        j = i + 1;
        while (j < input_record.nfields && input_record.field[j]->level == input_record.field[i]->level 
            && (input_record.field[j]->type == FIELD_CDATA || input_record.field[j]->type == FIELD_EXPELLED))
          j++;
        i = j - 1;
      }
    }
    else if (input_record.field[i]->type == FIELD_SUBRECORD)
    {
      for (current_level = input_record.field[i]->level + 1 ; current_level < nlevels ; current_level++)
        kvp_tag[current_level]->set = FALSE;
    }
  } /* end for/i */

  /* adjust levels to reflect the revised field layout */
  for (i = 0 ; i < input_record.nfields ; i++)
  {
    if (input_record.field[i]->type == FIELD_KEYTAG_ELEMENT)
    {
      /* backward update */
      j = i - 1;
      while (j >= 0 && input_record.field[j]->level > 1
          && ((input_record.field[j]->level <= input_record.field[i]->level && input_record.field[j]->type == FIELD_SUBRECORD)
          || (input_record.field[j]->level == input_record.field[i]->level && input_record.field[j]->type == FIELD_EXPELLED)))
        input_record.field[j--]->level--;
      /* forward update */
      input_record.field[i]->level--;
      i++;
      if (i < input_record.nfields)
        input_record.field[i]->level--;
    }
    else if (input_record.field[i]->type == FIELD_KEYTAG_RECORD)
    {
      /* check if 2 consecutive superfluous levels exist */
      flag_drop_level = TRUE;
      for (j = i + 1; flag_drop_level && j < input_record.nfields && input_record.field[j]->level >= input_record.field[i]->level ; j++)
      {
        if (input_record.field[j]->level == input_record.field[i]->level || input_record.field[j]->level == input_record.field[i]->level + 1) 
          if (input_record.field[j]->type == FIELD_CDATA || input_record.field[j]->type == FIELD_KEYTAG_ELEMENT || input_record.field[j]->type == FIELD_KEYTAG_RECORD)
            flag_drop_level = FALSE;
      }
          
      if (flag_drop_level)
      {
          /* remove the lower superfluous level that has no fields as it messes up the key/pair output format */
          for (k = i + 1 ; k < j ; k++)
            input_record.field[k]->level--;

          input_record.field[i]->level--;
      }
      else
      {
        /* backward update */
        j = i - 1;
        while (j >= 0 && input_record.field[j]->level > 1 
            && ((input_record.field[j]->level <= input_record.field[i]->level && input_record.field[j]->type == FIELD_SUBRECORD)
            || (input_record.field[j]->level == input_record.field[i]->level && input_record.field[j]->type == FIELD_EXPELLED)))
          input_record.field[j--]->level--;
        /* forward update */
        j = i + 1;
        while (j < input_record.nfields && input_record.field[j]->level == input_record.field[i]->level
            && (input_record.field[j]->type == FIELD_CDATA || input_record.field[j]->type == FIELD_EXPELLED))
           j++;
        input_record.field[i]->level--;
        i = j - 1;
      }
    }
  } /* end for/i */

  /* 
   * output adjusted fields of this record 
   */

  kvp_tag[1]->name     = sw_print[1]->tagname; 
  kvp_tag[1]->length   = sw_print[1]->tagname_length; 
  kvp_tag[1]->encoding = UTF_8;
  kvp_tag[1]->level_d  = 1;
  kvp_tag[1]->match_d  = delimiter[1]->match_d;
  record_type          = FIELD_KEYTAG_RECORD;

  /* find the first field for becoming the key of key/value */
  for (i = 0 ; i < input_record.nfields ; i++)
  {
    if (input_record.field[i]->type == FIELD_SUBRECORD || input_record.field[i]->type == FIELD_EXPELLED)
      continue;

    if (input_record.field[i]->level == 1)
    {
      /* set the parent tag name of this level to be the text of the first field of the current level */
      if (input_record.field[i]->type == FIELD_KEYTAG_ELEMENT || input_record.field[i]->type == FIELD_KEYTAG_RECORD)
      {
        /* set the key */
        if (input_record.field[i]->length)
        {
          kvp_tag[1]->name     = &input_record.line[input_record.field[i]->start];
          kvp_tag[1]->length   = input_record.field[i]->length;
          kvp_tag[1]->encoding = input_record.encoding;
        }
        else
        {
          kvp_tag[1]->name     = sw_print[1]->tagname;
          kvp_tag[1]->length   = sw_print[1]->tagname_length;
          kvp_tag[1]->encoding = UTF_8;
        }
        kvp_tag[1]->level_d = input_record.field[i]->level_d;
        kvp_tag[1]->match_d = input_record.field[i]->match_d;
        record_type         = input_record.field[i]->type;
      }
    }
    /* if the first valid field is not a tag then use the default */
    break;
  } /* end for/i */

  if (record_type != FIELD_KEYTAG_ELEMENT)
  {
    /* open record tag */
    mktag(TAG_START,
      1,
      kvp_tag[1]->name, 
      kvp_tag[1]->length, 
      kvp_tag[1]->encoding, 
      record_number,
      sw_print[1]->numbered_tag,
      sw_print[1]->numbered_attr,
      kvp_tag[1]->level_d,
      kvp_tag[1]->match_d,
      sw_print[1]->delimiter_attr,
      sw_print[1]->level_tag,
      sw_print[1]->level_attr,
      sw_print[1]->timestamp_attr,
      ninsert_attr,
      ninsert_tag);
    utf_print(STDOUT_FD, output_record.line, output_record.length);

    /* init level */
    current_level = 2;

    /* print any insert tags for this new depth level */
    if (ninsert_tag)
      print_insert_tag(current_level, FALSE);

    /* print a line tag */
    if (sw_line_tags) 
    {
      (void) sprintf(temp_array, "<%s>%ld</%s>%s", schema_linenumber, record_number, schema_linenumber, newline);
      strcpy_output_record(current_level, (unsigned char *) temp_array);
      utf_print(STDOUT_FD, output_record.line, output_record.length);
    }
  
    for (i = 0 ; i < input_record.nfields ; i++)
    {
      if (input_record.field[i]->level < 2 || input_record.field[i]->type == FIELD_EXPELLED)
        continue;

      if (input_record.field[i]->level < current_level)
      {
        while (input_record.field[i]->level != current_level)
        {
          /* print tree of closing field tags */
          current_level--;
          mktag(TAG_END,
            current_level,
            kvp_tag[current_level]->name, 
            kvp_tag[current_level]->length, 
            kvp_tag[current_level]->encoding, 
            output_record.field_cntr[current_level] + 1,
            sw_print[current_level]->numbered_tag,
            0,
            0,
            DELIMITER_INDEX_UNMATCHED,
            0,
            sw_print[current_level]->level_tag,
            0,
            0,
            0,
            0);
          utf_print(STDOUT_FD, output_record.line, output_record.length);
          output_record.field_cntr[current_level] += 1;
          /* reset tag name */
          kvp_tag[current_level + 1]->name     = sw_print[current_level + 1]->tagname; 
          kvp_tag[current_level + 1]->length   = sw_print[current_level + 1]->tagname_length; 
          kvp_tag[current_level + 1]->encoding = UTF_8;
          if (current_level == 2 && (input_record.field[i]->type == FIELD_CDATA || input_record.field[i]->type == FIELD_SUBRECORD))
          {
            kvp_tag[current_level]->name     = sw_print[current_level]->tagname; 
            kvp_tag[current_level]->length   = sw_print[current_level]->tagname_length; 
            kvp_tag[current_level]->encoding = UTF_8;
          }
        }
      }

      if (input_record.field[i]->type == FIELD_KEYTAG_RECORD || input_record.field[i]->type == FIELD_KEYTAG_ELEMENT)
      {
        /* set the key */
        if (input_record.field[i]->length)
        {
          kvp_tag[input_record.field[i]->level]->name     = &input_record.line[input_record.field[i]->start];
          kvp_tag[input_record.field[i]->level]->length   = input_record.field[i]->length;
          kvp_tag[input_record.field[i]->level]->encoding = input_record.encoding;
        }
        else
        {
          kvp_tag[input_record.field[i]->level]->name     = sw_print[input_record.field[i]->level]->tagname;
          kvp_tag[input_record.field[i]->level]->length   = sw_print[input_record.field[i]->level]->tagname_length;
          kvp_tag[input_record.field[i]->level]->encoding = UTF_8;
        }
        kvp_tag[input_record.field[i]->level]->level_d    = input_record.field[i]->level_d;
        kvp_tag[input_record.field[i]->level]->match_d    = input_record.field[i]->match_d;
        if (input_record.field[i]->type == FIELD_KEYTAG_RECORD && input_record.field[i]->level + 1 < nlevels)
        {
          kvp_tag[input_record.field[i]->level + 1]->name     = sw_print[input_record.field[i]->level + 1]->tagname; 
          kvp_tag[input_record.field[i]->level + 1]->length   = sw_print[input_record.field[i]->level + 1]->tagname_length; 
          kvp_tag[input_record.field[i]->level + 1]->encoding = UTF_8;
        }
      }

      if (input_record.field[i]->level > current_level)
      {
        while (input_record.field[i]->level != current_level)
        {
          /* print tree of opening field tags */
          if (input_record.field[i]->type == FIELD_CDATA || input_record.field[i]->type == FIELD_SUBRECORD)
          {
            /* set correct delimiter for this tag */
            kvp_tag[current_level]->level_d = input_record.field[i - 1]->level_d; 
            kvp_tag[current_level]->match_d = input_record.field[i - 1]->match_d; 
            /* NOTE-ON-SEGV: input_record.field[0]->level should always start at initialised 
             * current_level meaning this code is never reached when i = 0 
             * I could use 
             *   i ? input_record.field[i - 1]->level_d : 2 
             *   i ? input_record.field[i - 1]->match_d : delimiter[2]->match_d 
             * but the current setting has allowed me to identify wrong behaviour. */
          }
          mktag(TAG_START,
            current_level,
            kvp_tag[current_level]->name, 
            kvp_tag[current_level]->length, 
            kvp_tag[current_level]->encoding, 
            output_record.field_cntr[current_level] + 1,
            sw_print[current_level]->numbered_tag,
            sw_print[current_level]->numbered_attr,
            kvp_tag[current_level]->level_d,
            kvp_tag[current_level]->match_d,
            sw_print[current_level]->delimiter_attr,
            sw_print[current_level]->level_tag,
            sw_print[current_level]->level_attr,
            sw_print[current_level]->timestamp_attr,
            ninsert_attr,
            ninsert_tag);
          utf_print(STDOUT_FD, output_record.line, output_record.length);
          current_level++;
          /* print any insert tags for this new depth level */
          if (ninsert_tag)
            print_insert_tag(current_level, FALSE);
          /* initialise next deeper level numbering */
          if (current_level < nlevels)
            output_record.field_cntr[current_level] = 0;
        }
      }

      /* print a complete field element at its depth level */
      if (input_record.field[i]->type == FIELD_CDATA)
      {
        if (sw_blank_lines || input_record.field[i]->length > 0 || kvp_tag[input_record.field[i]->level]->name != sw_print[input_record.field[i]->level]->tagname)
        {
          mkelem(input_record.field[i]->level,
            kvp_tag[input_record.field[i]->level]->name, 
            kvp_tag[input_record.field[i]->level]->length, 
            kvp_tag[input_record.field[i]->level]->encoding, 
            &input_record.line[input_record.field[i]->start], 
            input_record.field[i]->length,
            input_record.encoding,
            output_record.field_cntr[input_record.field[i]->level] + 1,
            sw_print[input_record.field[i]->level]->numbered_tag,
            sw_print[input_record.field[i]->level]->numbered_attr,
            input_record.field[i]->level_d,
            input_record.field[i]->match_d,
            sw_print[input_record.field[i]->level]->delimiter_attr,
            sw_print[input_record.field[i]->level]->level_tag,
            sw_print[input_record.field[i]->level]->level_attr,
            sw_print[input_record.field[i]->level]->timestamp_attr,
            ninsert_attr,
            ninsert_tag);
          utf_print(STDOUT_FD, output_record.line, output_record.length);
          output_record.field_cntr[input_record.field[i]->level] += 1;

          if (current_level == 2)
          {
            /* reset tag name */
            kvp_tag[current_level]->name     = sw_print[current_level]->tagname; 
            kvp_tag[current_level]->length   = sw_print[current_level]->tagname_length; 
            kvp_tag[current_level]->encoding = UTF_8;
          }
        }
      }
    } /* end for/i */

    /* print tree of closing field tags */
    while (current_level > 2)
    {
      current_level--;
      mktag(TAG_END,
        current_level,
        kvp_tag[current_level]->name, 
        kvp_tag[current_level]->length, 
        kvp_tag[current_level]->encoding, 
        output_record.field_cntr[current_level] + 1,
        sw_print[current_level]->numbered_tag,
        0,
        0,
        DELIMITER_INDEX_UNMATCHED,
        0,
        sw_print[current_level]->level_tag,
        0,
        0,
        0,
        0);
      utf_print(STDOUT_FD, output_record.line, output_record.length);
      output_record.field_cntr[current_level] += 1;
      /* reset tag name */
      kvp_tag[current_level + 1]->name     = sw_print[current_level + 1]->tagname; 
      kvp_tag[current_level + 1]->length   = sw_print[current_level + 1]->tagname_length; 
      kvp_tag[current_level + 1]->encoding = UTF_8;
    }
    /* reset level 2 tag name */
    kvp_tag[2]->name     = sw_print[2]->tagname; 
    kvp_tag[2]->length   = sw_print[2]->tagname_length; 
    kvp_tag[2]->encoding = UTF_8;
  
    /* close record tag */
    mktag(TAG_END,
      1,
      kvp_tag[1]->name, 
      kvp_tag[1]->length, 
      kvp_tag[1]->encoding, 
      record_number,
      sw_print[1]->numbered_tag,
      0,
      0,
      DELIMITER_INDEX_UNMATCHED,
      0,
      sw_print[1]->level_tag,
      0,
      0,
      0,
      0);
    utf_print(STDOUT_FD, output_record.line, output_record.length);
  }
  else  /* record_type = FIELD_KEYTAG_ELEMENT */
  {
    for (i = 0 ; i < input_record.nfields ; i++)
    {
      if (input_record.field[i]->type == FIELD_KEYTAG_ELEMENT)
      {
        if (input_record.field[i]->length)
        {
          kvp_tag[input_record.field[i]->level]->name     = &input_record.line[input_record.field[i]->start];
          kvp_tag[input_record.field[i]->level]->length   = input_record.field[i]->length;
          kvp_tag[input_record.field[i]->level]->encoding = input_record.encoding;
        }
        else
        {
          kvp_tag[input_record.field[i]->level]->name      = sw_print[input_record.field[i]->level]->tagname;
          kvp_tag[input_record.field[i]->level]->length    = sw_print[input_record.field[i]->level]->tagname_length;
          kvp_tag[input_record.field[i]->level]->encoding  = UTF_8;
        }
        kvp_tag[input_record.field[i]->level]->level_d = input_record.field[i]->level_d;
        kvp_tag[input_record.field[i]->level]->match_d = input_record.field[i]->match_d;
      }
      else if (input_record.field[i]->type == FIELD_CDATA)
      {
        mkelem(input_record.field[i]->level,
          kvp_tag[input_record.field[i]->level]->name, 
          kvp_tag[input_record.field[i]->level]->length, 
          kvp_tag[input_record.field[i]->level]->encoding, 
          &input_record.line[input_record.field[i]->start], 
          input_record.field[i]->length,
          input_record.encoding,
          output_record.field_cntr[input_record.field[i]->level] + 1,
          sw_print[input_record.field[i]->level]->numbered_tag,
          sw_print[input_record.field[i]->level]->numbered_attr,
          kvp_tag[input_record.field[i]->level]->level_d,
          kvp_tag[input_record.field[i]->level]->match_d,
          sw_print[input_record.field[i]->level]->delimiter_attr,
          sw_print[input_record.field[i]->level]->level_tag,
          sw_print[input_record.field[i]->level]->level_attr,
          sw_print[input_record.field[i]->level]->timestamp_attr,
          ninsert_attr,
          ninsert_tag);
        utf_print(STDOUT_FD, output_record.line, output_record.length);
        output_record.field_cntr[input_record.field[i]->level] += 1;
      }
    } /* end for/i */
  } /* end if/record_type */
}


void process_output_record_schema(record_number)
long record_number;
{
  /* print an output record using schema for XML format */

  /* open schema record tag */
  mktag(TAG_START,
    1,
    schema_line_name, 
    xstrlen(schema_line_name), 
    UTF_8,
    record_number,
    sw_print[1]->numbered_tag,
    sw_print[1]->numbered_attr,
    1,
    delimiter[1]->match_d,
    sw_print[1]->delimiter_attr,
    sw_print[1]->level_tag,
    sw_print[1]->level_attr,
    sw_print[1]->timestamp_attr,
    ninsert_attr,
    ninsert_tag);
  utf_print(STDOUT_FD, output_record.line, output_record.length);

  /* print any insert tags for this new depth level */
  if (ninsert_tag)
    print_insert_tag(2, FALSE);

  /* print a line tag */
  if (sw_line_tags) 
  {
    (void) sprintf(temp_array, "<%s>%ld</%s>%s", schema_linenumber, record_number, schema_linenumber, newline);
    strcpy_output_record(2, (unsigned char *) temp_array);
    utf_print(STDOUT_FD, output_record.line, output_record.length);
  }

  /* print schema tree tags and data */
  i_curr_field = 0;
  process_schema_level(2);

  /* close schema record tag */
  mktag(TAG_END,
    1,
    schema_line_name, 
    xstrlen(schema_line_name), 
    UTF_8,
    record_number,
    sw_print[1]->numbered_tag,
    0,
    0,
    DELIMITER_INDEX_UNMATCHED,
    0,
    sw_print[1]->level_tag,
    0,
    0,
    0,
    0);
  utf_print(STDOUT_FD, output_record.line, output_record.length);
}


void process_schema_level(level)
int level;
{
  /* process the input fields that now have an association with the schema elements */
  long i, j, f_cntr, prev_sequence;
  int flag_process_deeper;

  prev_sequence = -1;
  flag_process_deeper = TRUE;

  /* we use i to capture the tree hierachy for the tag closure in the recursion */
  i = i_curr_field;

  while (i_curr_field < input_record.nfields_cdata && flag_process_deeper)
  {
    if (input_record.field[i_curr_field]->level == level && input_record.field[i_curr_field]->sequence > prev_sequence)
    {
      prev_sequence = input_record.field[i_curr_field]->sequence;

      /* re-calculate field number as fields may occur in punctuated sequences */
      f_cntr = 0;
      if (sw_print[level]->numbered_tag || sw_print[level]->numbered_attr)
      {
        for (j = i_curr_field ; j >= 0 && input_record.field[j]->level >= level ; j--) 
        {
          if (input_record.field[j]->level == level) 
          {
            if (xstrcmp(input_record.field[i_curr_field]->tagname, input_record.field[j]->tagname) == 0)
              f_cntr++;
            if (input_record.field[j]->sequence < 1)
               break; /* break for/j */
          }
        } /* end for/j */
      }

      mkelem(input_record.field[i_curr_field]->level,
        input_record.field[i_curr_field]->tagname, 
        xstrlen(input_record.field[i_curr_field]->tagname), 
        UTF_8,
        &input_record.line[input_record.field[i_curr_field]->start], 
        input_record.field[i_curr_field]->length,
        input_record.encoding,
        f_cntr,
        sw_print[input_record.field[i_curr_field]->level]->numbered_tag,
        sw_print[input_record.field[i_curr_field]->level]->numbered_attr,
        input_record.field[i_curr_field]->level_d,
        input_record.field[i_curr_field]->match_d,
        sw_print[input_record.field[i_curr_field]->level]->delimiter_attr,
        sw_print[input_record.field[i_curr_field]->level]->level_tag,
        sw_print[input_record.field[i_curr_field]->level]->level_attr,
        sw_print[input_record.field[i_curr_field]->level]->timestamp_attr,
        ninsert_attr,
        ninsert_tag);
      utf_print(STDOUT_FD, output_record.line, output_record.length);

      i_curr_field++;
      i = i_curr_field;
    }
    else if (input_record.field[i_curr_field]->level > level)
    {
      /* re-calculate field number as fields may occur in punctuated sequences */
      f_cntr = 0;
      if (sw_print[level]->numbered_tag || sw_print[level]->numbered_attr)
      {
        for (j = i ; j >= 0 && (input_record.field[j]->level > level || (schema_match_start_level == 0 && input_record.field[j]->level == level)) ; j--) 
          if (input_record.field[i]->level == input_record.field[j]->level && input_record.field[j]->sequence == 0) 
            if (input_record.field[i]->schema_xpath[level] == input_record.field[j]->schema_xpath[level])
              f_cntr++;
      }

      /* print opening field tag */
      /* NOTE: Tree delimiters are meaningless with flattened input field structure
       * as they would have exactly the same value as the field delimiter.
       *   input_record.field[i]->level_d,
       *   input_record.field[i]->match_d,
       *   sw_print[level]->delimiter_attr, 
       * The above is what would have been used if they weren't, but it is the 
       * schema that dictates the tree structure not the delimiter.
       */
      mktag(TAG_START,
        level,
        schema_element[input_record.field[i]->schema_xpath[level]].name, 
        xstrlen(schema_element[input_record.field[i]->schema_xpath[level]].name), 
        UTF_8,
        f_cntr,
        sw_print[level]->numbered_tag,
        sw_print[level]->numbered_attr,
        0,
        DELIMITER_INDEX_UNMATCHED,
        0,
        sw_print[level]->level_tag,
        sw_print[level]->level_attr,
        sw_print[level]->timestamp_attr,
        ninsert_attr,
        ninsert_tag);
      utf_print(STDOUT_FD, output_record.line, output_record.length);

      /* print any insert tags for this new depth level */
      if (ninsert_tag)
        print_insert_tag(level + 1, FALSE);

      process_schema_level(level + 1);

      /* print closing field tag */
      mktag(TAG_END,
        level,
        schema_element[input_record.field[i]->schema_xpath[level]].name, 
        xstrlen(schema_element[input_record.field[i]->schema_xpath[level]].name), 
        UTF_8,
        f_cntr,
        sw_print[level]->numbered_tag,
        0,
        0,
        DELIMITER_INDEX_UNMATCHED,
        0,
        sw_print[level]->level_tag,
        0,
        0,
        0,
        0);
      utf_print(STDOUT_FD, output_record.line, output_record.length);

      i = i_curr_field;
    }
    else
      flag_process_deeper = FALSE;
  }
}


void process_output_record_html(record_number)
long record_number;
{
  /* print an output record in HTML format */
  long i;
  int current_level;

  /* initialise counters */
  for (i = 0 ; i < nlevels ; i++)
    output_record.field_cntr[i] = 0;

  /* open record tag */
  mktag(TAG_START,
    1,
    html_parent_tagname, 
    xstrlen(html_parent_tagname), 
    UTF_8,
    record_number,
    0,
    0,
    0,
    DELIMITER_INDEX_UNMATCHED,
    0,
    0,
    0,
    0,
    ninsert_attr,
    ninsert_tag);
  utf_print(STDOUT_FD, output_record.line, output_record.length);

  /* init level */
  current_level = 2;

  /* print any insert tags for this new depth level */
  if (ninsert_tag)
    print_insert_tag(current_level, FALSE);

  /* print a line tag */
  if (sw_line_tags) 
  {
    (void) sprintf(temp_array, "<%s>%ld</%s>%s", schema_linenumber, record_number, schema_linenumber, newline);
    strcpy_output_record(3, (unsigned char *) temp_array);
    utf_print(STDOUT_FD, output_record.line, output_record.length);
  }

  for (i = 0 ; i < input_record.nfields ; i++)
  {
    while (input_record.field[i]->level != current_level)
    {
      if (input_record.field[i]->level > current_level)
      {
        /* print tree of opening field tags */
        mktag(TAG_START,
          current_level,
          html_parent_tagname, 
          xstrlen(html_parent_tagname), 
          UTF_8,
          output_record.field_cntr[current_level] + 1,
          0,
          0,
          0,
          DELIMITER_INDEX_UNMATCHED,
          0,
          0,
          0,
          0,
          ninsert_attr,
          ninsert_tag);
        utf_print(STDOUT_FD, output_record.line, output_record.length);
        current_level++;
        /* print any insert tags for this new depth level */
        if (ninsert_tag)
          print_insert_tag(current_level, FALSE);
        /* initialise next deeper level numbering */
        if (current_level < nlevels)
          output_record.field_cntr[current_level] = 0;
      }
      else if (input_record.field[i]->level < current_level)
      {
        /* print tree of closing field tags */
        current_level--;
        mktag(TAG_END,
          current_level,
          html_parent_tagname, 
          xstrlen(html_parent_tagname), 
          UTF_8,
          output_record.field_cntr[current_level] + 1,
          0,
          0,
          0,
          DELIMITER_INDEX_UNMATCHED,
          0,
          0,
          0,
          0,
          0,
          0);
        utf_print(STDOUT_FD, output_record.line, output_record.length);
        output_record.field_cntr[current_level] += 1;
      }
    }

    /* print a complete field element at its depth level */
    if (input_record.field[i]->type == FIELD_CDATA)
    {
      mkelem(input_record.field[i]->level,
        html_child_tagname, 
        xstrlen(html_child_tagname), 
        UTF_8,
        &input_record.line[input_record.field[i]->start], 
        input_record.field[i]->length,
        input_record.encoding,
        output_record.field_cntr[input_record.field[i]->level] + 1,
        0,
        0,
        0,
        DELIMITER_INDEX_UNMATCHED,
        0,
        0,
        0,
        0,
        ninsert_attr,
        ninsert_tag);
      utf_print(STDOUT_FD, output_record.line, output_record.length);
      output_record.field_cntr[input_record.field[i]->level] += 1;
    }
  }

  /* print tree of closing field tags */
  while (current_level > 2)
  {
    current_level--;
    mktag(TAG_END,
      current_level,
      html_parent_tagname, 
      xstrlen(html_parent_tagname), 
      UTF_8,
      output_record.field_cntr[current_level] + 1,
      0,
      0,
      0,
      DELIMITER_INDEX_UNMATCHED,
      0,
      0,
      0,
      0,
      0,
      0);
    utf_print(STDOUT_FD, output_record.line, output_record.length);
    output_record.field_cntr[current_level] += 1;
  }
 
  /* close record tag */
  mktag(TAG_END,
    1,
    html_parent_tagname, 
    xstrlen(html_parent_tagname), 
    UTF_8,
    record_number,
    0,
    0,
    0,
    DELIMITER_INDEX_UNMATCHED,
    0,
    0,
    0,
    0,
    0,
    0);
  utf_print(STDOUT_FD, output_record.line, output_record.length);
}


void mkelem(level, t, t_length, t_encoding, f, f_length, f_encoding, n, sw_numbered_tag, sw_numbered_attr, level_d, d, sw_delimiter_attr, sw_level_tag, sw_level_attr, sw_timestamp_attr, sw_custom_attr, sw_custom_tag)
int level;
unsigned char *t;
int t_length;
int t_encoding;
unsigned char *f;
long f_length;
int f_encoding;
long n;
int sw_numbered_tag;
int sw_numbered_attr;
int level_d;
int d;
int sw_delimiter_attr;
int sw_level_tag;
int sw_level_attr;
int sw_timestamp_attr;
int sw_custom_attr;
int sw_custom_tag;
{
  /* Construct an element tag array suitable for XML and UTF-8 printing */
  /* Array output_record.line and output_record.length are updated */
  long i;
  int j, f_charwidth;

  /* set the UTF charwidth of the array supplied in the argument. */
  f_charwidth = 1;
  if (f_encoding == UTF_16BE || f_encoding == UTF_16LE)
    f_charwidth = 2;
  else if (f_encoding == UTF_32BE || f_encoding == UTF_32LE)
    f_charwidth = 4;

  /* print re-insert tag */
  if (sw_custom_tag)
    print_insert_tag(level, TRUE);

  /* open element tag */
  if (sw_xml_format == SW_XML)
    strcpy_output_record(level, (unsigned char *) "<");
  else if (sw_xml_format == SW_SOAP)
  {
    strcpy_output_record(level + 2, (unsigned char *) "<");
    strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_XMLFY);
    strcat_output_record((unsigned char *) ":");
  }
  else if (sw_xml_format == SW_HTML)
    strcpy_output_record(level + 2, (unsigned char *) "<");
  if (sw_level_tag)
  {
    (void) sprintf(temp_array, "%s%d", DEFAULT_STR_TAG_LEVEL, level);
    strcat_output_record((unsigned char *) temp_array);
  }
  /* print tag name */
  if (t_encoding == UTF_8)
    strncat_output_record(t, t_length);
  else
  {
    for (i = 0 ; i < t_length ; i++)
      output_record.line[output_record.length++] = t[i];
  }
  if (n > 0)
  {
    if (sw_numbered_tag)
    {
      (void) sprintf(temp_array, "%ld", n);
      strcat_output_record((unsigned char *) temp_array);
    }
    if (sw_numbered_attr)
    {
      (void) sprintf(temp_array, " %s=\"%ld\"", DEFAULT_STR_ATTRIBUTE_NUMBER, n);
      strcat_output_record((unsigned char *) temp_array);
    }
  }
  if (sw_level_attr)
  {
    (void) sprintf(temp_array, " %s=\"%d\"", DEFAULT_STR_ATTRIBUTE_LEVEL, level);
    strcat_output_record((unsigned char *) temp_array);
  }
  if (sw_delimiter_attr)
  {
    if (d >= 0)
    {
      (void) sprintf(temp_array, " %s=\"", DEFAULT_STR_ATTRIBUTE_DELIMITER);
      strcat_output_record((unsigned char *) temp_array);
      append_escaped_string(delimiter[level_d]->arg_d[d], delimiter[level_d]->arg_length[d]);
      strcat_output_record((unsigned char *) "\"");
    }
    else if (d == DELIMITER_INDEX_WHITESPACE)
    {
      (void) sprintf(temp_array, " %s=\"WHITESPACE\"", DEFAULT_STR_ATTRIBUTE_DELIMITER);
      strcat_output_record((unsigned char *) temp_array);
    }
    else if (d == DELIMITER_INDEX_NEWLINE)
    {
      (void) sprintf(temp_array, " %s=\"NEWLINE\"", DEFAULT_STR_ATTRIBUTE_DELIMITER);
      strcat_output_record((unsigned char *) temp_array);
    }
    else if (d == DELIMITER_INDEX_COLUMN)
    {
      if (column[level_d].min > -1)
        (void) sprintf(temp_array, " %s=\"%ld", DEFAULT_STR_ATTRIBUTE_DELIMITER, column[level_d].min + 1);
      else
        (void) sprintf(temp_array, " %s=\"LAST", DEFAULT_STR_ATTRIBUTE_DELIMITER);
      strcat_output_record((unsigned char *) temp_array);
      if (column[level_d].max > -1)
        (void) sprintf(temp_array, "-%ld\"", column[level_d].max + 1);
      else
        (void) sprintf(temp_array, "-LAST\"");
      strcat_output_record((unsigned char *) temp_array);
    }
    else if (d == DELIMITER_INDEX_REGEX)
    {
      (void) sprintf(temp_array, " %s=\"", DEFAULT_STR_ATTRIBUTE_DELIMITER);
      strcat_output_record((unsigned char *) temp_array);
      append_escaped_string(column[level_d].pattern, column[level_d].pattern_length);
      strcat_output_record((unsigned char *) "\"");
    }
    else
    {
      (void) sprintf(temp_array, " %s=\"\"", DEFAULT_STR_ATTRIBUTE_DELIMITER);
      strcat_output_record((unsigned char *) temp_array);
    }
  }
  if (sw_timestamp_attr)
  {
    time(&ts);

    (void) sprintf(temp_array, " %s=\"", DEFAULT_STR_ATTRIBUTE_TIMESTAMP_DATE);
    strcat_output_record((unsigned char *) temp_array);
    make_timestamp_array(ts, TIMESTAMP_DATE);
    strcat_output_record((unsigned char *) temp_array);
    strcat_output_record((unsigned char *) "\"");

    (void) sprintf(temp_array, " %s=\"", DEFAULT_STR_ATTRIBUTE_TIMESTAMP_SECS);
    strcat_output_record((unsigned char *) temp_array);
    make_timestamp_array(ts, TIMESTAMP_SECS);
    strcat_output_record((unsigned char *) temp_array);
    strcat_output_record((unsigned char *) "\"");
  }
  if (sw_custom_attr)
  {
    print_insert_attr(level);
  }
  strcat_output_record((unsigned char *) ">");

  /* ============================================================
   * NOTE: from this point do not use clib routines which don't 
   * support UTF-* on the output_record.line array directly.
   * ============================================================ */

  /* field */
  for (i = 0 ; i < f_length ; i += f_charwidth)
  {
    /* XML characters that need escaping */
    if (utf_char_equals_2(i, '<', f, f_length, f_encoding) && sw_char_escape_lt)
      strcat_output_record((unsigned char *) "&lt;");
    else if (utf_char_equals_2(i, '>', f, f_length, f_encoding) && sw_char_escape_gt)
      strcat_output_record((unsigned char *) "&gt;");
    else if (utf_char_equals_2(i, '&', f, f_length, f_encoding) && sw_char_escape_amp)
      strcat_output_record((unsigned char *) "&amp;");
    else if (utf_char_equals_2(i, '"', f, f_length, f_encoding) && sw_char_escape_quot)
      strcat_output_record((unsigned char *) "&quot;");
    else if (utf_char_equals_2(i, '|', f, f_length, f_encoding) && sw_char_escape_brvbar)
      strcat_output_record((unsigned char *) "&brvbar;");
    else if (utf_char_equals_2(i, '\'', f, f_length, f_encoding) && sw_char_escape_apos)
      strcat_output_record((unsigned char *) "&apos;");
    else
    {
      if (output_record.encoding == UTF_8)
        output_record.line[output_record.length++] = f[i];
      else if (input_record.encoding == UTF_8 || f_encoding == UTF_8) /* and wide encoding */
      {
        if (f[i])
          strncat_output_record(&f[i], 1);
        else
        {
          /* print a null. NOTE: strncat does not handle nulls */
          output_record.line[output_record.length++] = 0;
          output_record.line[output_record.length++] = 0;
          if (output_record.encoding == UTF_32BE || output_record.encoding == UTF_32LE)
          {
            output_record.line[output_record.length++] = 0;
            output_record.line[output_record.length++] = 0;
          }
        }
      }
      else 
      {
        /* Wide encoding (UTF-16 or UTF-32).
         * NOTE: We don't do code point error correction and assume the
         * wide input chars fall within their correct boundaries.
         * Any input stream corruption or purposely formed delimiter
         * strings that match across code point boundaries will 
         * cause the ouput to get out of synchronisation and
         * warrant some investigation by the user. */
        for (j = 0 ; j < output_record.charwidth && i + j < f_length ; j++)
          output_record.line[output_record.length++] = f[i + j];
      }
    }
    /* mem checks */
    if (output_record.length > memseg_output_record)
      memadd_output_record(output_record.length);
  }

  /* close element tag */
  /* mem checks (18 = approx sizeof("</Lnnnnn") + sizeof("nnnnnnnn>\n") */
  if (output_record.length + t_length + (18 * output_record.charwidth) > memseg_output_record)
    memadd_output_record(output_record.length + t_length + (18 * output_record.charwidth));

  strcat_output_record((unsigned char *) "</");
  if (sw_xml_format == SW_SOAP)
  {
    strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_XMLFY);
    strcat_output_record((unsigned char *) ":");
  }
  if (sw_level_tag)
  {
    (void) sprintf(temp_array, "%s%d", DEFAULT_STR_TAG_LEVEL, level);
    strcat_output_record((unsigned char *) temp_array);
  }
  /* neaten close tag name of any attributes etc */
  if (sw_trimtagclose)
    t_length = utf_get_first_word_length(t, t_length, t_encoding);
  /* print tag name */
  if (t_encoding == UTF_8)
    strncat_output_record(t, t_length);
  else
  {
    for (i = 0 ; i < t_length ; i++)
      output_record.line[output_record.length++] = t[i];
  }
  if (n > 0)
  {
    if (sw_numbered_tag)
    {
      (void) sprintf(temp_array, "%ld", n);
      strcat_output_record((unsigned char *) temp_array);
    }
  }
  strcat_output_record((unsigned char *) ">");
  strcat_output_record(newline);
}


void mktag(tag_control, level, t, t_length, t_encoding, n, sw_numbered_tag, sw_numbered_attr, level_d, d, sw_delimiter_attr, sw_level_tag, sw_level_attr, sw_timestamp_attr, sw_custom_attr, sw_custom_tag)
int tag_control;
int level;
unsigned char *t;
int t_length;
int t_encoding;
long n;
int sw_numbered_tag;
int sw_numbered_attr;
int level_d;
int d;
int sw_delimiter_attr;
int sw_level_tag;
int sw_level_attr;
int sw_timestamp_attr;
int sw_custom_attr;
int sw_custom_tag;
{
  /* Construct an record tag array suitable for XML and UTF-8 printing */
  /* Array output_record.line and output_record.length are updated */
  long i;

  /* print re-insert tag */
  if (sw_custom_tag)
    print_insert_tag(level, TRUE);

  /* padding */
  if (sw_xml_format == SW_XML)
    strcpy_output_record(level, (unsigned char *) "<");
  else /* SW_SOAP and SW_HTML */
    strcpy_output_record(level + 2, (unsigned char *) "<");

  /* Not yet supported
   * if (tag_control == TAG_START || tag_control == TAG_EMPTY) */
  if (tag_control == TAG_START)
  {
    /* open record tag */
    if (sw_xml_format == SW_SOAP)
    {
      strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_XMLFY);
      strcat_output_record((unsigned char *) ":");
    }
    if (sw_level_tag)
    {
      (void) sprintf(temp_array, "%s%d", DEFAULT_STR_TAG_LEVEL, level);
      strcat_output_record((unsigned char *) temp_array);
    }
    /* print tag name */
    if (t_encoding == UTF_8)
      strncat_output_record(t, t_length);
    else
    {
      for (i = 0 ; i < t_length ; i++)
        output_record.line[output_record.length++] = t[i];
    }
    if (n > 0)
    {
      if (sw_numbered_tag)
      {
        (void) sprintf(temp_array, "%ld", n);
        strcat_output_record((unsigned char *) temp_array);
      }
      if (sw_numbered_attr)
      {
        (void) sprintf(temp_array, " %s=\"%ld\"", DEFAULT_STR_ATTRIBUTE_NUMBER, n);
        strcat_output_record((unsigned char *) temp_array);
      }
    }
    if (sw_level_attr)
    {
      (void) sprintf(temp_array, " %s=\"%d\"", DEFAULT_STR_ATTRIBUTE_LEVEL, level);
      strcat_output_record((unsigned char *) temp_array);
    }
    if (sw_delimiter_attr)
    {
      if (d >= 0)
      {
        (void) sprintf(temp_array, " %s=\"", DEFAULT_STR_ATTRIBUTE_DELIMITER);
        strcat_output_record((unsigned char *) temp_array);
        append_escaped_string(delimiter[level_d]->arg_d[d], delimiter[level_d]->arg_length[d]);
        strcat_output_record((unsigned char *) "\"");
      }
      else if (d == DELIMITER_INDEX_WHITESPACE)
      {
        (void) sprintf(temp_array, " %s=\"WHITESPACE\"", DEFAULT_STR_ATTRIBUTE_DELIMITER);
        strcat_output_record((unsigned char *) temp_array);
      }
      else if (d == DELIMITER_INDEX_NEWLINE)
      {
        (void) sprintf(temp_array, " %s=\"NEWLINE\"", DEFAULT_STR_ATTRIBUTE_DELIMITER);
        strcat_output_record((unsigned char *) temp_array);
      }
      else
      {
        (void) sprintf(temp_array, " %s=\"\"", DEFAULT_STR_ATTRIBUTE_DELIMITER);
        strcat_output_record((unsigned char *) temp_array);
      }
    }
    if (sw_timestamp_attr)
    {
      time(&ts);
      (void) sprintf(temp_array, " %s=\"", DEFAULT_STR_ATTRIBUTE_TIMESTAMP_DATE);
      strcat_output_record((unsigned char *) temp_array);
      make_timestamp_array(ts, TIMESTAMP_DATE);
      strcat_output_record((unsigned char *) temp_array);
      strcat_output_record((unsigned char *) "\"");

      (void) sprintf(temp_array, " %s=\"", DEFAULT_STR_ATTRIBUTE_TIMESTAMP_SECS);
      strcat_output_record((unsigned char *) temp_array);
      make_timestamp_array(ts, TIMESTAMP_SECS);
      strcat_output_record((unsigned char *) temp_array);
      strcat_output_record((unsigned char *) "\"");
    }
    if (sw_custom_attr)
    {
      print_insert_attr(level);
    }
  }
  else 
  {
    /* close record tag */
    strcat_output_record((unsigned char *) "/");
    if (sw_xml_format == SW_SOAP)
    {
      strcat_output_record((unsigned char *) DEFAULT_STR_SOAP_XS_XMLFY);
      strcat_output_record((unsigned char *) ":");
    }
    if (sw_level_tag)
    {
      (void) sprintf(temp_array, "%s%d", DEFAULT_STR_TAG_LEVEL, level);
      strcat_output_record((unsigned char *) temp_array);
    }
    /* neaten close tag name of any attributes etc */
    if (sw_trimtagclose)
      t_length = utf_get_first_word_length(t, t_length, t_encoding);
    /* print tag name */
    if (t_encoding == UTF_8)
      strncat_output_record(t, t_length);
    else
    {
      for (i = 0 ; i < t_length ; i++)
        output_record.line[output_record.length++] = t[i];
    }
    if (n > 0)
    {
      if (sw_numbered_tag)
      {
        (void) sprintf(temp_array, "%ld", n);
        strcat_output_record((unsigned char *) temp_array);
      }
    }
  }
  /* 
   * Not yet supported
   * if (tag_control == TAG_EMPTY)
   *   strcat_output_record((unsigned char *) " /");
   */
  strcat_output_record((unsigned char *) ">");
  strcat_output_record(newline);
}


void make_timestamp_array(t, type)
time_t t;
int type;
{
  /* put a timestamp into an array */
  int i;

  if (type == TIMESTAMP_SECS)
    (void) sprintf(temp_array, "%ld", t);
  else
  {
    (void) sprintf(temp_array, "%s", ctime(&t));
    /* strip new-line character from returned string */
    for (i = 0 ; i < strlen(temp_array) ; i++) 
      if (temp_array[i] == '\n') temp_array[i] = '\0';
  }
}


void append_escaped_string(s, s_length)
unsigned char *s;
int s_length;
{
  /* append arg_delimiter string to output_record.line array and 
   * ensure printed delimiter doesn't corrupt XML output */
  int i;
  char hex_array[4];

  for (i = 0 ; i < s_length ; i++)
  {
    /* only alphanum characters allowed */
    if (s[i] == '\n')
      strcat_output_record((unsigned char *) "\\n");
    else if (s[i] == '\r')
      strcat_output_record((unsigned char *) "\\r");
    else if (s[i] == '\t')
      strcat_output_record((unsigned char *) "\\t");
    else if (s[i] == '\"' ||
             s[i] == '\'' ||
             s[i] == '`' ||
             s[i] == '<' ||
             s[i] == '>' ||
             s[i] == '&' ||
             s[i] == '|')
    {
      /* print HEX equivalent in URI style */
      (void) sprintf(hex_array, "%%%2.2x", s[i]);
      hex_array[1] = toupper((int) hex_array[1]);
      hex_array[2] = toupper((int) hex_array[2]);
      strcat_output_record((unsigned char *) hex_array);
    }
    else if (isprint(s[i]))
      strncat_output_record((unsigned char *) &s[i], 1);
    else
    {
      /* print HEX equivalent in URI style */
      (void) sprintf(hex_array, "%%%2.2x", s[i]);
      hex_array[1] = toupper((int) hex_array[1]);
      hex_array[2] = toupper((int) hex_array[2]);
      strcat_output_record((unsigned char *) hex_array);
    }
  }
}


void format_key_value_tagname(i)
long i;
{
  /* neaten up what will become the new tag name */
  long j;

  /* re-trim field in case it was derived from a quoted field */
  trim_field(i);
  /* replace the whitespace between words with an underscore */
  for (j = input_record.field[i]->start ; j < input_record.field[i]->start + input_record.field[i]->length ; j += input_record.charwidth)
  {
    if (utf_isspace(j) && j < input_record.field[i]->start + input_record.field[i]->length)
    {
      if (input_record.encoding == UTF_8)
        input_record.line[j] = '_';
      else if (input_record.encoding == UTF_16BE)
      {
        input_record.line[j]     = 0;
        input_record.line[j + 1] = '_';
      }
      else if (input_record.encoding == UTF_16LE)
      {
        input_record.line[j]     = '_';
        input_record.line[j + 1] = 0;
      }
      else if (input_record.encoding == UTF_32BE)
      {
        input_record.line[j]     = 0;
        input_record.line[j + 1] = 0;
        input_record.line[j + 2] = 0;
        input_record.line[j + 3] = '_';
      }
      else if (input_record.encoding == UTF_32LE)
      {
        input_record.line[j]     = '_';
        input_record.line[j + 1] = 0;
        input_record.line[j + 2] = 0;
        input_record.line[j + 3] = 0;
      }
    }
  }
}


void print_insert_attr(level)
int level;
{
  /* iterate through insert_attr structs and print out custom attributes */
  int i, j;
  int flag_print_attr;

  for (i = 0 ; i < ninsert_attr ; i++)
  {
    if ((insert_attr[i].level < 0 && level > 0) || insert_attr[i].level == level)
    {
      flag_print_attr = TRUE;

      /* process scope */
      if (insert_attr[i].nscope)
      {
        /* check for scoping restrictions */
        flag_print_attr = FALSE;
        for (j = 0 ; j < insert_attr[i].nscope ; j++)
        {
          if (insert_attr[i].scope_seq >= insert_attr[i].scope[j].min 
             && (insert_attr[i].scope[j].max < 0 || insert_attr[i].scope_seq <= insert_attr[i].scope[j].max))
          {
            flag_print_attr = TRUE;
            break; /* break for/j */
          }
        } /* end for/j */

        /* increment scope sequence */
        insert_attr[i].scope_seq++;

        /* check for retarting scope and restart */
        if (flag_print_attr && insert_attr[i].scope[j].restart && insert_attr[i].scope[j].max >= 0 && insert_attr[i].scope[j].max == insert_attr[i].scope_seq - 1)
        {
          insert_attr[i].scope_seq = 0;
          for (j = 0 ; j < ninsert_attr ; j++)
          {
            if (insert_attr[j].level > level)
              insert_attr[j].scope_seq = 0;
          } /* end for/j */
        }
      }

      /* print custom attribute */
      if (flag_print_attr)
      {
        strcat_output_record((unsigned char *) " ");
        strncat_output_record(insert_attr[i].name, insert_attr[i].name_length);
        strcat_output_record((unsigned char *) "=\"");
        strncat_output_record(insert_attr[i].value, insert_attr[i].value_length);
        strcat_output_record((unsigned char *) "\"");
      }
    }
    else if (insert_attr[i].nscope && insert_attr[i].level > level)
    {
      /* reset scope sequence of higher levels */
      insert_attr[i].scope_seq = 0;
    }
  } /* end for/i */
}


void print_insert_tag(level, flag_reinsert)
int level;
int flag_reinsert;
{
  /* iterate through insert_tag structs and print out custom tags */
  int i, j;
  int flag_print_tag;

  for (i = 0 ; i < ninsert_tag ; i++)
  {
    if (insert_tag[i].reinsert == flag_reinsert)
    {
      if ((insert_tag[i].level < 0 && level > 0) || insert_tag[i].level == level)
      {
        flag_print_tag = TRUE;

        /* process scope */
        if (insert_tag[i].nscope)
        {
          /* check for scoping restrictions */
          flag_print_tag = FALSE;
          for (j = 0 ; j < insert_tag[i].nscope ; j++)
          {
            if (insert_tag[i].scope_seq >= insert_tag[i].scope[j].min 
               && (insert_tag[i].scope[j].max < 0 || insert_tag[i].scope_seq <= insert_tag[i].scope[j].max))
            {
              flag_print_tag = TRUE;
              break; /* break for/j */
            }
          } /* end for/j */

          /* increment scope sequence */
          insert_tag[i].scope_seq++;

          /* check for retarting scope and restart */
          if (flag_print_tag && insert_tag[i].scope[j].restart && insert_tag[i].scope[j].max >= 0 && insert_tag[i].scope[j].max == insert_tag[i].scope_seq - 1)
          {
            insert_tag[i].scope_seq = 0;
            for (j = 0 ; j < ninsert_tag ; j++)
              if (insert_tag[j].type == insert_tag[i].type && insert_tag[j].level > level)
                insert_tag[j].scope_seq = 0;
          }
        }

        /* print custom tag */
        if (flag_print_tag)
        {
          if (insert_tag[i].type == TAG_INSERT_ELEMENT)
          {
            /* no filtering for illegal XML */
            if (sw_xml_format == SW_XML)
              strcpy_output_record(level, (unsigned char *) "<");
            else /* SW_SOAP and SW_HTML */
              strcpy_output_record(level + 2, (unsigned char *) "<");
            strcat_output_record(insert_tag[i].name);
            strcat_output_record((unsigned char *) ">");
            strcat_output_record(insert_tag[i].value);
            strcat_output_record((unsigned char *) "</");
            strcat_output_record(insert_tag[i].name);
            strcat_output_record((unsigned char *) ">");
            strcat_output_record(newline);
            utf_print(STDOUT_FD, output_record.line, output_record.length);
          }
          else if (insert_tag[i].type == TAG_INSERT_FILE)
          {
            /* we filter file contents for illegal XML */
            mkelem(level, insert_tag[i].name, insert_tag[i].name_length, UTF_8, insert_tag[i].value, insert_tag[i].value_length, UTF_8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
            utf_print(STDOUT_FD, output_record.line, output_record.length);
          }
          else if (insert_tag[i].type == TAG_INSERT_XMLFILE)
          {
            /* no filtering for illegal XML */
            utf_print(STDOUT_FD, insert_tag[i].value, insert_tag[i].value_length);
          }
        }
      } 
      else if (insert_tag[i].reinsert && insert_tag[i].nscope && insert_tag[i].level > level)
      {
        /* reset scope sequence of higher levels but only for reinsertable tags */
        insert_tag[i].scope_seq = 0;
      }
    }
  } /* end for/i */
}
