/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2013, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert to XML on the fly.                                     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "xmlout.h"

void print_output_summary(argc, argv)
int argc;
unsigned char **argv;
{
  /* produce a summary tag */
  long i, j, k, record_cntr_processed;
  int parent_name_length;
  unsigned char digest[64];

  record_cntr_processed = 0;

  /* Stop the output digest from processing this summary.
   * Normally we would finalise all the digests here but that technique requires
   * an array for each digest result, so instead we use a global switch to 
   * stop output digest processing and save a bit of memory. */
  sw_process_output_digest = FALSE;
  
  /* ensure no user defined anomalies in parent tag name */
  parent_name_length = (int) utf_get_first_word_length(schema_parent_name, xstrlen(schema_parent_name), UTF_8);
   
  if (sw_summary == SUMMARY_FILE)
  {
    /* print a summary prologue */
    if (sw_bom)
      utf_print_bom(summary_fd);

    strcpy_output_record(0, (unsigned char *) "<?xml version=\"");
    if (version_xml == XML_1_0)
      strcat_output_record((unsigned char *) "1.0");
    else if (version_xml == XML_1_1)
      strcat_output_record((unsigned char *) "1.1");

    strcat_output_record((unsigned char *) "\" encoding=\"");
    strcat_output_record((unsigned char *) output_record.encoding_name);
    strcat_output_record((unsigned char *) "\"?>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
    strcpy_output_record(0, (unsigned char *) "<!DOCTYPE ");
    strncat_output_record(schema_parent_name, parent_name_length);
    strcat_output_record((unsigned char *) "_summary SYSTEM \"xmlfy.dtd\">");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
  }
  else if (sw_summary == SUMMARY_COMMENT)
  {
    /* open comment token */
    strcpy_output_record(0, (unsigned char *) "<!-- start-of-summary-comment");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
  }

  /* open summary tag */
  strcpy_output_record(0, (unsigned char *) "<");
  strncat_output_record(schema_parent_name, parent_name_length);
  strcat_output_record((unsigned char *) "_summary>");
  strcat_output_record(newline);
  utf_print(summary_fd, output_record.line, output_record.length);
  /* version */
  strcpy_output_record(1, (unsigned char *) "<version>");
  strcat_output_record((unsigned char *) VERSION);
  strcat_output_record((unsigned char *) "</version>");
  strcat_output_record(newline);
  utf_print(summary_fd, output_record.line, output_record.length);
  /* timestamp */
  strcpy_output_record(1, (unsigned char *) "<timestamp>");
  strcat_output_record(newline);
  utf_print(summary_fd, output_record.line, output_record.length);
  make_timestamp_array(start_ts, TIMESTAMP_DATE);
  mkelem(2, (unsigned char *) "start", strlen("start"), UTF_8, (unsigned char *) temp_array, strlen(temp_array), UTF_8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
  utf_print(summary_fd, output_record.line, output_record.length);
  time(&ts);
  make_timestamp_array(ts, TIMESTAMP_DATE);
  mkelem(2, (unsigned char *) "end", strlen("end"), UTF_8, (unsigned char *) temp_array, strlen(temp_array), UTF_8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
  utf_print(summary_fd, output_record.line, output_record.length);
  strcpy_output_record(1, (unsigned char *) "</timestamp>");
  strcat_output_record(newline);
  utf_print(summary_fd, output_record.line, output_record.length);
  if (sw_digest)
  {
    /* finalise input MD5 */
    MD5Final(digest, &md5_in_context);
    /* md5_input signature tag */
    strcpy_output_record(1, (unsigned char *) "<md5_input>");
    if (sw_printonly_records) /* MD5 input digest is only accurately calculated with record processing */
    {
      for (i = 0 ; i < 16 ; i++)
      {
        (void) sprintf(temp_array, "%02x", digest[i]);
        strcat_output_record((unsigned char *) temp_array);
      }
    }
    strcat_output_record((unsigned char *) "</md5_input>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
    /* finalise output MD5 */
    MD5Final(digest, &md5_out_context);
    /* md5_output signature tag */
    strcpy_output_record(1, (unsigned char *) "<md5_output>");
    for (i = 0 ; i < 16 ; i++)
    {
      (void) sprintf(temp_array, "%02x", digest[i]);
      strcat_output_record((unsigned char *) temp_array);
    }
    strcat_output_record((unsigned char *) "</md5_output>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
    /* finalise input SHA512 */
    j = SHA512Result(&sha512_in_context, digest);
    /* sha512_input signature tag */
    strcpy_output_record(1, (unsigned char *) "<sha512_input>");
    if (sw_printonly_records) /* SHA512 input digest is only accurately calculated with record processing */
    {
      if (j == 0)
      {
        for (i = 0 ; i < 64 ; i++)
        {
          (void) sprintf(temp_array, "%02x", digest[i]);
          strcat_output_record((unsigned char *) temp_array);
        }
      }
      else
        strcat_output_record((unsigned char *) "fail");
    }
    strcat_output_record((unsigned char *) "</sha512_input>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
    /* finalise output SHA512 */
    j = SHA512Result(&sha512_out_context, digest);
    /* sha512_output signature tag */
    strcpy_output_record(1, (unsigned char *) "<sha512_output>");
    if (j == 0)
    {
      for (i = 0 ; i < 64 ; i++)
      {
        (void) sprintf(temp_array, "%02x", digest[i]);
        strcat_output_record((unsigned char *) temp_array);
      }
    }
    else
      strcat_output_record((unsigned char *) "fail");
    strcat_output_record((unsigned char *) "</sha512_output>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
  }
  /* open args tag */
  strcpy_output_record(1, (unsigned char *) "<args>");
  strcat_output_record(newline);
  utf_print(summary_fd, output_record.line, output_record.length);
  strcpy_output_record(2, (unsigned char *) "<command>");
  strcat_output_record(newline);
  utf_print(summary_fd, output_record.line, output_record.length);
  /* command args */
  for (i = 0 ; i < argc ; i++)
  {
    mkelem(3, (unsigned char *) "arg", strlen("arg"), UTF_8, argv[i], xstrlen(argv[i]), UTF_8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
    utf_print(summary_fd, output_record.line, output_record.length);
  }
  strcpy_output_record(2, (unsigned char *) "</command>");
  strcat_output_record(newline);
  utf_print(summary_fd, output_record.line, output_record.length);
  if (sw_schema != SCHEMA_DEFAULT)
  {
    strcpy_output_record(2, (unsigned char *) "<schema>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
    /* schema args */
    for (i = 1 ; i < schema_argc ; i++)
    {
      mkelem(3, (unsigned char *) "arg", strlen("arg"), UTF_8, schema_argv[i], xstrlen(schema_argv[i]), UTF_8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
      utf_print(summary_fd, output_record.line, output_record.length);
    }
    strcpy_output_record(2, (unsigned char *) "</schema>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
  }
  /* close args tag */
  strcpy_output_record(1, (unsigned char *) "</args>");
  strcat_output_record(newline);
  utf_print(summary_fd, output_record.line, output_record.length);
  /* open separator tokens tag */
  strcpy_output_record(1, (unsigned char *) "<separator>");
  strcat_output_record(newline);
  utf_print(summary_fd, output_record.line, output_record.length);
  /* delimiter tokens */
  for (i = 1 ; i < nlevels ; i++)
  {
    if (i == 1 && delimiter[i]->ndelimiters == 0)
    {
      (void) sprintf(temp_array, "<level%ld>%s", i, newline);
      strcpy_output_record(2, (unsigned char *) temp_array);
      strcat_output_record((unsigned char *) "      <delimiter>NEWLINE</delimiter>");
      strcat_output_record(newline);
      (void) sprintf(temp_array, "    </level%ld>%s", i, newline);
      strcat_output_record((unsigned char *) temp_array);
      utf_print(summary_fd, output_record.line, output_record.length);
    }
    else if (i == 2 && delimiter[i]->ndelimiters == 0)
    {
      (void) sprintf(temp_array, "<level%ld>%s", i, newline);
      strcpy_output_record(2, (unsigned char *) temp_array);
      strcat_output_record((unsigned char *) "      <delimiter>WHITESPACE</delimiter>");
      strcat_output_record(newline);
      (void) sprintf(temp_array, "    </level%ld>%s", i, newline);
      strcat_output_record((unsigned char *) temp_array);
      utf_print(summary_fd, output_record.line, output_record.length);
    }
    else if (delimiter[i]->ndelimiters > 0)
    {
      (void) sprintf(temp_array, "<level%ld>%s", i, newline);
      strcpy_output_record(2, (unsigned char *) temp_array);
      utf_print(summary_fd, output_record.line, output_record.length);

      for (j = 0 ; j < delimiter[i]->ndelimiters ; j++)
      {
        strcpy_output_record(3, (unsigned char *) "<delimiter");
        if (delimiter[i]->nscope[j])
        {
          strcat_output_record((unsigned char *) " scope=\"");
          for (k = 0 ; k < delimiter[i]->nscope[j] ; k++)
          {
            if (k)
              strcat_output_record((unsigned char *) ",");
            if (delimiter[i]->scope[j][k].min == delimiter[i]->scope[j][k].max)
              (void) sprintf(temp_array, "%ld", delimiter[i]->scope[j][k].min + 1);
            else if (delimiter[i]->scope[j][k].max < 0)
              (void) sprintf(temp_array, "%ld-MAX", delimiter[i]->scope[j][k].min + 1);
            else
              (void) sprintf(temp_array, "%ld-%ld", delimiter[i]->scope[j][k].min + 1, delimiter[i]->scope[j][k].max + 1);
            strcat_output_record((unsigned char *) temp_array);
            if (delimiter[i]->scope[j][k].restart)
              strcat_output_record((unsigned char *) "r");
          }
          strcat_output_record((unsigned char *) "\"");
        }
        strcat_output_record((unsigned char *) ">");
        append_escaped_string(delimiter[i]->arg_d[j], delimiter[i]->arg_length[j]);
        strcat_output_record((unsigned char *) "</delimiter>");
        strcat_output_record(newline);
        utf_print(summary_fd, output_record.line, output_record.length);
      }

      (void) sprintf(temp_array, "</level%ld>%s", i, newline);
      strcpy_output_record(2, (unsigned char *) temp_array);
      utf_print(summary_fd, output_record.line, output_record.length);
    }
  }
  /* close separator tokens tag */
  strcpy_output_record(1, (unsigned char *) "</separator>");
  strcat_output_record(newline);
  utf_print(summary_fd, output_record.line, output_record.length);
  /* column option */
  if (ncolumn)
  {
    /* open column tokens tag */
    strcpy_output_record(1, (unsigned char *) "<columns>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
    k = 1;
    for (i = 0 ; i < ncolumn ; i++)
    {
      /* open field range tokens tag */
      (void) sprintf(temp_array, "<field%ld", k);
      strcpy_output_record(2, (unsigned char *) temp_array);
      if (column[i].nscope)
      {
        strcat_output_record((unsigned char *) " scope=\"");
        for (j = 0 ; j < column[i].nscope ; j++)
        {
          if (j)
            strcat_output_record((unsigned char *) ",");
          if (column[i].scope[j].min == column[i].scope[j].max)
            (void) sprintf(temp_array, "%ld", column[i].scope[j].min + 1);
          else if (column[i].scope[j].max < 0)
            (void) sprintf(temp_array, "%ld-MAX", column[i].scope[j].min + 1);
          else
            (void) sprintf(temp_array, "%ld-%ld", column[i].scope[j].min + 1, column[i].scope[j].max + 1);
          strcat_output_record((unsigned char *) temp_array);
          if (column[i].scope[j].restart)
            strcat_output_record((unsigned char *) "r");
        }
        strcat_output_record((unsigned char *) "\"");
      }
      strcat_output_record((unsigned char *) ">");
      strcat_output_record(newline);
      utf_print(summary_fd, output_record.line, output_record.length);
      if (column[i].type == OPT_COLUMN)
      {
        /* column type tag */
        strcpy_output_record(3, (unsigned char *) "<type>range</type>");
        strcat_output_record(newline);
        utf_print(summary_fd, output_record.line, output_record.length);
        /* open/close min tokens tag */
        if (column[i].min < 0)
          (void) sprintf(temp_array, "<min>LAST</min>%s", newline);
        else
          (void) sprintf(temp_array, "<min>%ld</min>%s", column[i].min + 1, newline);
        strcpy_output_record(3, (unsigned char *) temp_array);
        utf_print(summary_fd, output_record.line, output_record.length);
        /* open/close max tokens tag */
        if (column[i].max < 0)
          (void) sprintf(temp_array, "<max>LAST</max>%s", newline);
        else
          (void) sprintf(temp_array, "<max>%ld</max>%s", column[i].max + 1, newline);
        strcpy_output_record(3, (unsigned char *) temp_array);
        utf_print(summary_fd, output_record.line, output_record.length);
        /* open/close name tokens tag */
        if (column[i].name == NULL)
        {
          strcpy_output_record(3, (unsigned char *) "<name>DEFAULT</name>");
          strcat_output_record(newline);
        }
        else
          mkelem(3, (unsigned char *) "name", strlen("name"), UTF_8, column[i].name, xstrlen(column[i].name), UTF_8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
        utf_print(summary_fd, output_record.line, output_record.length);
      }
      else if (column[i].type == OPT_REGEX)
      {
        /* column type tag */
        strcpy_output_record(3, (unsigned char *) "<type>regex</type>");
        strcat_output_record(newline);
        utf_print(summary_fd, output_record.line, output_record.length);
        mkelem(3, (unsigned char *) "pattern", strlen("pattern"), UTF_8, column[i].pattern, column[i].pattern_length, UTF_8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
        utf_print(summary_fd, output_record.line, output_record.length);
        (void) sprintf(temp_array, "<subfields>%d</subfields>%s", (int) column[i].pcomp.re_nsub, newline);
        strcpy_output_record(3, (unsigned char *) temp_array);
        utf_print(summary_fd, output_record.line, output_record.length);
        /* regex names */
        for (j = 0 ; j < column[i].ntagname ; j++)
        {
          (void) sprintf(temp_array, "name%d", (int) j + 1);
          mkelem(3, (unsigned char *) temp_array, strlen(temp_array), UTF_8, column[i].tagname[j], xstrlen(column[i].tagname[j]), UTF_8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
          utf_print(summary_fd, output_record.line, output_record.length);
        }
      }
      /* close field range tokens tag */
      (void) sprintf(temp_array, "</field%ld>%s", k, newline);
      strcpy_output_record(2, (unsigned char *) temp_array);
      utf_print(summary_fd, output_record.line, output_record.length);
      if (column[i].type == OPT_REGEX && column[i].pcomp.re_nsub > 0)
        k += column[i].pcomp.re_nsub;
      else
        k++;
    }
    /* close column tokens tag */
    strcpy_output_record(1, (unsigned char *) "</columns>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
  }
  /* quotechars option */
  if (quotechars_length)
  {
    /* open quotechars tokens tag */
    if (quotechar_type == 2)
      strcpy_output_record(1, (unsigned char *) "<quotechars match=\"paired\">");
    else
      strcpy_output_record(1, (unsigned char *) "<quotechars match=\"any\">");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
    /* quotechars tokens */
    for (i = 0 ; i < quotechars_length ; i++)
    {
      mkelem(2, (unsigned char *) "char", strlen("char"), UTF_8, &quotechars[i], 1, UTF_8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
      utf_print(summary_fd, output_record.line, output_record.length);
    }
    /* close quotechars tokens tag */
    strcpy_output_record(1, (unsigned char *) "</quotechars>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
  }
  /* expel option */
  if (nexpel)
  {
    /* open expel tokens tag */
    strcpy_output_record(1, (unsigned char *) "<expel>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
    for (i = 0 ; i < nexpel ; i++)
    {
      /* open expel group tokens tag */
      (void) sprintf(temp_array, "<group%ld>%s", i + 1, newline);
      strcpy_output_record(2, (unsigned char *) temp_array);
      utf_print(summary_fd, output_record.line, output_record.length);
      /* open expel records tokens tag */
      strcpy_output_record(3, (unsigned char *) "<input_records>");
      strcat_output_record(newline);
      utf_print(summary_fd, output_record.line, output_record.length);
      for (j = 0 ; j < expel[i].n_erecords ; j++)
      {
        /* open range tokens tag */
        strcpy_output_record(4, (unsigned char *) "<range>");
        strcat_output_record(newline);
        utf_print(summary_fd, output_record.line, output_record.length);
        /* open/close type tokens tag */
        switch(expel[i].erecords[j].type)
        {
          case EXPEL_TYPE_N:
             strcpy_output_record(5, (unsigned char *) "<type>n</type>");
             strcat_output_record(newline);
             break;
  
          case EXPEL_TYPE_F:
             strcpy_output_record(5, (unsigned char *) "<type>f</type>");
             strcat_output_record(newline);
             break;

          case EXPEL_TYPE_C:
             strcpy_output_record(5, (unsigned char *) "<type>c</type>");
             strcat_output_record(newline);
             break;
  
          default:
             (void) sprintf(temp_array, "<type>%d</type>%s", expel[i].erecords[j].type, newline);
             strcpy_output_record(5, (unsigned char *) temp_array);
        }
        utf_print(summary_fd, output_record.line, output_record.length);
        /* open/close min tokens tag */
        (void) sprintf(temp_array, "<min>%ld</min>%s", expel[i].erecords[j].min + 1, newline);
        strcpy_output_record(5, (unsigned char *) temp_array);
        utf_print(summary_fd, output_record.line, output_record.length);
        /* open/close max tokens tag */
        if (expel[i].erecords[j].max < 0)
        {
          strcpy_output_record(5, (unsigned char *) "<max>LAST</max>");
          strcat_output_record(newline);
        }
        else
        {
          (void) sprintf(temp_array, "<max>%ld</max>%s", expel[i].erecords[j].max + 1, newline);
          strcpy_output_record(5, (unsigned char *) temp_array);
        }
        utf_print(summary_fd, output_record.line, output_record.length);
        /* open/close search string tokens tag */
        if (expel[i].erecords[j].s_length)
        {
          mkelem(5, (unsigned char *) "search", strlen("search"), UTF_8, expel[i].erecords[j].s, expel[i].erecords[j].s_length, UTF_8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
          utf_print(summary_fd, output_record.line, output_record.length);
        }
        /* close range tokens tag */
        strcpy_output_record(4, (unsigned char *) "</range>");
        strcat_output_record(newline);
        utf_print(summary_fd, output_record.line, output_record.length);
      }
      /* close expel records tokens tag */
      strcpy_output_record(3, (unsigned char *) "</input_records>");
      strcat_output_record(newline);
      utf_print(summary_fd, output_record.line, output_record.length);
      /* open expel fields tokens tag */
      strcpy_output_record(3, (unsigned char *) "<input_fields>");
      strcat_output_record(newline);
      utf_print(summary_fd, output_record.line, output_record.length);
      if (expel[i].n_efields)
      {
        for (j = 0 ; j < expel[i].n_efields ; j++)
        {
          /* open range tokens tag */
          strcpy_output_record(4, (unsigned char *) "<range>");
          strcat_output_record(newline);
          utf_print(summary_fd, output_record.line, output_record.length);
          /* open/close min tokens tag */
          if (expel[i].efields[j].min < 0)
          {
            strcpy_output_record(5, (unsigned char *) "<min>LAST</min>");
            strcat_output_record(newline);
          }
          else
          {
            (void) sprintf(temp_array, "<min>%ld</min>%s", expel[i].efields[j].min + 1, newline);
            strcpy_output_record(5, (unsigned char *) temp_array);
          }
          utf_print(summary_fd, output_record.line, output_record.length);
          /* open/close max tokens tag */
          if (expel[i].efields[j].max < 0)
          {
            strcpy_output_record(5, (unsigned char *) "<max>LAST</max>");
            strcat_output_record(newline);
          }
          else
          {
            (void) sprintf(temp_array, "<max>%ld</max>%s", expel[i].efields[j].max + 1, newline);
            strcpy_output_record(5, (unsigned char *) temp_array);
          }
          utf_print(summary_fd, output_record.line, output_record.length);
          /* close range tokens tag */
          strcpy_output_record(4, (unsigned char *) "</range>");
          strcat_output_record(newline);
          utf_print(summary_fd, output_record.line, output_record.length);
        }
      }
      else
      {
        /* open range tokens tag */
        strcpy_output_record(4, (unsigned char *) "<range>");
        strcat_output_record(newline);
        utf_print(summary_fd, output_record.line, output_record.length);
        /* open/close min tokens tag */
        strcpy_output_record(5, (unsigned char *) "<min>1</min>");
        strcat_output_record(newline);
        utf_print(summary_fd, output_record.line, output_record.length);
        /* open/close max tokens tag */
        strcpy_output_record(5, (unsigned char *) "<max>LAST</max>");
        strcat_output_record(newline);
        utf_print(summary_fd, output_record.line, output_record.length);
        /* close range tokens tag */
        strcpy_output_record(4, (unsigned char *) "</range>");
        strcat_output_record(newline);
        utf_print(summary_fd, output_record.line, output_record.length);
      }
      /* close expel fields tokens tag */
      strcpy_output_record(3, (unsigned char *) "</input_fields>");
      strcat_output_record(newline);
      utf_print(summary_fd, output_record.line, output_record.length);
      /* close expel group tokens tag */
      (void) sprintf(temp_array, "</group%ld>%s", i + 1, newline);
      strcpy_output_record(2, (unsigned char *) temp_array);
      utf_print(summary_fd, output_record.line, output_record.length);
    }
    /* close expel tokens tag */
    strcpy_output_record(1, (unsigned char *) "</expel>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
  }
  /* custom attribute option */
  if (ninsert_attr)
  {
    /* open insert_attr tokens tag */
    strcpy_output_record(1, (unsigned char *) "<custom_attributes>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
    /* attribute name tokens */
    for (i = 0 ; i < ninsert_attr ; i++)
    {
      /* open attribute tokens tag */
      (void) sprintf(temp_array, "<attribute%ld>%s", i + 1, newline);
      strcpy_output_record(2, (unsigned char *) temp_array);
      utf_print(summary_fd, output_record.line, output_record.length);
      /* open/close level tokens tag */
      if (insert_attr[i].level < 0)
      {
        strcpy_output_record(3, (unsigned char *) "<level>ALL</level>");
        strcat_output_record(newline);
      }
      else
      {
        (void) sprintf(temp_array, "<level>%d</level>%s", insert_attr[i].level, newline);
        strcpy_output_record(3, (unsigned char *) temp_array);
      }
      utf_print(summary_fd, output_record.line, output_record.length);
      /* open/close name tokens tag */
      mkelem(3, (unsigned char *) "name", strlen("name"), UTF_8, insert_attr[i].name, insert_attr[i].name_length, UTF_8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
      utf_print(summary_fd, output_record.line, output_record.length);
      /* open/close value tokens tag */
      mkelem(3, (unsigned char *) "value", strlen("value"), UTF_8, insert_attr[i].value, insert_attr[i].value_length, UTF_8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
      utf_print(summary_fd, output_record.line, output_record.length);
      /* close attribute tokens tag */
      (void) sprintf(temp_array, "</attribute%ld>%s", i + 1, newline);
      strcpy_output_record(2, (unsigned char *) temp_array);
      utf_print(summary_fd, output_record.line, output_record.length);
    }
    /* close insert_attr tokens tag */
    strcpy_output_record(1, (unsigned char *) "</custom_attributes>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
  }
  /* custom tag option */
  if (ninsert_tag)
  {
    /* open insert_tag tokens tag */
    strcpy_output_record(1, (unsigned char *) "<custom_tags>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
    /* tag name tokens */
    for (i = 0 ; i < ninsert_tag ; i++)
    {
      /* open tag tokens tag */
      (void) sprintf(temp_array, "<tag%ld", i + 1);
      strcpy_output_record(2, (unsigned char *) temp_array);
      if (insert_tag[i].nscope)
      {
        strcat_output_record((unsigned char *) " scope=\"");
        for (j = 0 ; j < insert_tag[i].nscope ; j++)
        {
          if (j)
            strcat_output_record((unsigned char *) ",");
          if (insert_tag[i].scope[j].min == insert_tag[i].scope[j].max)
            (void) sprintf(temp_array, "%ld", insert_tag[i].scope[j].min + 1);
          else if (insert_tag[i].scope[j].max < 0)
            (void) sprintf(temp_array, "%ld-MAX", insert_tag[i].scope[j].min + 1);
          else
            (void) sprintf(temp_array, "%ld-%ld", insert_tag[i].scope[j].min + 1, insert_tag[i].scope[j].max + 1);
          strcat_output_record((unsigned char *) temp_array);
          if (insert_tag[i].scope[j].restart)
            strcat_output_record((unsigned char *) "r");
        }
        strcat_output_record((unsigned char *) "\"");
      }
      strcat_output_record((unsigned char *) ">");
      strcat_output_record(newline);
      utf_print(summary_fd, output_record.line, output_record.length);
      /* open/close level tokens tag */
      if (insert_tag[i].level < 0)
      {
        strcpy_output_record(3, (unsigned char *) "<level>ALL</level>");
        strcat_output_record(newline);
      }
      else
      {
        (void) sprintf(temp_array, "<level>%d</level>%s", insert_tag[i].level, newline);
        strcpy_output_record(3, (unsigned char *) temp_array);
      }
      utf_print(summary_fd, output_record.line, output_record.length);
      /* open/close type tokens tag */
      strcpy_output_record(3, (unsigned char *) "<type>");
      if (insert_tag[i].type == TAG_INSERT_ELEMENT)
        strcat_output_record((unsigned char *) "element");
      else if (insert_tag[i].type == TAG_INSERT_FILE)
        strcat_output_record((unsigned char *) "file");
      else if (insert_tag[i].type == TAG_INSERT_XMLFILE)
        strcat_output_record((unsigned char *) "xmlfile");
      strcat_output_record((unsigned char *) "</type>");
      strcat_output_record(newline);
      utf_print(summary_fd, output_record.line, output_record.length);
      /* open/close reinsert tag */
      strcpy_output_record(3, (unsigned char *) "<reinsert>");
      if (insert_tag[i].reinsert)
        strcat_output_record((unsigned char *) "true");
      else
        strcat_output_record((unsigned char *) "false");
      strcat_output_record((unsigned char *) "</reinsert>");
      strcat_output_record(newline);
      utf_print(summary_fd, output_record.line, output_record.length);
      /* open/close name tokens tag */
      if (insert_tag[i].type == TAG_INSERT_XMLFILE)
      {
        (void) sprintf(temp_array, "<indent>%d</indent>%s", insert_tag[i].indent, newline);
        strcpy_output_record(3, (unsigned char *) temp_array);
      }
      else 
        mkelem(3, (unsigned char *) "name", strlen("name"), UTF_8, insert_tag[i].name, insert_tag[i].name_length, UTF_8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
      utf_print(summary_fd, output_record.line, output_record.length);
      /* open/close value tokens tag */
      if (insert_tag[i].type == TAG_INSERT_ELEMENT)
        mkelem(3, (unsigned char *) "value", strlen("value"), UTF_8, insert_tag[i].value, insert_tag[i].value_length, UTF_8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
      else
        mkelem(3, (unsigned char *) "filename", strlen("filename"), UTF_8, insert_tag[i].filename, xstrlen(insert_tag[i].filename), UTF_8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
      utf_print(summary_fd, output_record.line, output_record.length);
      /* close tag tokens tag */
      (void) sprintf(temp_array, "</tag%ld>%s", i + 1, newline);
      strcpy_output_record(2, (unsigned char *) temp_array);
      utf_print(summary_fd, output_record.line, output_record.length);
    }
    /* close insert_tag tokens tag */
    strcpy_output_record(1, (unsigned char *) "</custom_tags>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
  }
  /* BOM tag */
  if (sw_bom)
  {
    strcpy_output_record(1, (unsigned char *) "<byte_order_mark>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
    mkelem(2, (unsigned char *) "input", strlen("input"), UTF_8, input_record.bom_name, xstrlen(input_record.bom_name), UTF_8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
    utf_print(summary_fd, output_record.line, output_record.length);
    mkelem(2, (unsigned char *) "output", strlen("output"), UTF_8, output_record.encoding_name, xstrlen(output_record.encoding_name), UTF_8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
    utf_print(summary_fd, output_record.line, output_record.length);
    strcpy_output_record(1, (unsigned char *) "</byte_order_mark>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
  }
  /* Schema - open tag */
  strcpy_output_record(1, (unsigned char *) "<schema>");
  strcat_output_record(newline);
  utf_print(summary_fd, output_record.line, output_record.length);
  if (sw_schema == SCHEMA_DEFAULT)
  {
    /* Schema - default */
    strcpy_output_record(2, (unsigned char *) "<default>true</default>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
    record_cntr_processed = record_cntr_printed;
    /* Schema - open records tag */
    strcpy_output_record(2, (unsigned char *) "<records>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
  }
  else
  {
    /* Schema - not default */
    strcpy_output_record(2, (unsigned char *) "<default>false</default>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
    /* Schema - file name */
    strcpy_output_record(2, (unsigned char *) "<file>");
    strcat_output_record(schema_filename);
    strcat_output_record((unsigned char *) "</file>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
    /* Schema - type */
    if (sw_schema == SCHEMA_DTD)
      strcpy_output_record(2, (unsigned char *) "<interpreter>DTD</interpreter>");
    else if (sw_schema == SCHEMA_RNC)
      strcpy_output_record(2, (unsigned char *) "<interpreter>RNC</interpreter>");
    else if (sw_schema == SCHEMA_XSD)
      strcpy_output_record(2, (unsigned char *) "<interpreter>XSD</interpreter>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
    /* Schema - open parent element */
    strcpy_output_record(2, (unsigned char *) "<parent>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
    /* Schema - parent element name */
    strcpy_output_record(3, (unsigned char *) "<name>");
    strncat_output_record(schema_parent_name, parent_name_length);
    strcat_output_record((unsigned char *) "</name>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
    /* Schema - open elements tag */
    if (schema_match_start_level)
    {
      strcpy_output_record(3, (unsigned char *) "<matchmode>DEFAULT</matchmode>");
      strcat_output_record(newline);
      utf_print(summary_fd, output_record.line, output_record.length);
      strcpy_output_record(3, (unsigned char *) "<elements>");
      strcat_output_record(newline);
      utf_print(summary_fd, output_record.line, output_record.length);
      /* Schema - find all level 1 elements */
      for (i = 1 ; i < schema_parent.nelements ; i++)
      {
        for (k = 0 ; k < schema_parent.record->nfields ; k++)
        {
          if (xstrcmp(schema_element[i].name, schema_parent.record->field[k].name) == 0)
          {
            /* Schema - open element name tag */
            strcpy_output_record(4, (unsigned char *) "<");
            strcat_output_record(schema_element[i].name);
            strcat_output_record((unsigned char *) ">");
            strcat_output_record(newline);
            utf_print(summary_fd, output_record.line, output_record.length);
            /* Schema - match control */
            switch(schema_parent.record->field[k].match_control)
            { 
              case MATCH_ONE:          strcpy_output_record(5, (unsigned char *) "<match_control>MATCH_ONE</match_control>"); break;
              case MATCH_NONE_OR_ONE:  strcpy_output_record(5, (unsigned char *) "<match_control>MATCH_NONE_OR_ONE</match_control>"); break;
              case MATCH_NONE_OR_MANY: strcpy_output_record(5, (unsigned char *) "<match_control>MATCH_NONE_OR_MANY</match_control>"); break;
              case MATCH_ONE_OR_MANY:  strcpy_output_record(5, (unsigned char *) "<match_control>MATCH_ONE_OR_MANY</match_control>"); break;
            } 
            strcat_output_record(newline);
            utf_print(summary_fd, output_record.line, output_record.length);
            /* Schema - instances */
            for (j = 0 ; j < schema_element[i].nrecords ; j++)
            {
              /* Schema - open instance tag */
              (void) sprintf(temp_array, "<instance%ld>%s", j + 1, newline);
              strcpy_output_record(5, (unsigned char *) temp_array);
              utf_print(summary_fd, output_record.line, output_record.length);
              /* Schema - instance fields */
              (void) sprintf(temp_array, "<fields>%d</fields>%s", schema_element[i].record[j]->nfields, newline);
              strcpy_output_record(6, (unsigned char *) temp_array);
              utf_print(summary_fd, output_record.line, output_record.length);
              /* Schema - close instance tag */
              (void) sprintf(temp_array, "</instance%ld>%s", j + 1, newline);
              strcpy_output_record(5, (unsigned char *) temp_array);
              utf_print(summary_fd, output_record.line, output_record.length);
            }
            /* Schema - close element name tag */
            strcpy_output_record(4, (unsigned char *) "</");
            strcat_output_record(schema_element[i].name);
            strcat_output_record((unsigned char *) ">");
            strcat_output_record(newline);
            utf_print(summary_fd, output_record.line, output_record.length);
            /* element names are unique, don't compare any more */
            break;
          }
        }
      }
      /* Schema - close elements tag */
      strcpy_output_record(3, (unsigned char *) "</elements>");
      strcat_output_record(newline);
      utf_print(summary_fd, output_record.line, output_record.length);
    }
    else
    {
      strcpy_output_record(3, (unsigned char *) "<matchmode>DIRECT</matchmode>");
      strcat_output_record(newline);
      utf_print(summary_fd, output_record.line, output_record.length);
    }
    /* Schema - close parent element */
    strcpy_output_record(2, (unsigned char *) "</parent>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
    /* Schema - open records tag */
    strcpy_output_record(2, (unsigned char *) "<records>");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
    /* Schema - open elements tag */
    if (schema_match_start_level)
    {
      strcpy_output_record(3, (unsigned char *) "<elements>");
      strcat_output_record(newline);
      utf_print(summary_fd, output_record.line, output_record.length);
      /* Schema - elements */
      for (i = 1 ; i < schema_parent.nelements ; i++)
      {
        for (k = 0 ; k < schema_parent.record->nfields ; k++)
        {
          if (xstrcmp(schema_element[i].name, schema_parent.record->field[k].name) == 0)
          {
            /* Schema - open element name tag */
            strcpy_output_record(4, (unsigned char *) "<");
            strcat_output_record(schema_element[i].name);
            strcat_output_record((unsigned char *) ">");
            strcat_output_record(newline);
            utf_print(summary_fd, output_record.line, output_record.length);
            /* Schema - element total */
            (void) sprintf(temp_array, "<total>%d</total>%s", schema_element[i].element_count, newline);
            strcpy_output_record(5, (unsigned char *) temp_array);
            utf_print(summary_fd, output_record.line, output_record.length);
            /* Schema - element instance total */
            for (j = 0 ; j < schema_element[i].nrecords ; j++)
            {
              (void) sprintf(temp_array, "<instance%ld>%d</instance%ld>%s", j + 1, schema_element[i].record[j]->record_count, j + 1, newline);
              strcpy_output_record(5, (unsigned char *) temp_array);
              utf_print(summary_fd, output_record.line, output_record.length);
            }
            /* Schema - close element name tag */
            strcpy_output_record(4, (unsigned char *) "</");
            strcat_output_record(schema_element[i].name);
            strcat_output_record((unsigned char *) ">");
            strcat_output_record(newline);
            utf_print(summary_fd, output_record.line, output_record.length);
    
            record_cntr_processed += schema_element[i].element_count;

            break;
          }
        }
      }
      /* Schema - close elements tag */
      strcpy_output_record(3, (unsigned char *) "</elements>");
      strcat_output_record(newline);
      utf_print(summary_fd, output_record.line, output_record.length);
    }
    else
    {
      if (xstrcmp(schema_match_start_name, (unsigned char *) "0") == 0)
        record_cntr_processed += schema_element[0].element_count;
      else
      {
        for (i = 0 ; i < schema_parent.nelements ; i++)
        {
          if (xstrcmp(schema_match_start_name, schema_element[i].name) == 0)
          {
            record_cntr_processed += schema_element[i].element_count;
            break; /* for/i */
          }
        } /* end for/i */
      }
    }
  }
  /* records read */
  (void) sprintf(temp_array, "<read>%ld</read>%s", record_cntr_read, newline);
  strcpy_output_record(3, (unsigned char *) temp_array);
  utf_print(summary_fd, output_record.line, output_record.length);
  /* records expelled */
  (void) sprintf(temp_array, "<expelled>%ld</expelled>%s", record_cntr_expelled, newline);
  strcpy_output_record(3, (unsigned char *) temp_array);
  utf_print(summary_fd, output_record.line, output_record.length);
  /* records matched */
  (void) sprintf(temp_array, "<matched>%ld</matched>%s", record_cntr_processed, newline);
  strcpy_output_record(3, (unsigned char *) temp_array);
  utf_print(summary_fd, output_record.line, output_record.length);
  /* records unmatched */
  (void) sprintf(temp_array, "<unmatched>%ld</unmatched>%s", record_cntr_unmatched, newline);
  strcpy_output_record(3, (unsigned char *) temp_array);
  utf_print(summary_fd, output_record.line, output_record.length);
  /* records blank */
  (void) sprintf(temp_array, "<ignoredblank>%ld</ignoredblank>%s", record_cntr_blank, newline);
  strcpy_output_record(3, (unsigned char *) temp_array);
  utf_print(summary_fd, output_record.line, output_record.length);
  /* records printed */
  (void) sprintf(temp_array, "<printed>%ld</printed>%s", record_cntr_printed, newline);
  strcpy_output_record(3, (unsigned char *) temp_array);
  utf_print(summary_fd, output_record.line, output_record.length);
  /* close records tag */
  strcpy_output_record(2, (unsigned char *) "</records>");
  strcat_output_record(newline);
  utf_print(summary_fd, output_record.line, output_record.length);
  /* Schema - close tag */
  strcpy_output_record(1, (unsigned char *) "</schema>");
  strcat_output_record(newline);
  utf_print(summary_fd, output_record.line, output_record.length);
  /* close summary */
  strcpy_output_record(0, (unsigned char *) "</");
  strncat_output_record(schema_parent_name, parent_name_length);
  strcat_output_record((unsigned char *) "_summary>");
  strcat_output_record(newline);
  utf_print(summary_fd, output_record.line, output_record.length);

  /* close comment token */
  if (sw_summary == SUMMARY_COMMENT)
  {
    strcpy_output_record(0, (unsigned char *) "end-of-summary-comment -->");
    strcat_output_record(newline);
    utf_print(summary_fd, output_record.line, output_record.length);
  }
}
